// COMPlugin.cpp
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "COMPlugin.h"


namespace PCC
{
    namespace Plugins
    {
        //
        // Constructor. Will create an instance of the COM plugin.
        //
        // @param p_CLSID ID of COM co-class that implements the plugin.
        //
        COMPlugin::COMPlugin(const CLSID& p_CLSID)
            : m_Id(p_CLSID),
              m_cpPlugin(),
              m_cpPluginGroup(),
              m_cpPluginState(),
              m_Description()
        {
            // Immediately create the plugin instance and make sure it works.
            HRESULT hRes = m_cpPlugin.CoCreateInstance(p_CLSID);
            if (FAILED(hRes)) {
                throw COMPluginError(hRes);
            }
            if (m_cpPlugin == NULL) {
                throw COMPluginError(E_FAIL);
            }

            // Immediately query for other interface implementations. This can fail.
            m_cpPluginGroup = m_cpPlugin;
            m_cpPluginState = m_cpPlugin;

            // Immediately get plugin description. It must work and return a non-empty string.
            CComBSTR bstrDescription;
            hRes = m_cpPlugin->get_Description(&bstrDescription);
            if (FAILED(hRes)) {
                throw COMPluginError(hRes);
            }
            if (bstrDescription == NULL || bstrDescription.Length() == 0) {
                throw COMPluginError(E_UNEXPECTED);
            }
            m_Description = bstrDescription.m_str;
        }

        //
        // Returns the plugin's unique identifier.
        //
        // @return Unique identifier.
        //
        const GUID& COMPlugin::Id() const
        {
            return m_Id;
        }

        //
        // Returns the plugin description.
        //
        // @returns Plugin description.
        //
        std::wstring COMPlugin::Description() const
        {
            // Return cached description.
            return m_Description;
        }

        //
        // Returns the plugin help text.
        //
        // @returns Plugin description.
        //
        std::wstring COMPlugin::HelpText() const
        {
            // Get plugin help text. If it doesn't work, no worry.
            std::wstring helpText;
            CComBSTR bstrHelpText;
            HRESULT hRes = m_cpPlugin->get_HelpText(&bstrHelpText);
            if (SUCCEEDED(hRes) && bstrHelpText != NULL && bstrHelpText.Length() > 0) {
                helpText = bstrHelpText.m_str;
            }

            // Return help text or an empty string.
            return helpText;
        }

        //
        // Returns whether plugin should be enabled in the contextual menu.
        //
        // @param p_ParentPath Path of the parent directory of files that
        //                     triggered the contextual menu.
        // @param p_File Path of one file that was selected.
        // @returns Whether plugin should be enabled or not in the contextual menu.
        //
        bool COMPlugin::Enabled(const std::wstring& p_ParentPath,
                                const std::wstring& p_File) const
        {
            // Check if plugin supports state changes. Otherwise assume it is enabled.
            bool enabled = true;
            if (m_cpPluginState != NULL) {
                // Ask plugin if it should be enabled. If an error
                // code is returned, do not enable the plugin.
                CComBSTR bstrParentPath(p_ParentPath.c_str());
                CComBSTR bstrFile(p_File.c_str());
                VARIANT_BOOL bEnabled = VARIANT_FALSE;
                HRESULT hRes = m_cpPluginState->Enabled(bstrParentPath, bstrFile, &bEnabled);
                enabled = SUCCEEDED(hRes) && hRes != S_FALSE && bEnabled != VARIANT_FALSE;
            }
            return enabled;
        }

        //
        // Transforms the given path using the plugin.
        //
        // @param p_File Full path to file.
        // @returns Transformed path.
        //
        std::wstring COMPlugin::GetPath(const std::wstring& p_File) const
        {
            // Call method and make sure it works.
            // Note that it is legal for the method to return NULL or an empty string.
            CComBSTR bstrPath(p_File.c_str());
            CComBSTR bstrNewPath;
            HRESULT hRes = m_cpPlugin->GetPath(bstrPath, &bstrNewPath);
            if (FAILED(hRes)) {
                throw COMPluginError(hRes);
            }

            // Convert path if provided, otherwise reuse same one.
            std::wstring newPath(p_File);
            if (bstrNewPath != NULL && bstrNewPath.Length() > 0) {
                newPath = bstrNewPath.m_str;
            }
            return newPath;
        }

        //
        // Returns the ID of the group to which the plugin belongs.
        // If the plugin does not implementing grouping, will return 0.
        //
        // @return Group ID of plugin, or 0 if not supported.
        //
        ULONG COMPlugin::GroupId() const
        {
            ULONG groupId = 0;
            if (m_cpPluginGroup != NULL) {
                ULONG groupId2;
                HRESULT hRes = m_cpPluginGroup->get_GroupId(&groupId2);
                if (SUCCEEDED(hRes)) {
                    groupId = groupId2;
                }
            }
            return groupId;
        }

        //
        // Returns the position of the plugin within its group.
        // If the plugin does not implement grouping, will return 0.
        //
        // @return Plugin position in group, or 0 if not supported.
        //
        ULONG COMPlugin::GroupPosition() const
        {
            ULONG groupPos = 0;
            if (m_cpPluginGroup != NULL) {
                ULONG groupPos2;
                HRESULT hRes = m_cpPluginGroup->get_GroupPosition(&groupPos2);
                if (SUCCEEDED(hRes)) {
                    groupPos = groupPos2;
                }
            }
            return groupPos;
        }

        //
        // Constructor with HRESULT.
        //
        // @param p_Result Result of COM call that went wrong.
        //
        COMPluginError::COMPluginError(const HRESULT p_Result)
            : std::exception(),
              m_Result(p_Result)
        {
        }

        //
        // Copy constructor.
        //
        // @param p_Exception Exception to copy.
        //
        COMPluginError::COMPluginError(const COMPluginError& p_Exception)
            : std::exception(p_Exception),
              m_Result(p_Exception.m_Result)
        {
        }

        //
        // Destructor.
        //
        COMPluginError::~COMPluginError()
        {
        }

        //
        // Assignment operator.
        //
        // @param p_Exception Exception to copy.
        //
        COMPluginError& COMPluginError::operator=(const COMPluginError& p_Exception)
        {
            if (this != &p_Exception) {
                std::exception::operator=(p_Exception);
                m_Result = p_Exception.m_Result;
            }
            return *this;
        }

        //
        // Returns a textual description of the exception.
        //
        // @returns Exception textual description.
        //
        const char* COMPluginError::what() const
        {
            return "COMPluginError";
        }

    } // namespace Plugins

} // namespace PCC
