// PipelinePlugin.cpp
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PipelinePlugin.h"

#include <PluginPipelineDecoder.h>

#include <assert.h>


namespace PCC
{
    namespace Plugins
    {
        //
        // Constructor.
        //
        // @param p_PluginId ID of this pipeline plugin.
        // @param p_PluginDescription Description of the pipeline plugin.
        // @param p_EncodedElements String containing encoded pipeline elements.
        //
        PipelinePlugin::PipelinePlugin(const GUID& p_PluginId,
                                       const std::wstring& p_PluginDescription,
                                       const std::wstring& p_EncodedElements)
            : m_Id(p_PluginId),
              m_Description(p_PluginDescription),
              m_spPipeline()
        {
            // Try decoding the encoded pipeline. If that fails, we'll
            // keep the plugin alive but it will be disabled.
            try {
                m_spPipeline = PipelineSP(new PCC::Pipeline(p_EncodedElements));
            } catch (const PCC::InvalidPipelineException&) {
                assert(m_spPipeline.get() == 0);
            }
        }

        //
        // Returns the ID of the pipeline plugin, allowing it to be uniquely identified.
        //
        // @return Plugin ID.
        //
        const GUID& PipelinePlugin::Id() const
        {
            return m_Id;
        }

        //
        // Returns a description of the pipeline plugin, to be used to display it
        // in the contextual menu.
        //
        // @return Plugin description.
        //
        std::wstring PipelinePlugin::Description() const
        {
            return m_Description;
        }

        //
        // Checks whether the plugin should be enabled or not in the contextual menu.
        // In our case, if we failed to decode our pipeline or if the pipeline complains,
        // we'll show up as disabled.
        //
        // @param p_ParentPath Path of the parent folder of all files; unused.
        // @param p_File Path of one selected file; unused.
        //
        bool PipelinePlugin::Enabled(const std::wstring& p_ParentPath,
                                     const std::wstring& p_File) const
        {
            return m_spPipeline.get() != 0 &&
                   m_spPipeline->ShouldBeEnabledFor(p_ParentPath, p_File);
        }

        //
        // Modifies a path using all elements in our pipeline.
        //
        // @param p_File Path of file to modify.
        // @return Modified path.
        //
        std::wstring PipelinePlugin::GetPath(const std::wstring& p_File) const
        {
            assert(m_spPipeline.get() != 0);

            std::wstring modifiedPath(p_File);
            m_spPipeline->ModifyPath(modifiedPath);
            return modifiedPath;
        }

    } // namespace Plugins

} // namespace PCC
