// PathCopyCopyContextMenuExt.h
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once
#include "resource.h"       // main symbols

#include "PathCopyCopy_i.h"
#include <PathCopyCopyTypes.h>
#include <PathCopyCopySettings.h>
#include <PluginRegistry.h>
#include <Nullable.h>
#include <CriticalSection.h>
#include <StImage.h>


//
// CPathCopyCopyContextMenuExt
//
// Shell contextual menu extension that allows the user to copy the path of a file
// or a group of files into various formats.
//
class ATL_NO_VTABLE CPathCopyCopyContextMenuExt :
    public CComObjectRootEx<CComSingleThreadModel>,
    public CComCoClass<CPathCopyCopyContextMenuExt, &CLSID_PathCopyCopyContextMenuExt>,
    public IPathCopyCopyContextMenuExt,
    public IShellExtInit,
    public IContextMenu
{
public:
    CPathCopyCopyContextMenuExt();

    DECLARE_REGISTRY_RESOURCEID(IDR_PATHCOPYCOPYCONTEXTMENUEXT)

    DECLARE_NOT_AGGREGATABLE(CPathCopyCopyContextMenuExt)

    BEGIN_COM_MAP(CPathCopyCopyContextMenuExt)
        COM_INTERFACE_ENTRY(IPathCopyCopyContextMenuExt)
        COM_INTERFACE_ENTRY(IShellExtInit)
        COM_INTERFACE_ENTRY(IContextMenu)
    END_COM_MAP()

    DECLARE_PROTECT_FINAL_CONSTRUCT()

    HRESULT FinalConstruct()
    {
        // Save our registry as the local registry so that plugins can access it.
        PCC::PluginRegistry::SetLocalRegistry(&m_Registry);

        return S_OK;
    }

    void FinalRelease()
    {
        // Remove this instance from the map of instances modifying menus.
        RemoveFromExtToMenu();

        // Reset local registry.
        PCC::PluginRegistry::SetLocalRegistry(0);

        // Check for updates.
        CheckForUpdates();
    }

public:
    // IPathCopyCopyContextMenuExt methods
    STDMETHOD(RegisterPlugin)(REFCLSID p_CLSID);
    STDMETHOD(UnregisterPlugin)(REFCLSID p_CLSID);

    // IShellExtInit methods
    STDMETHOD(Initialize)(PCIDLIST_ABSOLUTE p_pFolderPIDL, IDataObject *p_pDataObject, HKEY p_hKeyFileClass);

    // IContextMenu methods
    STDMETHOD(QueryContextMenu)(HMENU p_hMenu, UINT p_Index, UINT p_FirstCmdId,
                                UINT p_LastCmdId, UINT p_Flags);
    STDMETHOD(InvokeCommand)(CMINVOKECOMMANDINFO* p_pCommandInfo);
    STDMETHOD(GetCommandString)(UINT_PTR p_CmdId, UINT p_Flags, UINT* p_pReserved,
                                LPSTR p_pBuffer, UINT p_BufferSize);

private:
    typedef std::map<UINT, PCC::PluginSP>   CmdIdPluginM;   // Map of plugins by command ID.

    typedef std::pair<CPathCopyCopyContextMenuExt*, HMENU>  ExtToMenuPair;  // Pairing an extension instance to an HMENU.
    typedef std::vector<ExtToMenuPair>                      ExtToMenuPairV; // Vector mapping extension instances to HMENUs.

    typedef std::auto_ptr<StImage>  StImageAP;              // AutoPtr to a Win32 image wrapper.

    PCC::PluginRegistry m_Registry;             // Registry of all plugins.
    PCC::Settings       m_Settings;             // Object to access program settings.

    PCC::FilesV         m_vFiles;               // Files selected in Shell.
    std::wstring        m_ParentPath;           // Path of the parent directory of all files selected.

    Nullable<UINT>      m_FirstCmdId;           // ID of first command menu item.
    Nullable<UINT>      m_SubMenuCmdId;         // ID of the menu item that opens our submenu.
    Nullable<UINT>      m_SettingsCmdId;        // ID of the menu item that triggers the options.
    CmdIdPluginM        m_mPluginsByCmdId;      // Map storing plugins by their command IDs.

    StImageAP           m_apPCCIcon;            // Holds the PCC icon that can be used in menus.

    static ExtToMenuPairV
                        s_vExtToMenus;          // Static map keeping track of which extensions modified which menu.
    static CriticalSection
                        s_ExtToMenusLock;       // Lock to protect the static map.

    HRESULT             AddPluginToMenu(const GUID& p_PluginId,
                                        const PCC::PluginSPV& p_vspPlugins,
                                        HMENU const p_hMenu,
                                        UINT& p_rCmdId,
                                        UINT& p_rPosition);
    HRESULT             AddPluginToMenu(const PCC::PluginSP& p_spPlugin,
                                        HMENU const p_hMenu,
                                        const bool p_UsePCCIcon,
                                        UINT& p_rCmdId,
                                        UINT& p_rPosition);

    HBITMAP             GetPCCIcon();

    void                RemoveFromExtToMenu();
    void                CheckForUpdates();
};

OBJECT_ENTRY_AUTO(__uuidof(PathCopyCopyContextMenuExt), CPathCopyCopyContextMenuExt)
