// PluginPipelineElements.h
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once

#include <PluginPipeline.h>

#include <regex>


namespace PCC
{
    //
    // QuotesPipelineElement
    //
    // Pipeline element that adds quotes around the path.
    //
    class QuotesPipelineElement : public PipelineElement
    {
    public:
                        QuotesPipelineElement();

        virtual void    ModifyPath(std::wstring& p_rPath) const;
    };

    //
    // BackToForwardSlashesPipelineElement
    //
    // Pipeline element that changes all backslashes in a path to
    // forward slashes.
    //
    class BackToForwardSlashesPipelineElement : public PipelineElement
    {
    public:
                        BackToForwardSlashesPipelineElement();

        virtual void    ModifyPath(std::wstring& p_rPath) const;
    };

    //
    // ForwardToBackslashesPipelineElement
    //
    // Pipeline element that changes all forward slashes in a path
    // to backslashes.
    //
    class ForwardToBackslashesPipelineElement : public PipelineElement
    {
    public:
                        ForwardToBackslashesPipelineElement();

        virtual void    ModifyPath(std::wstring& p_rPath) const;
    };

    //
    // FindReplacePipelineElement
    //
    // Pipeline element that replaces all instances of one string
    // in the path with another string.
    //
    class FindReplacePipelineElement : public PipelineElement
    {
    public:
                        FindReplacePipelineElement(const std::wstring& p_OldValue,
                                                   const std::wstring& p_NewValue);

        virtual void    ModifyPath(std::wstring& p_rPath) const;

    private:
        std::wstring    m_OldValue;     // Value to replace.
        std::wstring    m_NewValue;     // Replacement value.
    };

    //
    // RegexPipelineElement
    //
    // Pipeline element that performs find/replace operations
    // using regular expressions.
    //
    class RegexPipelineElement : public PipelineElement
    {
    public:
                        RegexPipelineElement(const std::wstring& p_Regex,
                                             const std::wstring& p_Format,
                                             const bool p_IgnoreCase);

        virtual void    ModifyPath(std::wstring& p_rPath) const;
        virtual bool    ShouldBeEnabledFor(const std::wstring& p_ParentPath,
                                           const std::wstring& p_File) const;

    private:
        std::wstring    m_Regex;        // Regex to use to find matches.
        std::wstring    m_Format;       // Format of replacement string.
        bool            m_IgnoreCase;   // Whether to ignore case when looking for matches.
        mutable std::auto_ptr<std::wregex>
                        m_apRegex;      // Regex object to use to perform lookups.
        mutable bool    m_RegexTested;  // true if we tried to create m_apRegex.

        void            InitRegex() const;

        // THESE METHODS ARE NOT IMPLEMENTED
                        RegexPipelineElement(const RegexPipelineElement&);
        RegexPipelineElement&
                        operator=(const RegexPipelineElement&);
    };

    //
    // ApplyPluginPipelineElement
    //
    // Pipeline element that fetches an existing plugin via its
    // ID and calls its GetPath method to modify the path.
    //
    class ApplyPluginPipelineElement : public PipelineElement
    {
    public:
        explicit        ApplyPluginPipelineElement(const GUID& p_PluginId);

        virtual void    ModifyPath(std::wstring& p_rPath) const;
        virtual bool    ShouldBeEnabledFor(const std::wstring& p_ParentPath,
                                           const std::wstring& p_File) const;

    private:
        GUID            m_PluginId;     // ID of plugin to apply.
    };

} // namespace PCC
