// Singleton.h
// (c) 2009, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once

#include <CriticalSection.h>

#include <memory>


//
// Singleton
//
// Template base class to create singleton classes.
// Inherit from this class and use the published Instance method to get the instance.
//
// Ex:
// class S : public Singleton<S>
// {
//     friend class Singleton<S>;
//     S() // private constructor
//     ...
//
// In your .cpp file, use the macro IMPLEMENT_SINGLETON to complete the implementation.
//
// Ex:
// IMPLEMENT_SINGLETON(S);
//
// Note: the Instance method of such singletons is thread-safe.
//
template<typename T>
class Singleton
{
    friend T;
public:
                        //
                        // Destructor.
                        //
    virtual             ~Singleton()
                        {
                        }

                        //
                        // Returns the global instance of the class.
                        //
                        // @returns Global singleton instance.
                        //
    static T&           Instance()
                        {
                            // Create on the first call.
                            if (s_Instance.get() == 0) {
                                CriticalSection::Locker lock(s_Lock);
                                if (s_Instance.get() == 0) {
                                    s_Instance.reset(new T);
                                }
                            }
                            return *s_Instance;
                        }

private:
    static std::auto_ptr<T>
                        s_Instance;     // Global instance. Defined using the IMPLEMENT_SINGLETON macro for a specific class.
    static CriticalSection
                        s_Lock;         // Lock to protect creation of the global instance.

                        //
                        // Private constructor.
                        //
                        Singleton()
                        {
                        }

    // THESE METHOD ARE NOT IMPLEMENTED
                        Singleton(const Singleton<T>&);
    Singleton<T>&       operator=(const Singleton<T>&);
};

//
// Macro used to define a singleton's static instance variable.
// Use this in your .cpp file when using the Singleton template class.
//
#define IMPLEMENT_SINGLETON(T)                      \
    std::auto_ptr<T> Singleton<T>::s_Instance;      \
    CriticalSection  Singleton<T>::s_Lock
