// StLock.h
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once


//
// StLock<T>
//
// Stack-based class that can acquire a lock on a synchronization object in a block of code.
// The constructor calls Lock() and the destructor calls Unlock().
//
template<typename T>
class StLock
{
public:
                        //
                        // Constructor with object. Calls Lock() on the synchronization object.
                        //
                        StLock(T& p_rObj)
                            : m_pObj(&p_rObj)
                        {
                            m_pObj->Lock();
                        }

                        //
                        // Constructor with pointer. Calls Lock() on the synchronization object if set.
                        //
                        StLock(T* const p_pObj)
                            : m_pObj(p_pObj)
                        {
                            if (m_pObj != 0) {
                                m_pObj->Lock();
                            }
                        }

                        //
                        // Destructor. Calls Unlock() on the synchronization object if set by the constructor.
                        //
                        ~StLock()
                        {
                            if (m_pObj != 0) {
                                m_pObj->Unlock();
                            }
                        }

private:
    T* const            m_pObj;     // Pointer to synchronization object.

    // THESE METHODS ARE NOT IMPLEMENTED.
                        StLock(const StLock&);
    StLock&             operator=(const StLock&);
};
