// StRegKey.h
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once


//
// StRegKey
//
// Stack-based class for proper disposal of a Windows registry key handle.
// The key will be closed by the destructor via RegCloseKey.
//
class StRegKey
{
public:
                        //
                        // Constructor.
                        // Initialized with no key handle.
                        //
                        StRegKey()
                            : m_hKey(0)
                        {
                        }

                        //
                        // Constructor.
                        // Assumes ownership of the given registry key. The
                        // handle will be closed in destructor via RegCloseKey.
                        //
                        // @param p_hKey Handle of key to save.
                        //
    explicit            StRegKey(const HKEY p_hKey)
                            : m_hKey(p_hKey)
                        {
                        }

                        //
                        // Destructor.
                        // Closes the key handle via RegCloseKey.
                        //
                        ~StRegKey()
                        {
                            ::RegCloseKey(m_hKey);
                        }

                        //
                        // Returns the handle of our registry key.
                        //
                        // @return Registry key handle.
                        //
    HKEY                Get() const
                        {
                            return m_hKey;
                        }

                        //
                        // Returns the handle of our registry key.
                        //
                        // @return Registry key handle.
                        //
                        operator HKEY()
                        {
                            return m_hKey;
                        }

                        //
                        // Returns a pointer to our registry key handle.
                        // Allows the object to be used as an HKEY in calls:
                        //
                        // StRegKey hKey;
                        // RegOpenKey(..., &hKey);
                        //
    HKEY*               operator&()
                        {
                            return &m_hKey;
                        }

private:
    HKEY                m_hKey;     // Handle of registry key.

    // THESE METHODS ARE NOT IMPLEMENTED.
                        StRegKey(const StRegKey&);
    StRegKey&           operator=(const StRegKey&);
};