// TlsSlot.h
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#pragma once


//
// TlsSlot
//
// Wrapper for a Windows TLS slot that can store a per-thread value. Once
// created, each thread accessing the value will have its own storage.
//
class TlsSlot
{
public:
                        TlsSlot();
    virtual             ~TlsSlot();

                        //
                        // Returns the value stored in this TLS slot
                        // for the calling thread.
                        //
                        // @return Thread-specific value for this slot.
                        //
    void*               Get() const
                        {
                            return ::TlsGetValue(m_Index);
                        }

                        //
                        // Changes the value of this TLS slot only for
                        // the calling thread.
                        //
                        // @param p_pValue New value to store.
                        //
    void                Set(void* const p_pValue)
                        {
                            ::TlsSetValue(m_Index, p_pValue);
                        }

private:
    DWORD               m_Index;    // TLS slot index needed to get/set values.

    // THESE METHODS ARE NOT IMPLEMENTED
                        TlsSlot(const TlsSlot&);
    TlsSlot&            operator=(const TlsSlot&);
};

//
// TlsSlotT<>
//
// Templated version of the TLS slot wrapper that provides
// strongly-typed getter and setter. The type used for this
// template must fit in a size_t.
//
template<typename T>
class TlsSlotT : public TlsSlot
{
public:
                        //
                        // Constructor.
                        //
                        TlsSlotT()
                            : TlsSlot()
                        {
                        }

                        //
                        // Returns the value stored in this TLS slot
                        // for the calling thread.
                        //
                        // @return Thread-specific value for this slot.
                        //
    typename T          Get() const
                        {
                            return (typename T) TlsSlot::Get();
                        }

                        //
                        // Changes the value of this TLS slot only for
                        // the calling thread.
                        //
                        // @param p_pValue New value to store.
                        //
    void                Set(typename T const p_Value)
                        {
                            TlsSlot::Set((void*) p_Value);
                        }

private:
    // THESE METHODS ARE NOT IMPLEMENTED
                        TlsSlotT(const TlsSlotT&);
    TlsSlotT&           operator=(const TlsSlotT&);
};
