// PathCopyCopyContextMenuExt.cpp
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PathCopyCopyContextMenuExt.h"

#include "StStgMedium.h"
#include "StClipboard.h"
#include "StGlobalBlock.h"
#include "StGlobalLock.h"
#include "ArrayAutoPtr.h"
#include "PathCopyCopyPluginsRegistration.h"
#include <PluginUtils.h>
#include "PathCopyCopySettingsApp.h"
#include "dllmain.h"

#include <DefaultPlugin.h>

#include <set>
#include <algorithm>


namespace {

const wchar_t   FILE_LINE_SEPARATOR[]   = L"\r\n";          // Separator used between lines when copying multiple file names.

}

// CPathCopyCopyContextMenuExt

// Static members
CPathCopyCopyContextMenuExt::ExtToMenuPairV CPathCopyCopyContextMenuExt::s_vExtToMenus;
CriticalSection                             CPathCopyCopyContextMenuExt::s_ExtToMenusLock;

//
// Constructor.
//
CPathCopyCopyContextMenuExt::CPathCopyCopyContextMenuExt()
    : m_Registry(),
      m_Settings(),
      m_vFiles(),
      m_ParentPath(),
      m_FirstCmdId(),
      m_SubMenuCmdId(),
      m_SettingsCmdId(),
      m_mPluginsByCmdId(),
      m_apPCCIcon()
{
}

//
// IPathCopyCopyContextMenuExt::RegisterPlugin
//
// Registers a plugin component in the Path Copy Copy registry. The plugin
// will be displayed in the Path Copy Copy contextual menu.
//
// @param p_pCLSID ID of plugin COM class, as a string.
// @returns S_OK if successful, S_FALSE if the plugin was already registered, otherwise an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::RegisterPlugin(
    REFCLSID p_CLSID)
{
    HRESULT hRes = E_FAIL;
    try {
        hRes = m_Settings.RegisterCOMPlugin(p_CLSID) ? S_OK : S_FALSE;
    } catch (PCC::SettingsException& se) {
        hRes = HRESULT_FROM_WIN32(se.ErrorCode());
    } catch (...) {
        hRes = E_UNEXPECTED;
    }
    return hRes;
}

//
// IPathCopyCopyContextMenuExt::UnregisterPlugin
//
// Removes a plugin component from the Path Copy Copy registry.
//
// @param p_pCLSID ID of plugin COM class, as a string.
// @returns S_OK if successful, S_FALSE if the component was not registered, otherwise an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::UnregisterPlugin(
    REFCLSID p_CLSID)
{
    HRESULT hRes = E_FAIL;
    try {
        hRes = m_Settings.UnregisterCOMPlugin(p_CLSID) ? S_OK : S_FALSE;
    } catch (PCC::SettingsException& se) {
        hRes = HRESULT_FROM_WIN32(se.ErrorCode());
    } catch (...) {
        hRes = E_UNEXPECTED;
    }
    return hRes;
}

//
// IShellExtInit::Initialize
//
// Called by the shell to initialize our contextual menu extension.
// We need to use this opportunity to look at the files/folders to act upon.
//
// @param p_pFolderPIDL Pointer to ITEMIDLIST representing selected folder; unused.
// @param p_pDataObject Pointer to data object containing information about selected files/folders.
// @param p_hKeyFileClass Handle to file class key; unused.
// @return S_OK if successful, otherwise an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::Initialize(
    PCIDLIST_ABSOLUTE p_pFolderPIDL,
    IDataObject *p_pDataObject,
    HKEY p_hKeyFileClass)
{
    HRESULT hRes = S_OK;

    try {
        // Register all known plugins in our registry.
        PCC::PluginsRegistration::Register(m_Registry, m_Settings);

        // Make sure we have a data object.
        if (p_pDataObject != 0) {
            // Extract HDROP from data object.
            StStgMedium stgMedium;
            FORMATETC formatEtc = {CF_HDROP, 0, DVASPECT_CONTENT, -1, TYMED_HGLOBAL};
            if (SUCCEEDED(p_pDataObject->GetData(&formatEtc, &stgMedium))) {
                // Get number of files included in the selection.
                UINT fileCount = ::DragQueryFileW(
                    static_cast<HDROP>(stgMedium.Get().hGlobal), 0xFFFFFFFF, 0, 0);
                if (fileCount > 0) {
                    // Pre-allocate space in vector to store files.
                    m_vFiles.reserve(fileCount);

                    // Get each file in turn.
                    wchar_t buffer[MAX_PATH + 1];
                    for(UINT i = 0; i < fileCount; ++i) {
                        UINT copiedCount = ::DragQueryFileW(static_cast<HDROP>(stgMedium.Get().hGlobal),
                            i, buffer, sizeof(buffer) / sizeof(wchar_t));
                        m_vFiles.push_back(std::wstring(buffer, copiedCount));
                    }

                    // Extract the parent path of the first file. We'll assume that all
                    // files have the same parent. This might not be strictly true in all
                    // cases (for example, in a custom shell view) but we're only using it
                    // for validation purposes, so it's good enough.
                    if (!m_vFiles.empty()) {
                        m_ParentPath = m_vFiles.front();
                        PCC::PluginUtils::ExtractFolderFromPath(m_ParentPath);
                    }
                } else {
                    // It's difficult to display a menu item without files to act upon.
                    hRes = E_FAIL;
                }
            }
        } else if (p_pFolderPIDL != 0) {
            // No data object, but maybe it's because user clicked on a folder's
            // background. Get the folder path from the ID list.
            wchar_t buffer[MAX_PATH + 1];
            if (::SHGetPathFromIDList(p_pFolderPIDL, buffer) != FALSE) {
                m_vFiles.push_back(std::wstring(buffer));
            } else {
                // Nothing like that either, problem.
                hRes = E_FAIL;
            }
        } else {
            hRes = E_POINTER;
        }
    } catch (...) {
        hRes = E_UNEXPECTED;
    }

    return hRes;
}

//
// IContextMenu::QueryContextMenu
//
// Invoked by the shell to populate a contextual menu.
// We need to use this opportunity to add our menu items.
//
// @param p_hMenu Handle to the contextual menu to populate.
// @param p_Index Index in the menu where to add items.
// @param p_FirstCmdId First available command ID for our commands.
// @param p_LastCmdId Last available command ID for our commands.
// @param p_Flags Flags indicating how the menu can be changed; see MSDN for details.
// @return If successful, a success code with code value set to the largest command ID
//         used, plus one; otherwise, an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::QueryContextMenu(
    HMENU p_hMenu,
    UINT p_Index,
    UINT p_FirstCmdId,
    UINT p_LastCmdId,
    UINT p_Flags)
{
    HRESULT hRes = S_OK;

    try {
        if (p_hMenu == 0) {
            hRes = E_INVALIDARG;
        } else {
            // Make sure this menu hasn't been modified by another instance.
            bool alreadyModified = false;
            {
                CriticalSection::Locker lock(s_ExtToMenusLock);
                ExtToMenuPairV::const_iterator it, end = s_vExtToMenus.end();
                for (it = s_vExtToMenus.begin(); it != end; ++it) {
                    if (it->second == p_hMenu) {
                        // Modified already, bail out.
                        alreadyModified = true;
                        break;
                    }
                }
            }

            // Do not add items if the default action is chosen, if we have no files,
            // if we have no registered plugins or if menu has been modified by another instance.
            if (m_vFiles.empty() || (p_Flags & CMF_DEFAULTONLY) != 0 || m_Registry.PluginCount() == 0 || alreadyModified) {
                hRes = E_FAIL;
            } else {
                UINT cmdId = p_FirstCmdId;
                UINT position = p_Index;

                // Fetch registered plugins from plugin registry.
                PCC::PluginSPV vspPlugins = m_Registry.Plugins();

                // Add all plugins requested to the main menu.
                PCC::CLSIDV vPluginIds;
                if (m_Settings.GetPluginsInMainMenu(vPluginIds)) {
                    if (vPluginIds.size() > 0) {
                        if (vPluginIds.size() != 1 || !::IsEqualCLSID(vPluginIds.front(), PCC::Plugins::LongPathPlugin().Id())) {
                            PCC::CLSIDV::const_iterator it, end = vPluginIds.end();
                            for (it = vPluginIds.begin(); SUCCEEDED(hRes) && it != end; ++it) {
                                hRes = AddPluginToMenu(*it, vspPlugins, p_hMenu, cmdId, position);
                            }
                        } else {
                            // Default plugin is specified, use our own instead.
                            PCC::PluginSP spDefaultPlugin(new PCC::Plugins::DefaultPlugin);
                            hRes = AddPluginToMenu(spDefaultPlugin, p_hMenu, m_Settings.GetUseIconForDefaultPlugin(), cmdId, position);
                        }
                    }
                } else {
                    // No setting specified for items in the main menu. Add our default plugin.
                    PCC::PluginSP spDefaultPlugin(new PCC::Plugins::DefaultPlugin);
                    hRes = AddPluginToMenu(spDefaultPlugin, p_hMenu, m_Settings.GetUseIconForDefaultPlugin(), cmdId, position);
                }

                // Create sub-menu to populate it with the other plugins.
                // If we don't always show the submenu, user needs to ask
                // for extended verbs to get it (by holding Shift).
                if (SUCCEEDED(hRes) && cmdId <= p_LastCmdId &&
                    (m_Settings.GetAlwaysShowSubmenu() || (p_Flags & CMF_EXTENDEDVERBS) != 0)) {

                    HMENU hSubMenu = ::CreatePopupMenu();
                    try {
                        // Fetch list of plugins to omit from the submenu.
                        vPluginIds.clear();
                        m_Settings.GetPluginsNotInSubmenu(vPluginIds);
                        if (!vPluginIds.empty()) {
                            // Scan list of plugins and remove all that are disabled by the user.
                            std::less<CLSID> clsidLess;
                            std::sort(vPluginIds.begin(), vPluginIds.end(), clsidLess);
                            PCC::PluginSPV::iterator pIt, pEnd = vspPlugins.end();
                            for (pIt = vspPlugins.begin(); pIt != pEnd; ++pIt) {
                                if (std::binary_search(vPluginIds.begin(), vPluginIds.end(), (*pIt)->Id(), clsidLess) &&
                                    !(*pIt)->IsSeparator()) {

                                    // Just set it to NULL, we'll handle that case below.
                                    pIt->reset();
                                }
                            }
                        }

                        // Iterate plugins and try to add them to the submenu.
                        UINT subPosition = 0;
                        PCC::PluginSPV::const_iterator it, end = vspPlugins.end();
                        bool prevWasSeparator = true;
                        for (it = vspPlugins.begin(); SUCCEEDED(hRes) && (cmdId <= p_LastCmdId) && (it != end); ++it) {
                            // Make sure it hasn't been set to NULL above.
                            const PCC::PluginSP& spPlugin = *it;
                            if (spPlugin.get() != 0) {
                                // Try to insert this plugin in the menu.
                                if (!spPlugin->IsSeparator()) {
                                    hRes = AddPluginToMenu(spPlugin, hSubMenu, false, cmdId, subPosition);
                                    prevWasSeparator = false;
                                } else {
                                    // This is a proxy to insert a separator.
                                    // Note: there's a chance that we may double up the separators
                                    // if not all plugins are shown in the submenu. Avoid this if possible.
                                    if (!prevWasSeparator) {
                                        if (::InsertMenuW(hSubMenu, subPosition, MF_BYPOSITION | MF_SEPARATOR, 0, 0)) {
                                            ++subPosition;
                                            // No need to increment cmdId.
                                        } else {
                                            hRes = E_FAIL;
                                        }
                                        prevWasSeparator = true;
                                    }
                                }
                            }
                        }

                        // Add item to open the settings app, unless editing
                        // the settings has been locked out by the administrator.
                        if (SUCCEEDED(hRes) && (cmdId <= p_LastCmdId) && !m_Settings.GetEditingDisabled()) {
                            if (!prevWasSeparator) {
                                if (::InsertMenuW(hSubMenu, subPosition, MF_BYPOSITION | MF_SEPARATOR, 0, 0)) {
                                    ++subPosition;
                                } else {
                                    hRes = E_FAIL;
                                }
                            }
                            if (SUCCEEDED(hRes)) {
                                CStringW settingsCaption(MAKEINTRESOURCEW(IDS_PCC_SETTINGS_DESCRIPTION));
                                if (::InsertMenuW(hSubMenu, subPosition, MF_BYPOSITION | MF_STRING, cmdId, settingsCaption)) {
                                    m_SettingsCmdId = cmdId;
                                    ++cmdId;
                                    ++subPosition;
                                } else {
                                    hRes = E_FAIL;
                                }
                            }
                        }
                    } catch (...) {
                        ::DestroyMenu(hSubMenu);
                        throw;
                    }

                    if (SUCCEEDED(hRes) && ::GetMenuItemCount(hSubMenu) > 0) {
                        // Submenu was populated. Add it to the contextual menu.
                        CStringW subMenuCaption(MAKEINTRESOURCEW(IDS_PATH_COPY_MENU_ITEM));
                        PCCDEBUGCODE(subMenuCaption += L" (DEBUG)");
                        MENUITEMINFOW menuItemInfo;
                        menuItemInfo.cbSize = sizeof(MENUITEMINFOW);
                        menuItemInfo.fMask = MIIM_FTYPE | MIIM_ID | MIIM_STRING | MIIM_SUBMENU;
                        menuItemInfo.fType = MFT_STRING;
                        menuItemInfo.wID = cmdId;
                        menuItemInfo.hSubMenu = hSubMenu;
                        menuItemInfo.dwTypeData = const_cast<LPWSTR>((LPCWSTR) subMenuCaption);
                        if (m_Settings.GetUseIconForSubmenu()) {
                            // Add an icon next to the submenu.
                            HBITMAP hIconBitmap = GetPCCIcon();
                            if (hIconBitmap != NULL) {
                                menuItemInfo.fMask |= MIIM_BITMAP;
                                menuItemInfo.hbmpItem = hIconBitmap;
                            }
                        }
                        if (::InsertMenuItemW(p_hMenu, position, TRUE, &menuItemInfo)) {
                            m_SubMenuCmdId = cmdId;
                            ++cmdId;
                            ++position;
                        } else {
                            hRes = E_FAIL;
                        }
                    }

                    if (FAILED(hRes)) {
                        ::DestroyMenu(hSubMenu);
                    }
                }

                if (SUCCEEDED(hRes)) {
                    // Strange return value requirement... see MSDN for details.
                    hRes = MAKE_HRESULT(SEVERITY_SUCCESS, 0, cmdId - p_FirstCmdId + 1);

                    // Mark us as being the instance that modified this menu.
                    CriticalSection::Locker lock(s_ExtToMenusLock);
                    s_vExtToMenus.push_back(ExtToMenuPair(this, p_hMenu));
                }
            }
        }
    } catch (...) {
        hRes = E_UNEXPECTED;
    }

    return hRes;
}

//
// IContextMenu::InvokeCommand
//
// Invoked by the shell when the user selects one of our menu items.
// We do our stuff here.
//
// @param p_pCommandInfo Pointer to struct containing command information.
// @return S_OK if successful, otherwise an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::InvokeCommand(
    CMINVOKECOMMANDINFO* p_pCommandInfo)
{
    HRESULT hRes = S_OK;

    try {
        if ((p_pCommandInfo == 0) || (p_pCommandInfo->cbSize < sizeof(CMINVOKECOMMANDINFO))) {
            hRes = E_INVALIDARG;
        } else {
            // Get offset of invoked command.
            UINT cmdOffset = (UINT) p_pCommandInfo->lpVerb;
            if ((cmdOffset & 0xFFFF0000) != 0) {
                // We do not support verb invokation.
                hRes = E_FAIL;
            } else if (!m_FirstCmdId.HasValue()) {
                // We do not have menu items, so we can't invoke anything.
                hRes = E_INVALIDARG;
            } else {
                // Check which command it is that is invoked.
                UINT cmdId = cmdOffset + m_FirstCmdId;
                CmdIdPluginM::const_iterator itId = m_mPluginsByCmdId.find(cmdId);
                if (itId == m_mPluginsByCmdId.end()) {
                    // This is not a recognized plugin command ID.
                    if (m_SettingsCmdId.HasValue() && m_SettingsCmdId == cmdId) {
                        PCC::SettingsApp().Launch();
                        hRes = S_OK;
                    } else if (m_SubMenuCmdId.HasValue() && m_SubMenuCmdId == cmdId) {
                        hRes = E_INVALIDARG;
                    } else {
                        hRes = E_FAIL;
                    }
                } else {
                    // Fetch plugin reference for this command ID.
                    const PCC::PluginSP& spPlugin = itId->second;

                    // Loop through files and compute filenames using plugin.
                    bool addQuotes = m_Settings.GetAddQuotesAroundPaths();
                    std::wstring newFiles;
                    PCC::FilesV::const_iterator it, end = m_vFiles.end();
                    for (it = m_vFiles.begin(); it != end; ++it) {
                        // Ask plugin to compute filename using its scheme and save it.
                        const std::wstring& oldName = *it;
                        if (!newFiles.empty()) {
                            newFiles += FILE_LINE_SEPARATOR;
                        }
                        if (addQuotes) {
                            newFiles += L"\"";
                        }
                        newFiles += spPlugin->GetPath(oldName);
                        if (addQuotes) {
                            newFiles += L"\"";
                        }
                    }

                    // Now store the copied paths in the clipboard.
                    StClipboard acquireClipboard(p_pCommandInfo->hwnd);
                    if (acquireClipboard.InitResult()) {
                        // Allocate global block to store text.
                        const size_t blockNumElements = newFiles.size() + 1;
                        const size_t blockSize = blockNumElements * sizeof(wchar_t);
                        StGlobalBlock memBlock(GMEM_MOVEABLE, blockSize);
                        if (memBlock.Get() != 0) {
                            // Lock block and copy text in it.
                            {
                                StGlobalLock lockBlock(memBlock.Get());
                                void* pBlock = lockBlock.GetPtr();
                                if (pBlock != 0) {
                                    errno_t copyErr = ::wcscpy_s(static_cast<wchar_t*>(pBlock),
                                                                 blockNumElements,
                                                                 newFiles.c_str());
                                    if (copyErr != 0) {
                                        // Could not copy data in block.
                                        hRes = E_FAIL;
                                    }
                                } else {
                                    // Could not copy data in block.
                                    hRes = E_FAIL;
                                }
                            }

                            if (SUCCEEDED(hRes)) {
                                // Save data in clipboard.
                                HANDLE hSavedData = ::SetClipboardData(CF_UNICODETEXT, memBlock.Get());
                                if (hSavedData != 0) {
                                    // Clipboard now owns the data, avoid freeing it.
                                    memBlock.Release();
                                } else {
                                    // Could not save clipboard data.
                                    hRes = E_FAIL;
                                }
                            }
                        } else {
                            // Could not allocate memory for clipboard data.
                            hRes = E_FAIL;
                        }
                    } else {
                        // Failed to acquire clipboard, can't set data.
                        hRes = E_FAIL;
                    }
                }
            }
        }
    } catch (...) {
        hRes = E_UNEXPECTED;
    }

    return hRes;
}

//
// IContextMenu::GetCommandString
//
// Invoked by the shell to get a text description for our menu item.
//
// @param p_CmdId Offset of ID of command for which to query text, relative
//                to our first command ID.
// @param p_Flags Type of information or action requested; see MSDN for details.
// @param p_pReserved Reserved; unused.
// @param p_pBuffer Pointer to memory buffer where to copy a null-terminated string.
//                  Note: if GCS_UNICODE is found in p_Flags, the string must
//                  be cast to a LPWSTR and a Unicode string must be stored there.
// @param p_BufferSize Max size of p_pBuffer, in characters.
// @return S_OK if successful, otherwise an error code.
//
STDMETHODIMP CPathCopyCopyContextMenuExt::GetCommandString(
    UINT_PTR p_CmdId,
    UINT p_Flags,
    UINT* p_pReserved,
    LPSTR p_pBuffer,
    UINT p_BufferSize)
{
    HRESULT hRes = S_OK;

    try {
        // Check what is requested.
        if ((p_Flags == GCS_VERBA) || (p_Flags == GCS_VERBW)) {
            // We do not support verb invokation.
            hRes = E_NOTIMPL;
        } else if ((p_Flags == GCS_VALIDATEA) || (p_Flags == GCS_VALIDATEW)) {
            // We need to validate command ID.
            hRes = S_FALSE;
            if (m_FirstCmdId.HasValue()) {
                if ((m_mPluginsByCmdId.find(static_cast<UINT>(m_FirstCmdId + p_CmdId)) != m_mPluginsByCmdId.end()) ||
                    (m_SettingsCmdId.HasValue() && (m_FirstCmdId + p_CmdId) == m_SettingsCmdId)) {

                    // Either it's a plugin or a special menu item.
                    hRes = S_OK;
                }
            }
        } else if (p_Flags == GCS_HELPTEXTA) {
            // Call this method to get the Unicode version.
            ArrayAutoPtr<wchar_t> wBuffer(new wchar_t[p_BufferSize]);
            hRes = this->GetCommandString(p_CmdId,
                                          GCS_HELPTEXTW,
                                          p_pReserved,
                                          reinterpret_cast<LPSTR>(wBuffer.Get()),
                                          p_BufferSize);
            if (SUCCEEDED(hRes)) {
                // Convert it to a single-byte string and return it.
                CStringA aBuffer(wBuffer.Get());
                if (::strcpy_s(p_pBuffer, p_BufferSize, aBuffer) == 0) {
                    hRes = S_OK;
                } else {
                    hRes = E_FAIL;
                }
            }
        } else if (p_Flags == GCS_HELPTEXTW) {
            // A Unicode help string is requested.
            if (p_pBuffer != 0) {
                // Try finding the plugin that handles this command ID.
                CmdIdPluginM::const_iterator it = m_mPluginsByCmdId.end();
                if (m_FirstCmdId.HasValue()) {
                    it = m_mPluginsByCmdId.find(static_cast<UINT>(m_FirstCmdId + p_CmdId));
                }
                if (it != m_mPluginsByCmdId.end()) {
                    // Found the plugin, ask for its help text.
                    std::wstring helpText = it->second->HelpText();

                    // Return help text.
                    if (::wcscpy_s((LPWSTR) p_pBuffer, p_BufferSize, helpText.c_str()) != 0) {
                        hRes = E_FAIL;
                    }
                } else {
                    // Check for non-plugin menu items.
                    if (m_FirstCmdId.HasValue() && m_SubMenuCmdId.HasValue() && (m_FirstCmdId + p_CmdId) == m_SubMenuCmdId) {
                        // The sub-menu itself.
                        CStringW subMenuHint(MAKEINTRESOURCEW(IDS_PATH_COPY_HINT));
                        if (::wcscpy_s((LPWSTR) p_pBuffer, p_BufferSize, subMenuHint) != 0) {
                            hRes = E_FAIL;
                        }
                    } else if (m_FirstCmdId.HasValue() && m_SettingsCmdId.HasValue() && (m_FirstCmdId + p_CmdId) == m_SettingsCmdId) {
                        // The item to open the settings application.
                        CStringW settingsHint(MAKEINTRESOURCEW(IDS_PCC_SETTINGS_HINT));
                        if (::wcscpy_s((LPWSTR) p_pBuffer, p_BufferSize, settingsHint) != 0) {
                            hRes = E_FAIL;
                        }
                    } else {
                        hRes = E_INVALIDARG;
                    }
                }
            } else {
                hRes = E_INVALIDARG;
            }
        } else {
            // Unknown, unsupported flag.
            hRes = E_NOTIMPL;
        }
    } catch (...) {
        hRes = E_UNEXPECTED;
    }

    return hRes;
}

//
// Adds a new item to the given menu (either the main contextual menu or
// our submenu can be used) and assigns the given plugin to it. This will
// be called multiple times by the QueryContextMenu method to fill in the
// menu with all the items.
//
// @param p_PluginId ID of plugin to add to the menu.
// @param p_vPlugins Container where to look for the plugin object.
// @param p_hMenu Handle of menu to add the item to.
// @param p_rCmdId Command ID to use for the item. Upon return, this
//                 will be incremented to the next command ID to use.
// @param p_rPosition Position where to insert the item in the menu.
//                    Upon return, this will be incremented to the
//                    next position to use.
// @returns Error code, or S_OK if the item was added to the menu.
//
HRESULT CPathCopyCopyContextMenuExt::AddPluginToMenu(const GUID& p_PluginId,
                                                     const PCC::PluginSPV& p_vspPlugins,
                                                     HMENU const p_hMenu,
                                                     UINT& p_rCmdId,
                                                     UINT& p_rPosition)
{
    // Look for the plugin in the container.
    PCC::PluginSP spPlugin;
    PCC::PluginSPV::const_iterator pIt, pEnd = p_vspPlugins.end();
    for (pIt = p_vspPlugins.begin(); pIt != pEnd; ++pIt) {
        if (::IsEqualGUID((*pIt)->Id(), p_PluginId)) {
            spPlugin = *pIt;
            break;
        }
    }

    // If we have a plugin, continue.
    HRESULT hRes = E_INVALIDARG;
    if (spPlugin.get() != 0) {
        hRes = AddPluginToMenu(spPlugin, p_hMenu, false, p_rCmdId, p_rPosition);
    }

    return hRes;
}

//
// Adds a new item to the given menu (either the main contextual menu or
// our submenu can be used) and assigns the given plugin to it. This will
// be called multiple times by the QueryContextMenu method to fill in the
// menu with all the items.
//
// @param p_spPlugin Plugin to add to the menu.
// @param p_hMenu Handle of menu to add the item to.
// @param p_UsePCCIcon Whether to add the PCC icon next to the plugin in the menu.
// @param p_rCmdId Command ID to use for the item. Upon return, this
//                 will be incremented to the next command ID to use.
// @param p_rPosition Position where to insert the item in the menu.
//                    Upon return, this will be incremented to the
//                    next position to use.
// @returns Error code, or S_OK if the item was added to the menu.
//
HRESULT CPathCopyCopyContextMenuExt::AddPluginToMenu(const PCC::PluginSP& p_spPlugin,
                                                     HMENU const p_hMenu,
                                                     const bool p_UsePCCIcon,
                                                     UINT& p_rCmdId,
                                                     UINT& p_rPosition)
{
    HRESULT hRes = S_OK;

    // Compile info about the menu item using the plugin object.
    const std::wstring description = p_spPlugin->Description();
    MENUITEMINFOW menuItemInfo;
    menuItemInfo.cbSize = sizeof(MENUITEMINFOW);
    menuItemInfo.fMask = MIIM_FTYPE | MIIM_ID | MIIM_STATE | MIIM_STRING;
    menuItemInfo.fType = MFT_STRING;
    menuItemInfo.fState = p_spPlugin->Enabled(m_ParentPath, m_vFiles.front()) ? MFS_ENABLED : MFS_DISABLED;
    menuItemInfo.wID = p_rCmdId;
    menuItemInfo.dwTypeData = const_cast<LPWSTR>(description.c_str());
    if (p_UsePCCIcon) {
        HBITMAP hIconBitmap = GetPCCIcon();
        if (hIconBitmap != NULL) {
            menuItemInfo.fMask |= MIIM_BITMAP;
            menuItemInfo.hbmpItem = hIconBitmap;
        }
    }

    // Insert the item in the menu.
    if (::InsertMenuItemW(p_hMenu, p_rPosition, TRUE, &menuItemInfo)) {
        m_mPluginsByCmdId[p_rCmdId] = p_spPlugin;
        if (!m_FirstCmdId.HasValue()) {
            m_FirstCmdId = p_rCmdId;
        }
        ++p_rCmdId;
        ++p_rPosition;
    } else {
        hRes = E_FAIL;
    }

    return hRes;
}

//
// Returns a handle to the bitmap containing the Path Copy Copy icon
// that can be used for contextual menu items, loading it if necessary.
//
// @return Handle of bitmap containing the icon. Will be NULL if loading failed.
//
HBITMAP CPathCopyCopyContextMenuExt::GetPCCIcon()
{
    // Load on first call.
    if (m_apPCCIcon.get() == 0) {
        m_apPCCIcon.reset(new StImage(CPathCopyCopyModule::HInstance(),
            MAKEINTRESOURCEW(IDB_PCCICON), IMAGE_BITMAP, 16, 16, LR_DEFAULTCOLOR));
    }
    return (m_apPCCIcon->GetLoadResult() == ERROR_SUCCESS) ? m_apPCCIcon->GetBitmap() : NULL;
}

//
// Scans the vector keeping tracks of instances modifying menus
// and removes any reference to this instance from it.
//
void CPathCopyCopyContextMenuExt::RemoveFromExtToMenu()
{
    CriticalSection::Locker lock(s_ExtToMenusLock);
    ExtToMenuPairV::iterator it, end = s_vExtToMenus.end();
    for (it = s_vExtToMenus.begin(); it != end; ++it) {
        if (it->first == this) {
            s_vExtToMenus.erase(it);
            break;
        }
    }
}

//
// Checks for software updates if needed.
//
void CPathCopyCopyContextMenuExt::CheckForUpdates()
{
    // See if we're due for an update check.
    if (m_Settings.NeedsUpdateCheck()) {
        // Mark the last update check as now so that we don't check too often.
        m_Settings.SetLastUpdateCheckNow();

        // Launch the settings app. It will handle the check.
        PCC::SettingsApp().Launch(PCC::SettingsApp::Command_UpdateCheck);
    }
}
