// PathCopyCopyPluginsRegistration.cpp
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PathCopyCopyPluginsRegistration.h"

#include <PathCopyCopySettings.h>
#include <PluginRegistry.h>
#include <PluginSeparator.h>

#include <ShortNamePlugin.h>
#include <LongNamePlugin.h>
#include <ShortPathPlugin.h>
#include <LongPathPlugin.h>
#include <ShortFolderPlugin.h>
#include <LongFolderPlugin.h>
#include <ShortUNCPathPlugin.h>
#include <LongUNCPathPlugin.h>
#include <ShortUNCFolderPlugin.h>
#include <LongUNCFolderPlugin.h>
#include <InternetPathPlugin.h>
#include <UnixPathPlugin.h>
#include <CygwinPathPlugin.h>

#include <algorithm>


namespace PCC
{
    //
    // Registers all plugins currently supported by PCC in the given PluginRegistry.
    //
    // @param p_rRegistry Registry where to add the plugins.
    // @param p_rSettings Object to access PCC settings.
    //
    void PluginsRegistration::Register(PluginRegistry& p_rRegistry,
                                       Settings& p_rSettings)
    {
        p_rRegistry.Clear();
        PluginSP spSeparator(new PluginSeparator);

        // Name plugins
        p_rRegistry.Register(PluginSP(new Plugins::ShortNamePlugin));
        p_rRegistry.Register(PluginSP(new Plugins::LongNamePlugin));

        // Path plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::ShortPathPlugin));
        p_rRegistry.Register(PluginSP(new Plugins::LongPathPlugin));

        // Folder plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::ShortFolderPlugin));
        p_rRegistry.Register(PluginSP(new Plugins::LongFolderPlugin));

        // UNC path plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::ShortUNCPathPlugin));
        p_rRegistry.Register(PluginSP(new Plugins::LongUNCPathPlugin));

        // UNC folder plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::ShortUNCFolderPlugin));
        p_rRegistry.Register(PluginSP(new Plugins::LongUNCFolderPlugin));

        // Internet plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::InternetPathPlugin));

        // *NIX plugins
        p_rRegistry.Register(spSeparator);
        p_rRegistry.Register(PluginSP(new Plugins::UnixPathPlugin));
        p_rRegistry.Register(PluginSP(new Plugins::CygwinPathPlugin));

        // COM plugins
        RegisterCOMPlugins(spSeparator, p_rRegistry, p_rSettings);

        // Pipeline plugins
        RegisterPipelinePlugins(spSeparator, p_rRegistry, p_rSettings);
    }

    //
    // Registers all COM plugins in the given PluginRegistry.
    // Takes care of sorting them by group info (ID and position).
    //
    // @param p_spSeparator Reference to a separator plugin, to avoid recreating it.
    // @param p_rRegistry Registry where to add the plugins.
    // @param p_rSettings Object to access PCC settings.
    //
    void PluginsRegistration::RegisterCOMPlugins(const PluginSP& p_spSeparator,
                                                 PluginRegistry& p_rRegistry,
                                                 Settings& p_rSettings)
    {
        // Get list of COM plugins from settings.
        CLSIDV vPluginCLSIDs;
        p_rSettings.GetCOMPlugins(vPluginCLSIDs);
        if (!vPluginCLSIDs.empty()) {
            // Load list of plugins by creating the COM objects and store group IDs and positions.
            COMPluginSPGroupInfoPairV vCOMPluginsWithGroupInfo;
            CLSIDV::const_iterator clsidIt, clsidEnd = vPluginCLSIDs.end();
            for (clsidIt = vPluginCLSIDs.begin(); clsidIt != clsidEnd; ++clsidIt) {
                try {
                    COMPluginSP spPlugin(new Plugins::COMPlugin(*clsidIt));
                    // Only care about position if the group ID is set.
                    GroupIdPositionPair idPositionPair(spPlugin->GroupId(), spPlugin->GroupPosition());
                    if (idPositionPair.first == 0) {
                        idPositionPair.second = 0;
                    }
                    vCOMPluginsWithGroupInfo.push_back(COMPluginSPGroupInfoPair(spPlugin, idPositionPair));
                } catch (const Plugins::COMPluginError&) {
                    // The plugin is not functional, simply skip it.
                }
            }

            // Check if we have some plugins left. We might fail to load them.
            if (!vCOMPluginsWithGroupInfo.empty()) {
                // Sort plugins by group ID and position.
                if (vCOMPluginsWithGroupInfo.size() > 1) {
                    std::stable_sort(vCOMPluginsWithGroupInfo.begin(), vCOMPluginsWithGroupInfo.end(), COMPluginGroupInfoLess());
                }

                // Make sure we have a separator.
                PluginSP spSeparator = p_spSeparator;
                if (spSeparator.get() == 0) {
                    spSeparator = PluginSP(new PluginSeparator);
                }

                // Register all plugins. Insert a separator between each group.
                ULONG currentGroup = 0xFFFFFFFF;
                COMPluginSPGroupInfoPairV::const_iterator pluginIt, pluginEnd = vCOMPluginsWithGroupInfo.end();
                for (pluginIt = vCOMPluginsWithGroupInfo.begin(); pluginIt != pluginEnd; ++pluginIt) {
                    ULONG pluginGroup = pluginIt->second.first;
                    if (pluginGroup != currentGroup) {
                        p_rRegistry.Register(spSeparator);
                    }
                    p_rRegistry.Register(PluginSP(pluginIt->first));
                    currentGroup = pluginGroup;
                }
            }
        }
    }

    //
    // Registers all pipeline plugins in the plugin registry.
    //
    // @param p_spSeparator Reference to a separator plugin, to avoid recreating it.
    // @param p_rRegistry Registry where to add the plugins.
    // @param p_rSettings Object to access PCC settings.
    //
    void PluginsRegistration::RegisterPipelinePlugins(const PluginSP& p_spSeparator,
                                                      PluginRegistry& p_rRegistry,
                                                      Settings& p_rSettings)
    {
        // Settings object can do this natively.
        PluginSPV vspPipelinePlugins;
        p_rSettings.GetPipelinePlugins(vspPipelinePlugins);
        if (!vspPipelinePlugins.empty()) {
            PluginSP spSeparator = p_spSeparator;
            if (spSeparator.get() == 0) {
                spSeparator = PluginSP(new PluginSeparator());
            }
            p_rRegistry.Register(spSeparator);
            PluginSPV::const_iterator it, end = vspPipelinePlugins.end();
            for (it = vspPipelinePlugins.begin(); it != end; ++it) {
                p_rRegistry.Register(*it);
            }
        }
    }

    //
    // Compares two plugins with associated group info, sorting
    // them by ascending order of group ID and position.
    //
    // @param p_Plugin1 First plugin to compare
    // @param p_Plugin2 Second plugin to compare
    //
    bool PluginsRegistration::COMPluginGroupInfoLess::operator()(
        const COMPluginSPGroupInfoPair& p_Plugin1,
        const COMPluginSPGroupInfoPair& p_Plugin2)
    {
        const GroupIdPositionPair& groupInfo1 = p_Plugin1.second;
        const GroupIdPositionPair& groupInfo2 = p_Plugin2.second;
        return (groupInfo1.first < groupInfo2.first) ||
               (groupInfo1.first == groupInfo2.first && groupInfo1.second < groupInfo2.second);
    }

} // namespace PCC
