// PathCopyCopyRunDll32EntryPoints.cpp
// (c) 2012, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PathCopyCopyRunDll32EntryPoints.h"

#include <PluginRegistry.h>
#include <PathCopyCopySettings.h>
#include <PathCopyCopyPluginsRegistration.h>
#include <StClipboard.h>
#include <StGlobalBlock.h>
#include <StGlobalLock.h>


namespace {
    // Separator between guid and path in the rundll32 command-line.
    const wchar_t   GUID_PATH_SEPARATOR     = L',';
}

//
// GetPathWithPluginW
//
// Function that can be called with rundll32.exe to invoke a specific
// plugin. The command-line must first contain the plugin's GUID followed
// by the path to convert, separated by a comma. Call like this:
//
// rundll32.exe path\to\PCCxx.dll,GetPathWithPlugin {guid},path\to\convert
//
// The resulting path will be stored to the clipboard before the function returns.
//
// p_hWnd         - Window handle to use as parent for our windows.
// p_hDllInstance - Instance handle for our DLL; ignored.
// p_pCmdLine     - Command-line passed to rundll32.
// p_ShowCmd      - How to show any window; ignored.
//
void CALLBACK GetPathWithPluginW(HWND p_hWnd,
                                 HINSTANCE /*p_hDllInstance*/,
                                 LPWSTR p_pCmdLine,
                                 int /*p_ShowCmd*/)
{
    // Assume we won't be able to convert the path.
    std::wstring resultingPath;

    try {
        // Parse command-line and separate the guid from the path.
        std::wstring cmdLine(p_pCmdLine);
        std::wstring::size_type sepPos = cmdLine.find(GUID_PATH_SEPARATOR);
        if (sepPos != std::wstring::npos) {
            // Convert GUID and get the plugin.
            cmdLine[sepPos] = L'\0';
            CLSID pluginId = { 0 };
            if (SUCCEEDED(::CLSIDFromString(&*cmdLine.begin(), &pluginId))) {
                PCC::PluginRegistry registry;
                PCC::Settings settings;
                PCC::PluginsRegistration::Register(registry, settings);
                PCC::PluginSP spPlugin = registry.FindPlugin(pluginId);
                if (spPlugin.get() != 0) {
                    // We got a plugin, now call its GetPath method.
                    resultingPath = spPlugin->GetPath(std::wstring(cmdLine.begin() + sepPos + 1, cmdLine.end()));
                }
            }
        }
    } catch (...) {
        // Assume plugin won't work.
        resultingPath.clear();
    }

    // Copy resulting path to the clipboard.
    StClipboard acquireClipboard(p_hWnd);
    if (acquireClipboard.InitResult()) {
        // Allocate global block to store text.
        const size_t blockNumElements = resultingPath.size() + 1;
        const size_t blockSize = blockNumElements * sizeof(wchar_t);
        StGlobalBlock memBlock(GMEM_MOVEABLE, blockSize);
        if (memBlock.Get() != 0) {
            // Lock block and copy text in it.
            bool copied = false;
            {
                StGlobalLock lockBlock(memBlock.Get());
                void* pBlock = lockBlock.GetPtr();
                if (pBlock != 0) {
                    errno_t copyErr = ::wcscpy_s(static_cast<wchar_t*>(pBlock),
                                                 blockNumElements,
                                                 resultingPath.c_str());
                    copied = copyErr == 0;
                }
            }

            if (copied) {
                // Save data in clipboard.
                HANDLE hSavedData = ::SetClipboardData(CF_UNICODETEXT, memBlock.Get());
                if (hSavedData != 0) {
                    // Clipboard now owns the data, avoid freeing it.
                    memBlock.Release();
                }
            }
        }
    }
}
