// PathCopyCopySettings.cpp
// (c) 2009-2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PathCopyCopySettings.h"

#include <StOleStr.h>
#include <StCoInitialize.h>
#include <StRegKey.h>
#include <PluginUtils.h>
#include <PathCopyCopy_i.h>
#include <PipelinePlugin.h>
#include <PluginSeparator.h>

#include <sstream>
#include <algorithm>
#include <time.h>


namespace
{
    // Name of registry keys storing the PCC settings.
    const wchar_t   PCC_SETTINGS_KEY[]                          = L"Software\\clechasseur\\PathCopyCopy";
    const wchar_t   PCC_PLUGINS_KEY[]                           = L"Software\\clechasseur\\PathCopyCopy\\Plugins";
    const wchar_t   PCC_PIPELINE_PLUGINS_KEY[]                  = L"Software\\clechasseur\\PathCopyCopy\\PipelinePlugins";

    // Values used for PCC settings.
    const wchar_t   SETTING_REVISIONS[]                         = L"Revisions";
    const wchar_t   SETTING_USE_HIDDEN_SHARES[]                 = L"UseHiddenShares";
    const wchar_t   SETTING_ADD_QUOTES[]                        = L"AddQuotes";
    const wchar_t   SETTING_USE_ICON_FOR_DEFAULT_PLUGIN[]       = L"UseIconForDefaultPlugin";
    const wchar_t   SETTING_USE_ICON_FOR_SUBMENU[]              = L"UseIconForSubmenu";
    const wchar_t   SETTING_ALWAYS_SHOW_SUBMENU[]               = L"AlwaysShowSubmenu";
    const wchar_t   SETTING_PLUGINS_NOT_IN_SUBMENU[]            = L"PluginsNotInSubmenu";
    const wchar_t   SETTING_PLUGINS_IN_MAIN_MENU[]              = L"PluginsInMainMenu";
    const wchar_t   SETTING_LAST_PLUGIN_UPDATE_ID[]             = L"LastUpdate";
    const wchar_t   SETTING_PIPELINE_DESCRIPTION[]              = L"Description";
    const wchar_t   SETTING_PIPELINE_DISPLAY_ORDER[]            = L"DisplayOrder";
    const wchar_t   SETTING_LAST_UPDATE_CHECK[]                 = L"LastUpdateCheck";
    const wchar_t   SETTING_UPDATE_INTERVAL[]                   = L"UpdateInterval";
    const wchar_t   SETTING_DISABLE_SOFTWARE_UPDATE[]           = L"DisableSoftwareUpdate";

    // Deprecated PCC setting values. Used for revising.        
    const wchar_t   OLD_SETTING_DISABLED_PLUGINS[]              = L"DisabledPlugins";
    const wchar_t   OLD_SETTING_DEFAULT_PLUGIN[]                = L"DefaultPlugin";

    // Default values for PCC settings.
    const bool      SETTING_USE_HIDDEN_SHARES_DEFAULT           = false;
    const bool      SETTING_ADD_QUOTES_DEFAULT                  = false;
    const bool      SETTING_USE_ICON_FOR_DEFAULT_PLUGIN_DEFAULT = false;
    const bool      SETTING_USE_ICON_FOR_SUBMENU_DEFAULT        = true;
    const bool      SETTING_ALWAYS_SHOW_SUBMENU_DEFAULT         = true;
    const double    SETTING_UPDATE_INTERVAL_DEFAULT             = 604800.0;     // One week, in seconds.
    const bool      SETTING_DISABLE_SOFTWARE_UPDATE_DEFAULT     = false;

    // Constants used to parse data.
    const wchar_t   PLUGINS_SEPARATOR                           = L',';
    const wchar_t   REVISIONS_SEPARATOR                         = L',';

    // Constants used to generate plugin info for the UI.
    const wchar_t   INFO_GROUP_INFO_SEPARATOR                   = L',';
    const wchar_t   INFO_DESCRIPTION_SEPARATOR                  = L'|';

    //
    // Predicate used to sort pipeline plugins according to their sort order.
    // Uses an ID vector of ordered plugin IDs to know whether two plugins
    // should be before one another.
    //
    class PipelinePluginLess
    {
    public:
                        //
                        // Constructor.
                        //
                        // @param p_pvOrderedPluginIds Vector of ordered pipeline plugin IDs.
                        //
        explicit        PipelinePluginLess(const PCC::CLSIDV& p_vOrderedPluginIds)
                            : m_pvOrderedPluginIds(&p_vOrderedPluginIds)
                        {
                        }

                        //
                        // Comparison operator that sorts pipeline plugins according to the
                        // position of their IDs in an ordered vector of plugin IDs.
                        //
                        // @param p_spPlugin1 First pipeline plugin to compare.
                        // @param p_spPlugin2 Second pipeline plugin to compare.
                        // @return true if p_spPlugin1 should be displayed before p_spPlugin2.
                        //
        bool            operator()(const PCC::PluginSP& p_spPlugin1, const PCC::PluginSP& p_spPlugin2) const
                        {
                            // Find each plugin ID in our ordered vector and compare the positions.
                            PCC::CLSIDV::const_iterator it1 = std::find(m_pvOrderedPluginIds->begin(),
                                m_pvOrderedPluginIds->end(), p_spPlugin1->Id());
                            PCC::CLSIDV::const_iterator it2 = std::find(m_pvOrderedPluginIds->begin(),
                                m_pvOrderedPluginIds->end(), p_spPlugin2->Id());
                            return it1 < it2;
                        }

    private:
        const PCC::CLSIDV*
                        m_pvOrderedPluginIds;
    };

} // anonymous namespace

namespace PCC
{
    //
    // Constructor. Opens the user key as an overrideable key (see class for details).
    // Also opens the plugins key in read/write mode if possible, otherwise read-only.
    //
    Settings::Settings()
        : m_UserKey(PCC_SETTINGS_KEY),
          m_PipelinePluginsKey(PCC_PIPELINE_PLUGINS_KEY),
          m_PluginsKey(),
          m_PluginsKeyReadOnly(false)
    {
        m_PluginsKey.Create(HKEY_LOCAL_MACHINE, PCC_PLUGINS_KEY, 0, 0, KEY_READ | KEY_WRITE);
        if (m_PluginsKey.m_hKey == NULL) {
            // Perhaps this is because we're running as a restricted user who doesn't
            // have full access to HKLM. Open as read-only.
            m_PluginsKey.Open(HKEY_LOCAL_MACHINE, PCC_PLUGINS_KEY, KEY_READ);
            m_PluginsKeyReadOnly = true;
        }

        // Apply revisions immediately before user can access the data.
        ApplyRevisions();
    }

    //
    // Checks whether user wants to consider hidden shares when
    // returning paths for the UNC plugins.
    //
    // @return true if hidden shares should be considered, false otherwise.
    //
    bool Settings::GetUseHiddenShares()
    {
        // Check if value exists. If so, read it, otherwise use default value.
        bool useHiddenShares = SETTING_USE_HIDDEN_SHARES_DEFAULT;
        DWORD regUseHiddenShares = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_USE_HIDDEN_SHARES, regUseHiddenShares) == ERROR_SUCCESS) {
            useHiddenShares = regUseHiddenShares != 0;
        }
        return useHiddenShares;
    }

    //
    // Checks whether user wants us to add quotes around the paths
    // saved to the clipboard. This applies to all plugins.
    //
    // @return true if we should add quotes around the paths.
    //
    bool Settings::GetAddQuotesAroundPaths()
    {
        // Check if value exists. If so, read it, otherwise use default value.
        bool addQuotes = SETTING_ADD_QUOTES_DEFAULT;
        DWORD regAddQuotes = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_ADD_QUOTES, regAddQuotes) == ERROR_SUCCESS) {
            addQuotes = regAddQuotes != 0;
        }
        return addQuotes;
    }

    //
    // Checks whether we want to show an icon next to the default plugin
    // in the contextual menu.
    //
    // @return true if we should show an icon next to default plugin.
    //
    bool Settings::GetUseIconForDefaultPlugin()
    {
        // Check if value exists. If so, read it, otherwise use default value.
        bool useIcon = SETTING_USE_ICON_FOR_DEFAULT_PLUGIN_DEFAULT;
        DWORD regUseIcon = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_USE_ICON_FOR_DEFAULT_PLUGIN, regUseIcon) == ERROR_SUCCESS) {
            useIcon = regUseIcon != 0;
        }
        return useIcon;
    }

    //
    // Checks whether we want to show an icon next to the submenu in the contextual menu.
    //
    // @return true if we should show an icon next to submenu.
    //
    bool Settings::GetUseIconForSubmenu()
    {
        // Check if value exists. If so, read it, otherwise use default value.
        bool useIcon = SETTING_USE_ICON_FOR_SUBMENU_DEFAULT;
        DWORD regUseIcon = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_USE_ICON_FOR_SUBMENU, regUseIcon) == ERROR_SUCCESS) {
            useIcon = regUseIcon != 0;
        }
        return useIcon;
    }

    //
    // Checks whether we should always show the submenu with all plugins.
    // If not, we will only show the submenu when the Shift key is pressed.
    //
    // @return true to always show the submenu.
    //
    bool Settings::GetAlwaysShowSubmenu()
    {
        // Check if value exists. If so, read it, otherwise use default value.
        bool alwaysShowSubmenu = SETTING_ALWAYS_SHOW_SUBMENU_DEFAULT;
        DWORD regAlwaysShowSubmenu = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_ALWAYS_SHOW_SUBMENU, regAlwaysShowSubmenu) == ERROR_SUCCESS) {
            alwaysShowSubmenu = regAlwaysShowSubmenu != 0;
        }
        return alwaysShowSubmenu;
    }

    //
    // Returns the IDs of plugins that should not be shown in the PCC submenu.
    //
    // @param p_rvPluginIds Upon return, will contain a list of plugin IDs.
    //                      If the method returns false, this list is untouched.
    // @return true if there were plugins in the settings and they were copied to p_rvPlugins.
    //
    bool Settings::GetPluginsNotInSubmenu(CLSIDV& p_rvPluginIds)
    {
        std::wstring pluginsAsString;
        bool hasValues = PluginUtils::ReadRegistryStringValue(m_UserKey, SETTING_PLUGINS_NOT_IN_SUBMENU, pluginsAsString) == ERROR_SUCCESS;
        if (hasValues) {
            p_rvPluginIds.clear();
            if (!pluginsAsString.empty()) {
                PluginUtils::StringToCLSIDs(pluginsAsString, PLUGINS_SEPARATOR, p_rvPluginIds);
            }
        }
        return hasValues;
    }

    //
    // Returns the IDs of plugins that should be shown in the main contextual menu.
    //
    // @param p_rvPluginIds Upon return, will contain a list of plugin IDs.
    //                      If the method returns false, this list is untouched.
    // @return true if there were plugins in the settings and they were copied to p_rvPlugins.
    //
    bool Settings::GetPluginsInMainMenu(CLSIDV& p_rvPluginIds)
    {
        std::wstring pluginsAsString;
        bool hasValues = PluginUtils::ReadRegistryStringValue(m_UserKey, SETTING_PLUGINS_IN_MAIN_MENU, pluginsAsString) == ERROR_SUCCESS;
        if (hasValues) {
            p_rvPluginIds.clear();
            if (!pluginsAsString.empty()) {
                PluginUtils::StringToCLSIDs(pluginsAsString, PLUGINS_SEPARATOR, p_rvPluginIds);
            }
        }
        return hasValues;
    }

    //
    // Checks if we need to perform a software update check according to the last time we did.
    //
    // @return true if an update check is needed.
    //
    bool Settings::NeedsUpdateCheck()
    {
        bool needsUpdate = false;

        // Check if software update is disabled.
        bool updateDisabled = SETTING_DISABLE_SOFTWARE_UPDATE_DEFAULT;
        DWORD storedUpdateDisabled = 0;
        if (m_UserKey.QueryDWORDValue(SETTING_DISABLE_SOFTWARE_UPDATE, storedUpdateDisabled) == ERROR_SUCCESS) {
            updateDisabled = storedUpdateDisabled != 0;
        }
        if (!updateDisabled) {
            // Get last time it was performed. If we don't have info on that, assume that it's been a hell of a while.
            __time64_t lastUpdateCheck;
            ULONGLONG storedLastUpdate = 0;
            if (m_UserKey.QueryQWORDValue(SETTING_LAST_UPDATE_CHECK, storedLastUpdate) == ERROR_SUCCESS) {
                lastUpdateCheck = static_cast<__time64_t>(storedLastUpdate);
            } else {
                lastUpdateCheck = 0;
            }

            // Get current time.
            __time64_t now;
            ::_time64(&now);

            // Get update interval.
            double updateInterval = SETTING_UPDATE_INTERVAL_DEFAULT;
            DWORD storedUpdateInterval = 0;
            if (m_UserKey.QueryDWORDValue(SETTING_UPDATE_INTERVAL, storedUpdateInterval) == ERROR_SUCCESS) {
                updateInterval = static_cast<double>(storedUpdateInterval);
            }

            // Compare and check if enough time has passed.
            needsUpdate = ::_difftime64(now, lastUpdateCheck) >= updateInterval;
        }

        return needsUpdate;
    }

    //
    // Sets a value in the settings indicating that we performed a software update check now.
    // This will enable the NeedsUpdateCheck method to compare up to this point.
    //
    void Settings::SetLastUpdateCheckNow()
    {
        __time64_t now;
        ::_time64(&now);
        m_UserKey.SetQWORDValue(SETTING_LAST_UPDATE_CHECK, static_cast<ULONGLONG>(now));
    }

    //
    // Checks whether the administrator has disabled editing of the
    // settings by the settings app.
    //
    // @return true if editing is disabled.
    //
    bool Settings::GetEditingDisabled()
    {
        return m_UserKey.Locked();
    }

    //
    // Returns a GUID corresponding to the last time plugins were registered/unregistered.
    // This can serve as an indicator as to when to reload plugins.
    //
    // @param p_rUpdateId ID of last plugin update (a GUID).
    //
    void Settings::LastPluginUpdateId(GUID& p_rUpdateId)
    {
        LONG loadRes = m_PluginsKey.QueryGUIDValue(SETTING_LAST_PLUGIN_UPDATE_ID, p_rUpdateId);
        if (loadRes != ERROR_SUCCESS) {
            // Failed to load for some reason, return default value.
            p_rUpdateId = GUID_NULL;
        }
    }

    //
    // Returns the list of COM plugins registered to be used with Path Copy Copy.
    //
    // @param p_rvCLSIDs Will contain the IDs of COM plugin classes.
    //
    void Settings::GetCOMPlugins(CLSIDV& p_rvCLSIDs)
    {
        // Plugins are stored as value names in the plugins registry key.
        wchar_t valueName[16384];   // See MSDN
        LONG res = ERROR_SUCCESS;
        for (DWORD index = 0; res == ERROR_SUCCESS; ++index) {
            DWORD valueNameSize = 16384;
            res = ::RegEnumValueW(m_PluginsKey.m_hKey, index, valueName, &valueNameSize, 0, 0, 0, 0);
            if (res == ERROR_SUCCESS) {
                // Check if this value name is a CLSID.
                CLSID clsid;
                if (SUCCEEDED(::CLSIDFromString(valueName, &clsid))) {
                    // It is, copy it to returned vector.
                    p_rvCLSIDs.push_back(clsid);
                }
            }
        }
    }

    //
    // Registers a plugin in the Path Copy Copy registry so that it is used in the contextual menu.
    //
    // @param p_CLSID ID of plugin COM class.
    // @returns true if the plugin has been registered, false it if was already registered.
    //
    bool Settings::RegisterCOMPlugin(const CLSID& p_CLSID)
    {
        bool registered = false;

        if (!m_PluginsKeyReadOnly) {
            // Convert CLSID to string.
            StOleStr clsidAsString;
            HRESULT hRes = ::StringFromCLSID(p_CLSID, &clsidAsString);
            if (SUCCEEDED(hRes)) {
                // Check if plugin was already registered.
                DWORD valueType = 0;
                if (::RegQueryValueExW(m_PluginsKey.m_hKey, clsidAsString.Get(), 0, &valueType, 0, 0) != ERROR_SUCCESS) {
                    // Register plugin.
                    m_PluginsKey.SetStringValue(clsidAsString.Get(), GetCOMPluginInfo(p_CLSID).c_str());

                    // Also update last update ID.
                    GUID newUpdateId;
                    ::CoCreateGuid(&newUpdateId);
                    m_PluginsKey.SetGUIDValue(SETTING_LAST_PLUGIN_UPDATE_ID, newUpdateId);

                    // We successfully registered.
                    registered = true;
                }
            } else {
                throw SettingsException(static_cast<LONG>(hRes));
            }
        }

        return registered;
    }

    //
    // Unregisters a plugin, removing it from our registry.
    //
    // @param p_CLSID ID of plugin COM class.
    // @returns true if plugin has been unregistered, false if it wasn't registered.
    //
    bool Settings::UnregisterCOMPlugin(const CLSID& p_CLSID)
    {
        bool unregistered = false;

        if (!m_PluginsKeyReadOnly) {
            // Convert CLSID to string.
            StOleStr clsidAsString;
            HRESULT hRes = ::StringFromCLSID(p_CLSID, &clsidAsString);
            if (SUCCEEDED(hRes)) {
                // Unregister the plugin and check if it worked in one swoop.
                unregistered = m_PluginsKey.DeleteValue(clsidAsString.Get()) == ERROR_SUCCESS;
                if (unregistered) {
                    // Also update last update ID.
                    GUID newUpdateId;
                    ::CoCreateGuid(&newUpdateId);
                    m_PluginsKey.SetGUIDValue(SETTING_LAST_PLUGIN_UPDATE_ID, newUpdateId);
                }
            } else {
                throw SettingsException(static_cast<LONG>(hRes));
            }
        }

        return unregistered;
    }

    //
    // Loads pipeline plugins found in the registry as saved there by the settings app.
    //
    // @param p_rvspPlugins Where to save all pipeline plugins.
    // @param p_AddSeparator If true and there are already plugins in p_rvspPlugins,
    //                       a separator plugin will be added before adding the
    //                       pipeline plugins.
    //
    void Settings::GetPipelinePlugins(PluginSPV& p_rvspPlugins,
                                      const bool p_AddSeparator /*= true*/)
    {
        // Pipeline plugins are stored in the registry in this way:
        //
        // clechasseur\PathCopyCopy
        // \- PipelinePlugins
        //    |   val DisplayOrder = <guid>,<guid>...
        //    \- <guid>
        //    |     val '' = <encoded pipeline>
        //    |     val Description = <description>
        //    \- <guid>
        //          ...
        //
        // We'll need to enumerate the subkeys of the pipeline plugins key to find plugins.
        PluginSPV vspPipelinePlugins;
        UserOverrideableRegKey::SubkeyInfoV vSubkeyInfos;
        m_PipelinePluginsKey.GetSubKeys(vSubkeyInfos);
        UserOverrideableRegKey::SubkeyInfoV::iterator it, end = vSubkeyInfos.end();
        for (it = vSubkeyInfos.begin(); it != end; ++it) {
            // Convert key name into a GUID and make sure it's valid.
            GUID pluginId = { 0 };
            if (::CLSIDFromString(const_cast<wchar_t*>(it->m_KeyName.c_str()), &pluginId) == S_OK) {
                // Get values for the pipeline encoded elements as well as the plugin description.
                std::wstring encodedElements, description;
                StRegKey pluginKey;
                LONG res = ::RegOpenKeyExW(it->m_hParent, it->m_KeyName.c_str(), 0,
                    KEY_QUERY_VALUE, &pluginKey);
                if (res == ERROR_SUCCESS) {
                    res = PluginUtils::ReadRegistryStringValue(pluginKey.Get(),
                        SETTING_PIPELINE_DESCRIPTION, description);
                }
                if (res == ERROR_SUCCESS) {
                    res = PluginUtils::ReadRegistryStringValue(pluginKey.Get(), 0, encodedElements);
                }
                if (res == ERROR_SUCCESS) {
                    // We have all the info we need, create the plugin and add it to the temp list.
                    vspPipelinePlugins.push_back(PluginSP(new PCC::Plugins::PipelinePlugin(
                        pluginId, description, encodedElements)));
                }
            }
        }

        // Get value containing the display order. If found, we'll have to reorder the
        // pipeline plugins according to this value.
        std::wstring displayOrder;
        LONG res = PluginUtils::ReadRegistryStringValue(m_PipelinePluginsKey,
            SETTING_PIPELINE_DISPLAY_ORDER, displayOrder);
        if (res == ERROR_SUCCESS && !displayOrder.empty()) {
            // The value contains a comma-separated list of pipeline plugin IDs.
            CLSIDV vOrderedPluginIds;
            PluginUtils::StringToCLSIDs(displayOrder, PLUGINS_SEPARATOR, vOrderedPluginIds);

            // Sort plugins using our special predicate that will order them properly.
            PipelinePluginLess lessPredicate(vOrderedPluginIds);
            std::sort(vspPipelinePlugins.begin(), vspPipelinePlugins.end(), lessPredicate);
        }

        // If we have pipeline plugins, insert them in the provided return vector.
        // Add a separator as appropriate.
        if (!p_rvspPlugins.empty() && !vspPipelinePlugins.empty() && p_AddSeparator) {
            p_rvspPlugins.push_back(PluginSP(new PluginSeparator()));
        }
        p_rvspPlugins.insert(p_rvspPlugins.end(), vspPipelinePlugins.begin(), vspPipelinePlugins.end());
    }

    //
    // Returns the info to write in the value for a registered
    // COM plugin. This info will be used by the UI when loading plugins.
    //
    // @param p_CLSID ID of the plugin COM class.
    // @return Info string.
    //
    std::wstring Settings::GetCOMPluginInfo(const CLSID& p_CLSID) const
    {
        // Since the UI that allows user to choose plugins doesn't have a way
        // to communicate with us through COM, we'll write some info for the UI
        // in the plugin's value in the registry. The format is:
        //
        // <group_id>,<group_pos>|<description>
        //
        // If the plugin doesn't support group info, that will amount to
        //
        // |<description>

        std::wostringstream wos;

        CComPtr<IPathCopyCopyPlugin> cpPlugin;
        if (SUCCEEDED(cpPlugin.CoCreateInstance(p_CLSID)) && cpPlugin.p != 0) {
            // First check for group info.
            CComQIPtr<IPathCopyCopyPluginGroupInfo> cpPluginGI(cpPlugin);
            if (cpPluginGI.p != 0) {
                ULONG groupId = 0, groupPos = 0;
                if (SUCCEEDED(cpPluginGI->get_GroupId(&groupId)) &&
                    SUCCEEDED(cpPluginGI->get_GroupPosition(&groupPos))) {

                    wos << groupId << INFO_GROUP_INFO_SEPARATOR << groupPos;
                }
            }

            // Now include description.
            CComBSTR bstrDesc;
            wos << INFO_DESCRIPTION_SEPARATOR;
            if (SUCCEEDED(cpPlugin->get_Description(&bstrDesc))) {
                wos << bstrDesc.m_str;
            }
        }

        std::wstring s = wos.str();
        return s;
    }

    //
    // Applies a series of revisions to the config. Revisions are used to update
    // the config when the structure changes. Every time this happens, add a revise
    // step here to ensure the data is updated before settings are accessed.
    //
    void Settings::ApplyRevisions()
    {
        // Read value of current revisions list.
        UIntV vCurrentRevisions;
        std::wstring currentRevisionsAsString;
        PluginUtils::ReadRegistryStringValue(m_UserKey, SETTING_REVISIONS, currentRevisionsAsString);
        PluginUtils::StringToUInts(currentRevisionsAsString, REVISIONS_SEPARATOR, vCurrentRevisions);

        // Sort the list so that we can perform quick finds.
        std::sort(vCurrentRevisions.begin(), vCurrentRevisions.end());

        // Look for and apply missing revisions.
        UIntV vNewRevisions(vCurrentRevisions);
        if (!std::binary_search(vCurrentRevisions.begin(), vCurrentRevisions.end(), 201201311ul)) {
            ApplyNewPluginOrganizationRevision201201311();
            vNewRevisions.push_back(201201311ul);
        }
        if (!std::binary_search(vCurrentRevisions.begin(), vCurrentRevisions.end(), 201204051ul)) {
            ApplyPluginsInMainMenuAdjustment201204051();
            vNewRevisions.push_back(201204051ul);
        }

        // Add any new revisions here.

        // Check if we applied some new revisions. If so, save them.
        if (vNewRevisions.size() != vCurrentRevisions.size()) {
            std::wstring newRevisionsAsString = PluginUtils::UIntsToString(vNewRevisions, REVISIONS_SEPARATOR);
            m_UserKey.SetStringValue(SETTING_REVISIONS, newRevisionsAsString.c_str());
        }
    }

    //
    // Revises the config by changing legacy settings for disabled and default plugins
    // to the new format of "in the main menu"/"not in the contextual menu".
    //
    void Settings::ApplyNewPluginOrganizationRevision201201311()
    {
        // The old format used the following registry values:
        // - DisabledPlugins: comma-separated list of plugins to remove from submenu.
        // - DefaultPlugin: ID of plugin to show in the main menu.
        //
        // The new structure uses the following values instead:
        // - PluginsNotInSubmenu: comma-separated list of plugins to remove from submenu.
        // - PluginsInMainMenu: comma-separated list of plugins to show in the main menu.
        //
        // To revise:
        // - Copy value of "DisabledPlugins" (if there is one) to "PluginsNotInSubmenu"
        // - Copy value of "DefaultPlugin" (if there is one) to "PluginsInMainMenu"

        DWORD dummyValueType = 0;
        if (m_UserKey.QueryValue(SETTING_PLUGINS_NOT_IN_SUBMENU, &dummyValueType, 0, 0) == ERROR_FILE_NOT_FOUND) {
            std::wstring disabledPluginsAsString;
            if (PluginUtils::ReadRegistryStringValue(m_UserKey, OLD_SETTING_DISABLED_PLUGINS, disabledPluginsAsString) == ERROR_SUCCESS) {
                m_UserKey.SetStringValue(SETTING_PLUGINS_NOT_IN_SUBMENU, disabledPluginsAsString.c_str());
                m_UserKey.DeleteValue(OLD_SETTING_DISABLED_PLUGINS);
            }
        }
        if (m_UserKey.QueryValue(SETTING_PLUGINS_IN_MAIN_MENU, &dummyValueType, 0, 0) == ERROR_FILE_NOT_FOUND) {
            std::wstring defaultPluginAsString;
            if (PluginUtils::ReadRegistryStringValue(m_UserKey, OLD_SETTING_DEFAULT_PLUGIN, defaultPluginAsString) == ERROR_SUCCESS) {
                m_UserKey.SetStringValue(SETTING_PLUGINS_IN_MAIN_MENU, defaultPluginAsString.c_str());
                m_UserKey.DeleteValue(OLD_SETTING_DEFAULT_PLUGIN);
            }
        }
    }

    //
    // Revises the config by making it possible to specify an empty list of plugins in main menu
    // and make this different from not having a registry value at all.
    //
    void Settings::ApplyPluginsInMainMenuAdjustment201204051()
    {
        // Formely, an empty list of plugins in main menu was the same as if the value
        // did not exist: we used the default plugin. Now, however, we will only use
        // the default value if no registry value exists; an empty list will be honored.
        // So if we currently have an empty list, delete the value.
        std::wstring pluginsList;
        if (PluginUtils::ReadRegistryStringValue(m_UserKey, SETTING_PLUGINS_IN_MAIN_MENU, pluginsList) && pluginsList.empty()) {
            m_UserKey.DeleteValue(SETTING_PLUGINS_IN_MAIN_MENU);
        }
    }

    //
    // Constructor with error code.
    //
    // @param p_ErrorCode The Windows error code associated with the exception.
    //
    SettingsException::SettingsException(const LONG p_ErrorCode)
        : std::exception(),
          m_ErrorCode(p_ErrorCode)
    {
    }

    //
    // Destructor.
    //
    SettingsException::~SettingsException()
    {
    }

    //
    // Returns the associated error code, returned by a Win32 API call.
    //
    // @returns Associated error code.
    //
    LONG SettingsException::ErrorCode() const
    {
        return m_ErrorCode;
    }

    //
    // Returns a textual description of the exception.
    //
    // @returns Exception description.
    //
    const char* SettingsException::what() const
    {
        return "PCC::SettingsException";
    }

} // namespace PCC
