// PluginPipelineElements.cpp
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PluginPipelineElements.h"

#include <StringUtils.h>
#include <PluginRegistry.h>

#include <algorithm>
#include <assert.h>


namespace PCC
{
    //
    // Constructor.
    //
    QuotesPipelineElement::QuotesPipelineElement()
        : PipelineElement()
    {
    }

    //
    // Modifies the given path by surrounding it with quotes.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void QuotesPipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        p_rPath.insert(p_rPath.begin(), 1, L'\"');
        p_rPath.append(1, L'\"');
    }

    //
    // Constructor.
    //
    BackToForwardSlashesPipelineElement::BackToForwardSlashesPipelineElement()
        : PipelineElement()
    {
    }

    //
    // Modifies the given path by replacing all backslashes by forward slashes.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void BackToForwardSlashesPipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        std::replace(p_rPath.begin(), p_rPath.end(), L'\\', L'/');
    }

    //
    // Constructor.
    //
    ForwardToBackslashesPipelineElement::ForwardToBackslashesPipelineElement()
        : PipelineElement()
    {
    }

    //
    // Modifies the given path by replacing all forward slashes by backslashes.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void ForwardToBackslashesPipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        std::replace(p_rPath.begin(), p_rPath.end(), L'/', L'\\');
    }

    //
    // Constructor.
    //
    // @param p_OldValue Value to replace.
    // @param p_NewValue Replacement value.
    //
    FindReplacePipelineElement::FindReplacePipelineElement(const std::wstring& p_OldValue,
                                                           const std::wstring& p_NewValue)
        : PipelineElement(),
          m_OldValue(p_OldValue),
          m_NewValue(p_NewValue)
    {
    }

    //
    // Modifies the given path by replacing all instances of our old value
    // with our new value.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void FindReplacePipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        StringUtils::ReplaceAll(p_rPath, m_OldValue, m_NewValue);
    }

    //
    // Constructor.
    //
    // @param p_Regex Regular expression to use to find matches.
    // @param p_Format Format of replacement string.
    // @param p_IgnoreCase Whether to ignore case when looking for matches.
    //
    RegexPipelineElement::RegexPipelineElement(const std::wstring& p_Regex,
                                               const std::wstring& p_Format,
                                               const bool p_IgnoreCase)
        : PipelineElement(),
          m_Regex(p_Regex),
          m_Format(p_Format),
          m_IgnoreCase(p_IgnoreCase),
          m_apRegex(),
          m_RegexTested(false)
    {
    }

    //
    // Modifies the given path by looking for matches using our regular
    // expression and replacing them using our format string.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void RegexPipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        // Check if regex is valid.
        InitRegex();
        if (m_apRegex.get() != 0) {
            try {
                // Perform the find-replace and return the modified string.
                p_rPath = std::regex_replace(p_rPath, *m_apRegex, m_Format);
            } catch (const std::regex_error&) {
                // Nothing much we can do, we didn't get this at init time...
                // Probably a problem with the replacement expression.
                // Leave path as-is.
            }
        }
    }

    //
    // Checks if a plugin using this pipeline element should be enabled or not.
    // In our case, we see if our regex is valid.
    //
    // @param p_ParentPath Path of the parent folder for the file to check; unused.
    // @param p_File Path of file to use for the check; unused.
    // @returns false if regular expression is invalid.
    //
    bool RegexPipelineElement::ShouldBeEnabledFor(const std::wstring& /*p_ParentPath*/,
                                                  const std::wstring& /*p_File*/) const
    {
        InitRegex();
        return m_apRegex.get() != 0;
    }

    //
    // Initializes the m_apRegex member using the other members.
    // Call this method before needing to access the regex object.
    //
    // Note: m_apRegex will remain null if the regular expression is invalid.
    //
    void RegexPipelineElement::InitRegex() const
    {
        // Only init once.
        if (!m_RegexTested) {
            // Try creating regex. Keep null if the regex is invalid.
            try {
                std::regex_constants::syntax_option_type reOptions = std::regex_constants::ECMAScript;
                if (m_IgnoreCase) {
                    reOptions |= std::regex_constants::icase;
                }
                m_apRegex.reset(new std::wregex(m_Regex, reOptions));
            } catch (const std::regex_error&) {
                assert(m_apRegex.get() == 0);
            }
            m_RegexTested = true;
        }
    }

    //
    // Constructor.
    //
    // @param p_PluginId ID of plugin to apply.
    //
    ApplyPluginPipelineElement::ApplyPluginPipelineElement(const GUID& p_PluginId)
        : PipelineElement(),
          m_PluginId(p_PluginId)
    {
    }

    //
    // Modifies the given path by fetching a reference to the plugin we need
    // to apply and call its GetPath method on our path.
    //
    // @param p_rPath Path to modify (in-place).
    //
    void ApplyPluginPipelineElement::ModifyPath(std::wstring& p_rPath) const
    {
        // Fetch a reference to the local plugin registry. We can't do anything
        // without it.
        const PluginRegistry* pRegistry = PluginRegistry::GetLocalRegistry();
        if (pRegistry != 0) {
            // Try finding the plugin we need in the registry.
            PluginSP spPlugin = pRegistry->FindPlugin(m_PluginId);
            if (spPlugin.get() != 0) {
                // Success, call the plugin's GetPath method.
                p_rPath = spPlugin->GetPath(p_rPath);
            }
        }
    }

    //
    // Checks if a plugin using this pipeline element should be enabled or not.
    // In our case, we see if our plugin complains.
    //
    // @param p_ParentPath Path of the parent folder for the file to check.
    // @param p_File Path of file to use for the check.
    // @returns false if pipeline says plugin should be disabled for this path.
    //
    bool ApplyPluginPipelineElement::ShouldBeEnabledFor(const std::wstring& p_ParentPath,
                                                        const std::wstring& p_File) const
    {
        // Fetch a reference to the local plugin registry. We can't do anything
        // without it.
        bool enabled = false;
        const PluginRegistry* pRegistry = PluginRegistry::GetLocalRegistry();
        if (pRegistry != 0) {
            // Try finding the plugin we need in the registry.
            PluginSP spPlugin = pRegistry->FindPlugin(m_PluginId);
            if (spPlugin.get() != 0) {
                // Success, call the plugin's Enabled method.
                enabled = spPlugin->Enabled(p_ParentPath, p_File);
            }
        }
        return enabled;
    }

} // namespace PCC
