// PluginRegistry.cpp
// (c) 2008, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PluginRegistry.h"

#include "Plugin.h"
#include <TlsSlot.h>


namespace
{
    // TLS slot used to store the thread-local plugin registry.
    TlsSlotT<const PCC::PluginRegistry*> g_LocalRegistryTls;

} // anonymous namespace

namespace PCC
{
    //
    // Constructor.
    //
    PluginRegistry::PluginRegistry()
        : m_vspPlugins()
    {
    }

    //
    // Registers the given plugin, adding it to the registry's
    // list of supported plugins.
    //
    // @param p_spPlugin Plugin to register. If null is passed,
    //                   nothing happens.
    //
    void PluginRegistry::Register(const PluginSP& p_spPlugin)
    {
        if (p_spPlugin.get() != 0) {
            m_vspPlugins.push_back(p_spPlugin);
        }
    }

    //
    // Clears the entire registry of all its plugins.
    //
    void PluginRegistry::Clear()
    {
        m_vspPlugins.clear();
    }

    //
    // Returns the number of plugins in the registry.
    //
    // @return Number of plugins.
    //
    size_t PluginRegistry::PluginCount() const
    {
        return m_vspPlugins.size();
    }

    //
    // Returns the vector of registered plugins.
    //
    // @return Vector of registered plugins.
    //
    const PluginSPV& PluginRegistry::Plugins() const
    {
        return m_vspPlugins;
    }

    //
    // Looks for a specific plugin in the registry.
    //
    // @param p_PluginId ID of plugin to look for.
    // @return Plugin object if found, otherwise null.
    //
    PluginSP PluginRegistry::FindPlugin(const GUID& p_PluginId) const
    {
        PluginSP spPlugin;
        PluginSPV::const_iterator it, end = m_vspPlugins.end();
        for (it = m_vspPlugins.begin(); it != end; ++it) {
            if (::IsEqualGUID(p_PluginId, (*it)->Id())) {
                spPlugin = *it;
                break;
            }
        }
        return spPlugin;
    }

    //
    // Saves a pointer to a local plugin registry for the calling thread only.
    // Can be useful to allow plugins to access the registry for the calling
    // thread without having to develop custom methods for passing it around.
    //
    // @param p_pRegistry Pointer to local registry. Pass null to reset it.
    //
    void PluginRegistry::SetLocalRegistry(const PluginRegistry* const p_pRegistry)
    {
        if (p_pRegistry == 0 || GetLocalRegistry() == 0) {
            g_LocalRegistryTls.Set(p_pRegistry);
        }
    }

    //
    // Resets the pointer to the local registry to null. Call this from the thread
    // that called SetLocalRegistry, once the registry would no longer be valid.
    //
    // @param p_pRegistry - Registry that should be removed.
    //
    void PluginRegistry::ResetLocalRegistry(const PluginRegistry* const p_pRegistry)
    {
        if (GetLocalRegistry() == p_pRegistry) {
            SetLocalRegistry(0);
        }
    }

    //
    // Returns a pointer to the local registry that has been set for the calling thread.
    //
    // @return Pointer to plugin registry, or null if it has not been set.
    //
    const PluginRegistry* PluginRegistry::GetLocalRegistry()
    {
        return g_LocalRegistryTls.Get();
    }

} // namespace PCC
