// PluginUtils.cpp
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "PluginUtils.h"

#include <ArrayAutoPtr.h>
#include <StringUtils.h>

#include <lm.h>

#include <sstream>


namespace
{
    const DWORD         INITIAL_BUFFER_SIZE     = 1024;     // Initial size of buffer used to fetch UNC name.
    const DWORD         MAX_REG_KEY_NAME_SIZE   = 255;      // Max size of a registry key's name.

    const std::wstring  SHARES_KEY_NAME     = L"SYSTEM\\CurrentControlSet\\Services\\Lanmanserver\\Shares"; // Name of key storing network shares
    const std::wstring  SHARE_PATH_VALUE    = L"Path=";     // Part of a share key's value containing the share path.
    const wchar_t       HIDDEN_SHARE_SUFFIX = L'$';         // Suffix used for hidden shares; we will not consider them unless specified.

    const std::wstring  HIDDEN_DRIVE_SHARES_REGEX   = L"^([A-Za-z])\\:((\\\\|/).*)$";   // Regex used to convert hidden drive shares.
    const std::wstring  HIDDEN_DRIVE_SHARES_FORMAT  = L"$1$$$2";                        // Format string used to convert hidden drive shares.

    const ULONG         REG_BUFFER_CHUNK_SIZE = 512;        // Size of chunks allocated to read the registry.

} // anonymous namespace

namespace PCC
{
    // Static members of PluginUtils
    CriticalSection PluginUtils::s_Lock;
    std::wstring    PluginUtils::s_ComputerName;
    bool            PluginUtils::s_HasComputerName = false;
    std::wregex     PluginUtils::s_HiddenDriveShareRegex(HIDDEN_DRIVE_SHARES_REGEX, std::regex_constants::ECMAScript);

    //
    // Given a path to a file or folder, will return the path to its
    // parent folder. Ex:
    // C:\Foo\Bar.txt => C:\Foo
    //
    // @param p_rPath Complete path on input, parent's path on output.
    // @return true if we did find a parent path and returned it.
    //
    bool PluginUtils::ExtractFolderFromPath(std::wstring& p_rPath)
    {
        // Find the last delimiter in the path and truncate path
        // as appropriate to return only the parent's path.
        std::wstring::size_type lastDelimPos = p_rPath.find_last_of(L"/\\");
        bool found = (lastDelimPos != std::wstring::npos);
        if (found) {
            // We found a delimiter, clear everything after that
            // (and the delimiter as well). Exception: if we're left
            // with only a drive letter, keep the delimiter.
            if (lastDelimPos <= 2) {
                ++lastDelimPos;
            }
            p_rPath.erase(lastDelimPos);
        }
        return found;
    }

    //
    // Checks if the given path is a UNC path in the form
    // \\server\share[\...]
    //
    // @param p_FilePath Path to verify.
    // @returns true if the path is a UNC path.
    //
    bool PluginUtils::IsUNCPath(const std::wstring& p_FilePath)
    {
        return p_FilePath.find(L"\\\\") == 0 && p_FilePath.find(L'\\', 2) > 2;
    }

    //
    // Checks if the given file resides on a mapped network drive.
    // If it does, returns its corresponding network path.
    // Ex: N:\Data\File.txt -> \\server\share\Data\File.txt
    //
    // @param p_rFilePath Local file path. Upon exit, will contain network path.
    // @returns true if the file was on a mapped network drive and we fetched its network path.
    //
    bool PluginUtils::GetMappedDriveFilePath(std::wstring& p_rFilePath)
    {
        // WNetGetUniversalName allows us to get the network path
        // if file is on a mapped drive.
        bool converted = false;
        DWORD bufferSize = INITIAL_BUFFER_SIZE;
        ArrayAutoPtr<char> apBuffer;
        DWORD ret = ERROR_MORE_DATA;
        while (ret == ERROR_MORE_DATA) {
            apBuffer = new char[bufferSize];
            ret = ::WNetGetUniversalNameW(p_rFilePath.c_str(),
                                          UNIVERSAL_NAME_INFO_LEVEL,
                                          apBuffer.Get(),
                                          &bufferSize);
        }
        if (ret == NO_ERROR) {
            // Got UNC path, return it.
            p_rFilePath.assign(reinterpret_cast<UNIVERSAL_NAME_INFO*>(apBuffer.Get())->lpUniversalName);
            converted = true;
        }
        return converted;
    }

    //
    // Checks if the given file resides in a directory in a network share.
    // If it does, returns its corresponding network path.
    // Ex: C:\SharedDir\File.txt -> \\thiscomputer\SharedDir\File.txt
    //
    // @param p_rFilePath Local file path. Upon exit, will contain network path.
    // @param p_UseHiddenShares Whether to consider hidden shares when looking for valid shares.
    // @returns true if the file was in a network share and we fetched its network path.
    //
    bool PluginUtils::GetNetworkShareFilePath(std::wstring& p_rFilePath,
                                              const bool p_UseHiddenShares)
    {
        bool converted = false;

        // Scan registry to see if we can find a share that contains this path.
        // Shares are stored in multi-string registry values in the Lanmanserver service keys.
        CRegKey shareKey;
        if (shareKey.Open(HKEY_LOCAL_MACHINE, SHARES_KEY_NAME.c_str(), KEY_READ) == ERROR_SUCCESS) {
            // Iterate registry values to check each share.
            wchar_t valueName[MAX_REG_KEY_NAME_SIZE + 1];
            std::wstring multiStringValue;
            std::wstring path;
            LONG ret = 0;
            DWORD i = 0;
            do {
                DWORD valueNameSize = MAX_REG_KEY_NAME_SIZE;
                DWORD valueType = 0;
                ret = ::RegEnumValue(shareKey, i, valueName, &valueNameSize, 0, &valueType, 0, 0);
                if (ret == ERROR_SUCCESS && valueType == REG_MULTI_SZ) {
                    // Get the multi-string values.
                    ULONG bufferSize = INITIAL_BUFFER_SIZE;
                    ArrayAutoPtr<wchar_t> buffer;
                    do {
                        buffer = new wchar_t[bufferSize];
                        ret = shareKey.QueryMultiStringValue(valueName, buffer.Get(), &bufferSize);
                    } while (ret == ERROR_MORE_DATA);
                    if (ret == ERROR_SUCCESS) {
                        // Make sure this is not a hidden share (unless we use them).
                        if (valueNameSize != 0 && (p_UseHiddenShares || valueName[valueNameSize - 1] != HIDDEN_SHARE_SUFFIX)) {
                            // Find the "Path=" part of the mult-string. This contains the share path.
                            multiStringValue.assign(buffer.Get(), bufferSize);
                            GetMultiStringLineBeginningWith(multiStringValue, SHARE_PATH_VALUE, path);
                            if (!path.empty()) {
                                // Check if our path is in that share.
                                PCCDEBUGCODE(std::wstring::size_type pathPos = p_rFilePath.find(path));
                                if (p_rFilePath.find(path) == 0) {
                                    // Success: this is a share that contains our path.
                                    // Replace the start of the path with the computer and share name.
                                    std::wstringstream ss;
                                    ss << L"\\\\" << GetLocalComputerName() << L"\\" << valueName;
                                    if (*path.rbegin() == L'\\' || *path.rbegin() == L'/') {
                                        // The substr below will remove the terminator if the share path
                                        // ends with one (for example, for drives' administrative shares).
                                        // We'll have to add an extra one manually.
                                        ss << L'\\';
                                    }
                                    ss << p_rFilePath.substr(path.size());
                                    std::wstring newPath = ss.str();
                                    p_rFilePath = newPath;
                                    converted = true;
                                    break;
                                }
                            }
                        }
                    }
                }

                // Go to next share.
                ++i;
            } while (ret == ERROR_SUCCESS);
        }

        return converted;
    }

    //
    // Checks if the given file resides in a directory on a local drive.
    // If it does, returns its corresponding network path using a hidden drive share.
    // Ex: C:\Dir\File.txt -> \\thiscomputer\C$\Dir\File.txt
    //
    // @param p_rFilePath Local file path. Upon exit, will contain network path.
    // @returns true if the file was on a local drive and we fetched its network path.
    //
    bool PluginUtils::GetHiddenDriveShareFilePath(std::wstring& p_rFilePath)
    {
        // Try to perform the replacement in one shot. If it works, the path is converted.
        bool converted = false;

        try {
            std::wstring replaced = std::regex_replace(p_rFilePath, s_HiddenDriveShareRegex, HIDDEN_DRIVE_SHARES_FORMAT);
            if (!replaced.empty()) {
                std::wstringstream ss;
                ss << L"\\\\" << GetLocalComputerName() << L"\\" << replaced;
                p_rFilePath = ss.str();
                converted = true;
            }
        } catch (const std::regex_error&) {
            // Didn't work.
        }

        return converted;
    }

    //
    // Returns the name of the local computer.
    //
    // @returns Name of local computer.
    //
    const std::wstring& PluginUtils::GetLocalComputerName()
    {
        if (!s_HasComputerName) {
            CriticalSection::Locker lock(s_Lock);
            if (!s_HasComputerName) {
                DWORD length = MAX_COMPUTERNAME_LENGTH + 1;
                wchar_t name[MAX_COMPUTERNAME_LENGTH + 1];
                if (::GetComputerNameW(name, &length) != FALSE) {
                    if (::_wcslwr_s(name, length + 1) == 0) {
                        s_ComputerName = name;
                    }
                }
                s_HasComputerName = true;
            }
        }
        return s_ComputerName;
    }

    //
    // Reads the content of a string registry value and returns it in
    // a std::wstring so that it's easier to manage. Will take care of
    // reallocating buffers as needed.
    //
    // @param p_hKey Handle of key containing the value to read.
    // @param p_pValueName Name of value.
    // @param p_rValue Upon exit, will contain the value.
    // @return Error code, or ERROR_SUCCESS if all goes well.
    //
    long PluginUtils::ReadRegistryStringValue(HKEY const p_hKey,
                                              const wchar_t* const p_pValueName,
                                              std::wstring& p_rValue)
    {
        // Clear the content to assume value doesn't exist.
        p_rValue.clear();

        // Loop until we are able to read the value.
        LONG lRes = ERROR_MORE_DATA;
        ULONG curSize = 0;
        while (lRes == ERROR_MORE_DATA) {
            curSize += REG_BUFFER_CHUNK_SIZE;
            ArrayAutoPtr<wchar_t> aapBuffer(new wchar_t[curSize]);
            DWORD valueType = REG_SZ;
            DWORD curSizeInBytes = curSize * sizeof(wchar_t);
            lRes = ::RegQueryValueExW(p_hKey, p_pValueName, 0, &valueType,
                                      reinterpret_cast<LPBYTE>(aapBuffer.Get()), &curSizeInBytes);
            if (lRes == ERROR_SUCCESS) {
                // Make sure it is a string.
                if (valueType == REG_SZ && (curSizeInBytes % sizeof(wchar_t)) == 0) {
                    // Success, copy resulting string.
                    p_rValue.assign(aapBuffer.Get());
                } else {
                    lRes = ERROR_INVALID_DATATYPE;
                }
            }
        }

        return lRes;
    }

    //
    // Reads the content of a string registry value and returns it in
    // a std::wstring so that it's easier to manage. Will take care of
    // reallocating buffers as needed.
    //
    // @param p_rKey User-overrideable key containing the value to read.
    // @param p_pValueName Name of value.
    // @param p_rValue Upon exit, will contain the value.
    // @return Error code, or ERROR_SUCCESS if all goes well.
    //
    long PluginUtils::ReadRegistryStringValue(UserOverrideableRegKey& p_rKey,
                                              const wchar_t* const p_pValueName,
                                              std::wstring& p_rValue)
    {
        // Clear the content to assume value doesn't exist.
        p_rValue.clear();

        // Loop until we are able to read the value.
        LONG lRes = ERROR_MORE_DATA;
        ULONG curSize = 0;
        while (lRes == ERROR_MORE_DATA) {
            curSize += REG_BUFFER_CHUNK_SIZE;
            ArrayAutoPtr<wchar_t> aapBuffer(new wchar_t[curSize]);
            DWORD valueType = REG_SZ;
            DWORD curSizeInBytes = curSize * sizeof(wchar_t);
            lRes = p_rKey.QueryValue(p_pValueName, &valueType, aapBuffer.Get(), &curSizeInBytes);
            if (lRes == ERROR_SUCCESS) {
                // Make sure it is a string.
                if (valueType == REG_SZ && (curSizeInBytes % sizeof(wchar_t)) == 0) {
                    // Success, copy resulting string.
                    p_rValue.assign(aapBuffer.Get());
                } else {
                    lRes = ERROR_INVALID_DATATYPE;
                }
            }
        }

        return lRes;
    }

    //
    // Given a multi-line string read from a REG_MULTI_SZ registry value,
    // finds a line that begins with a given prefix and returns it.
    //
    // @param p_MultiStringValue The multi-string value.
    // @param p_Prefix Prefix of line to look for.
    // @param p_rLine Returns the entire matching line (excluding prefix), or an empty string if line is not found.
    //
    void PluginUtils::GetMultiStringLineBeginningWith(const std::wstring& p_MultiStringValue,
                                                      const std::wstring& p_Prefix,
                                                      std::wstring&       p_rLine)
    {
        bool found = false;

        // Multi-line values contain embedded NULLs to separate the lines.
        std::wstring::size_type offset = 0;
        do {
            std::wstring::size_type endOffset = p_MultiStringValue.find_first_of(L'\0', offset);
            PCCDEBUGCODE(const wchar_t* pCur = &p_MultiStringValue[offset]);
            if (endOffset == offset) {
                // We're at the end of the value.
                break;
            }

            if (p_MultiStringValue.find(p_Prefix, offset) == offset) {
                // This is the line we're looking for.
                std::wstring::size_type prefixSize = p_Prefix.size();
                p_rLine = p_MultiStringValue.substr(offset + prefixSize, endOffset - offset - prefixSize);
                found = true;
                break;
            } else {
                // Not the line, go to next one.
                offset = endOffset + 1;
            }
        } while (offset < p_MultiStringValue.size());

        if (!found) {
            p_rLine.empty();
        }
    }

    //
    // Converts a string containing a list of CLSIDs to a vector of CLSID structs.
    //
    // @param p_rClsidsAsString String containing the CLSIDs. Upon return, this
    //                          string is unusable.
    // @param p_Separator Character used to separate the CLSIDs in the string.
    // @param p_rvClsids Upon return, will contain the CLSIDs.
    //
    void PluginUtils::StringToCLSIDs(std::wstring& p_rClsidsAsString,
                                     const wchar_t p_Separator,
                                     CLSIDV&       p_rvClsids)
    {
        // Assume there are no CLSIDs.
        p_rvClsids.clear();

        // First split the string.
        WStringV vStringParts;
        StringUtils::Split(p_rClsidsAsString, p_Separator, vStringParts);

        // Scan parts and convert to CLSID.
        CLSID oneClsid = { 0 };
        WStringV::iterator it, end = vStringParts.end();
        for (it = vStringParts.begin(); it != end; ++it) {
            if (SUCCEEDED(::CLSIDFromString(&*it->begin(), &oneClsid))) {
                p_rvClsids.push_back(oneClsid);
            }
        }
    }

    //
    // Converts a string containing a list of unsigned integers to a vector.
    //
    // @param p_rUIntsAsString String containing the integers. Upon return,
    //                         this string is unusable.
    // @param p_Separator Character used to separate the integers in the string.
    // @param p_rvUInts Upon return, will contain the integers.
    //
    void PluginUtils::StringToUInts(std::wstring& p_rUIntsAsString,
                                    const wchar_t p_Separator,
                                    UIntV& p_rvUInts)
    {
        // Assume there are no integers.
        p_rvUInts.clear();

        // First split the string.
        WStringV vStringParts;
        StringUtils::Split(p_rUIntsAsString, p_Separator, vStringParts);

        // Scan parts and convert to integers.
        WStringV::const_iterator it, end = vStringParts.end();
        for (it = vStringParts.begin(); it != end; ++it) {
            std::wistringstream wis(it->c_str());
            UIntV::value_type integer = 0;
            wis >> integer;
            p_rvUInts.push_back(integer);
        }
    }

    //
    // Converts a list of unsigned integers to a string containing them.
    // This is the opposite of StringToUInts.
    //
    // @param p_vUInts List of integers to convert.
    // @param p_Separator Character used to separate the integers in the string.
    // @returns String with merged integers.
    //
    std::wstring PluginUtils::UIntsToString(const UIntV& p_vUInts,
                                            const wchar_t p_Separator)
    {
        // Assume there are no integers.
        std::wstring resultingString;

        // Insert the first integer without separator.
        std::wostringstream wos;
        if (!p_vUInts.empty()) {
            wos << p_vUInts.front();

            // Insert the other elements with separators.
            UIntV::const_iterator it, end = p_vUInts.end();
            for (it = p_vUInts.begin() + 1; it != end; ++it) {
                wos << p_Separator << *it;
            }
        }

        // Copy the resulting string and return it.
        std::wstring ws = wos.str();
        return ws;
    }

} // namespace PCC
