// UserOverrideableRegKey.cpp
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include "stdafx.h"
#include "UserOverrideableRegKey.h"


namespace
{
    // Registry value that is used to lock out users of overriding a key.
    const wchar_t   VALUE_NAME_LOCKED_OUT[]     = L"KeyLock";

} // anonymous namespace

//
// Constructor. Opens the registry keys immediately. The global key in HKLM
// is only open for reading; the user key is open in read/write mode.
//
// @param p_pKeyPath Path of registry key.
// @param p_pUserKeyPath If set to a non-null value, we will use p_pKeyPath
//                       for the global key only and this value for the user key.
//
UserOverrideableRegKey::UserOverrideableRegKey(const wchar_t* const p_pKeyPath,
                                               const wchar_t* const p_pUserKeyPath /*= 0*/)
    : m_GlobalKey(),
      m_UserKey()
{
    // Open global key in HLKM for reading.
    m_GlobalKey.Open(HKEY_LOCAL_MACHINE, p_pKeyPath, KEY_QUERY_VALUE | KEY_ENUMERATE_SUB_KEYS);

    // Open user key in HKCU, creating it if necessary.
    m_UserKey.Create(HKEY_CURRENT_USER, (p_pUserKeyPath != 0) ? p_pUserKeyPath : p_pKeyPath);
}

//
// Checks whether the user registry key is currently locked in the global key.
// This allows administrators to set options and disallow changing them.
//
// @return true if the user key is locked and cannot be edited.
//
bool UserOverrideableRegKey::Locked()
{
    bool locked = false;
    DWORD regLocked = 0;
    if (m_GlobalKey.m_hKey != NULL &&
        m_GlobalKey.QueryDWORDValue(VALUE_NAME_LOCKED_OUT, regLocked) == ERROR_SUCCESS) {

        locked = regLocked != 0;
    }
    return locked;
}

//
// Tries to load a DWORD value from the registry key.
//
// @param p_pValueName Name of value to load.
// @param p_rValue Where to store value.
// @return Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::QueryDWORDValue(const wchar_t* const p_pValueName,
                                             DWORD& p_rValue)
{
    long res = ERROR_FILE_NOT_FOUND;
    if (!Locked()) {
        res = m_UserKey.QueryDWORDValue(p_pValueName, p_rValue);
    } else {
        res = ERROR_ACCESS_DENIED;
    }
    if (res != ERROR_SUCCESS && m_GlobalKey != NULL) {
        res = m_GlobalKey.QueryDWORDValue(p_pValueName, p_rValue);
    }
    return res;
}

//
// Tries to load a QWORD value from the registry key.
//
// @param p_pValueName Name of value to load.
// @param p_rValue Where to store value.
// @return Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::QueryQWORDValue(const wchar_t* const p_pValueName,
                                             ULONGLONG& p_rValue)
{
    long res = ERROR_FILE_NOT_FOUND;
    if (!Locked()) {
        res = m_UserKey.QueryQWORDValue(p_pValueName, p_rValue);
    } else {
        res = ERROR_ACCESS_DENIED;
    }
    if (res != ERROR_SUCCESS && m_GlobalKey != NULL) {
        res = m_GlobalKey.QueryQWORDValue(p_pValueName, p_rValue);
    }
    return res;
}

//
// Tries to load a GUID value from the registry key (stored as a string).
//
// @param p_pValueName Name of value to load.
// @param p_rValue Where to store value.
// @return Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::QueryGUIDValue(const wchar_t* const p_pValueName,
                                            GUID& p_rValue)
{
    long res = ERROR_FILE_NOT_FOUND;
    if (!Locked()) {
        res = m_UserKey.QueryGUIDValue(p_pValueName, p_rValue);
    } else {
        res = ERROR_ACCESS_DENIED;
    }
    if (res != ERROR_SUCCESS && m_GlobalKey != NULL) {
        res = m_GlobalKey.QueryGUIDValue(p_pValueName, p_rValue);
    }
    return res;
}

//
// Tries to load a value from the registry key.
//
// @param p_pValueName Name of value to load.
// @param p_pValueType If set, will receive the type of value.
// @param p_pValue Pointer to buffer where to store value. Can be null.
// @param p_pValueSize Pointer to variable containing the size of p_pValue.
//                     Upon exit, will contain the actual size of the
//                     value copied to p_pValue. Can be null only if p_pValue is too.
// @return Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::QueryValue(const wchar_t* const p_pValueName,
                                        DWORD* const p_pValueType,
                                        void* const p_pValue,
                                        DWORD* const p_pValueSize)
{
    long res = ERROR_FILE_NOT_FOUND;
    if (!Locked()) {
        res = m_UserKey.QueryValue(p_pValueName, p_pValueType, p_pValue, p_pValueSize);
    } else {
        res = ERROR_ACCESS_DENIED;
    }
    if (res != ERROR_SUCCESS && res != ERROR_MORE_DATA && m_GlobalKey != NULL) {
        res = m_GlobalKey.QueryValue(p_pValueName, p_pValueType, p_pValue, p_pValueSize);
    }
    return res;
}

//
// Scans both the user and global keys to find subkeys and returns
// a list of all such subkeys.
//
// @param p_rvSubkeys Where to store information about the subkeys.
//
void UserOverrideableRegKey::GetSubKeys(SubkeyInfoV& p_rvSubkeys)
{
    if (m_GlobalKey != NULL) {
        GetSubKeys(m_GlobalKey, p_rvSubkeys);
    }
    if (!Locked()) {
        GetSubKeys(m_UserKey, p_rvSubkeys);
    }
}

//
// Tries to save a QWORD value in the registry key.
// This will always write the value to the user key.
//
// @param p_pValueName Name of value to save.
// @param p_pValue Value to save.
// @returns Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::SetQWORDValue(const wchar_t* const p_pValueName,
                                           const ULONGLONG p_Value)
{
    long res = ERROR_ACCESS_DENIED;
    if (!Locked()) {
        res = m_UserKey.SetQWORDValue(p_pValueName, p_Value);
    }
    return res;
}

//
// Tries to save a GUID value in the registry key (as a string).
// This will always write the value to the user key.
//
// @param p_pValueName Name of value to save.
// @param p_pValue Value to save.
// @returns Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::SetGUIDValue(const wchar_t* const p_pValueName,
                                          const GUID& p_Value)
{
    long res = ERROR_ACCESS_DENIED;
    if (!Locked()) {
        res = m_UserKey.SetGUIDValue(p_pValueName, p_Value);
    }
    return res;
}

//
// Tries to save a string value in the registry key. This will always
// write the value to the user key.
//
// @param p_pValueName Name of value to save.
// @param p_pValue Value to save.
// @returns Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::SetStringValue(const wchar_t* const p_pValueName,
                                            const wchar_t* const p_pValue)
{
    long res = ERROR_ACCESS_DENIED;
    if (!Locked()) {
        res = m_UserKey.SetStringValue(p_pValueName, p_pValue);
    }
    return res;
}

//
// Tries to delete a value from the user registry key.
// NOTE: if a similar value is found in the global registry key,
// that value is unchanged and will thus come out on the next Query!
//
// @param p_pValueName Name of value to delete.
// @returns Result code (ERROR_SUCCESS if it worked).
//
long UserOverrideableRegKey::DeleteValue(const wchar_t* const p_pValueName)
{
    long res = ERROR_ACCESS_DENIED;
    if (!Locked()) {
        res = m_UserKey.DeleteValue(p_pValueName);
    }
    return res;
}

//
// Scans the subkeys of the given registry key and returns info about them.
//
// @param p_rKey Key to enumerate the subkeys of.
// @param p_rvSubkeys Where to store information about the subkeys.
//
void UserOverrideableRegKey::GetSubKeys(CRegKey& p_rKey,
                                        SubkeyInfoV& p_rvSubkeys)
{
    // Scan subkeys, starting with the first one.
    wchar_t subkeyName[256];        // See MSDN's RegEnumKeyEx.
    DWORD subkeyNameSize = 256;
    DWORD i = 0;
    LONG res = p_rKey.EnumKey(i, subkeyName, &subkeyNameSize);
    while (res == ERROR_SUCCESS) {
        // Add info about this subkey to the vector.
        p_rvSubkeys.push_back(SubkeyInfo(p_rKey.m_hKey, subkeyName));

        // Go to next subkey.
        subkeyNameSize = 256;
        res = p_rKey.EnumKey(++i, subkeyName, &subkeyNameSize);
    }
}

//
// Default constructor.
//
UserOverrideableRegKey::SubkeyInfo::SubkeyInfo()
    : m_hParent(0),
      m_KeyName()
{
}

//
// Parametrized constructor.
//
// @param p_hParent Handle of parent key.
// @param p_pKeyName Name of subkey.
//
UserOverrideableRegKey::SubkeyInfo::SubkeyInfo(HKEY const p_hParent,
                                               const wchar_t* const p_pKeyName)
    : m_hParent(p_hParent),
      m_KeyName(p_pKeyName)
{
}
