﻿// COMPlugins.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using Microsoft.Win32;
using PathCopyCopy.Settings;

namespace PCC
{
    //************************************************************************
    /// <summary>
    /// Helper to get a list of all COM plugins in Path Copy Copy.
    /// </summary>
    //************************************************************************
    public static class COMPlugins
    {
        /// Regex used to parse the content of a registry value for a COM plugin.
        private static readonly Regex COM_PLUGIN_CONTENT_REGEX = new Regex(
            @"(?:(?<groupId>[0-9]+),(?<groupPosition>[0-9]+))?\|(?<description>.*)", RegexOptions.Compiled);

        /// Name of the groups used in the content regex.
        private const string GROUP_ID_GROUP_NAME = "groupId";
        private const string GROUP_POSITION_GROUP_NAME = "groupPosition";
        private const string DESCRIPTION_GROUP_NAME = "description";

        /// Path of the Plugins key in the registry.
        private const string PCC_PLUGINS_KEY = @"Software\clechasseur\PathCopyCopy\Plugins";

        /// Path of the Plugins key in the registry when running under WOW64.
        private const string PCC_PLUGINS_KEY_WOW64 = @"Software\Wow6432Node\clechasseur\PathCopyCopy\Plugins";

        /// Name of the value storing a marker to know when the plugins were last updated. We skip that.
        private const string PLUGINS_LAST_UPDATE_VALUE_NAME = "LastUpdate";

        /// Bean containing info about a COM plugin.
        private struct COMPluginInfo
        {
            public Plugin plugin;
            public int groupId;
            public int groupPosition;
        }

        //************************************************************************
        /// <summary>
        /// Returns a list of all registered COM plugins in Path Copy Copy.
        /// </summary>
        /// <returns>List of COM plugins.</returns>
        //************************************************************************
        public static List<Plugin> GetCOMPlugins()
        {
            List<COMPluginInfo> pluginsInfo = new List<COMPluginInfo>();
            SeparatorPlugin separator = new SeparatorPlugin();

            // Get path to the registry key that contains the plugins.
            // If we're called by a 32-bit extension running under WOW64 we'll need to adjust.
            string regKeyPath;
            if (!PCCEnvironment.Is64Bit && PCCEnvironment.IsWow64) {
                regKeyPath = PCC_PLUGINS_KEY_WOW64;
            } else {
                regKeyPath = PCC_PLUGINS_KEY;
            }

            // Open the registry key.
            using (RegistryKey pluginsKey = Registry.LocalMachine.OpenSubKey(regKeyPath, false)) {
                if (pluginsKey != null) {
                    // Get names of all values. Each name is a plugin ID except for the marker.
                    string[] ids = pluginsKey.GetValueNames();
                    foreach (string id in ids) {
                        // Try converting this ID to a Guid. Note that there are other values
                        // in that key so if it fails, simply skip it.
                        Guid? idAsGuid = null;
                        try {
                            if (id != PLUGINS_LAST_UPDATE_VALUE_NAME) {
                                idAsGuid = new Guid(id);
                            }
                        } catch (FormatException) {
                            idAsGuid = null;
                        } catch (OverflowException) {
                            idAsGuid = null;
                        }
                        if (idAsGuid != null) {
                            // Get value content which we'll use to interpret the plugin.
                            string value = (string) pluginsKey.GetValue(id);

                            // Parse the value using our regex.
                            Match match = COM_PLUGIN_CONTENT_REGEX.Match(value);

                            // Check if we have group info.
                            Group groupIdGroup = match.Groups[GROUP_ID_GROUP_NAME];
                            Group groupPosGroup = match.Groups[GROUP_POSITION_GROUP_NAME];

                            // Get description which is always present.
                            string description = match.Groups[DESCRIPTION_GROUP_NAME].Value;

                            // Construct bean for this plugin and add it to the list.
                            COMPluginInfo info;
                            info.plugin = new Plugin(idAsGuid.Value, description);
                            if (groupIdGroup.Success) {
                                info.groupId = int.Parse(groupIdGroup.Value);
                            } else {
                                info.groupId = 0;
                            }
                            if (groupPosGroup.Success) {
                                info.groupPosition = int.Parse(groupPosGroup.Value);
                            } else {
                                info.groupPosition = 0;
                            }
                            pluginsInfo.Add(info);
                        }
                    }
                }
            }

            // Sort list of plugins first by group ID then by group position.
            pluginsInfo.Sort(delegate (COMPluginInfo info1, COMPluginInfo info2) {
                int cmp = info1.groupId - info2.groupId;
                if (cmp == 0) {
                    cmp = info1.groupPosition - info2.groupPosition;
                }
                return cmp;
            });

            // Copy all plugins to a list of plugins only. Insert a separator between groups.
            List<Plugin> plugins = new List<Plugin>();
            if (pluginsInfo.Count != 0) {
                int currentGroup = pluginsInfo[0].groupId;
                foreach (var info in pluginsInfo) {
                    if (currentGroup != info.groupId) {
                        plugins.Add(separator);
                    }
                    plugins.Add(info.plugin);
                    currentGroup = info.groupId;
                }
            }

            return plugins;
        }
    }
}
