﻿// MainForm.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using PathCopyCopy.Settings;
using PathCopyCopy.Settings.Properties;
using PCC;
using System.IO;
using System.Reflection;

namespace PathCopyCopySettings
{
    //************************************************************************
    /// <summary>
    /// Main form of the Path Copy Copy settings application.
    /// </summary>
    //************************************************************************
    public partial class MainForm : Form
    {
        /// Name of our LICENSE file.
        private const string LICENSE_TXT_FILE = "LICENSE.TXT";

        // URI of the Donations page.
        private const string DONATIONS_PAGE_URI = @"https://www.paypal.com/cgi-bin/webscr?cmd=_donations&business=LM5B9WNTH4KN4&lc=CA&item_name=Charles%20Lechasseur&item_number=PathCopyCopy&currency_code=CAD&bn=PP%2dDonationsBF%3abtn_donateCC_LG%2egif%3aNonHosted";

        /// The standard DPI values in Windows.
        private const int STANDARD_DPI = 96;

        /// Delegate that will be used to handle events related to changes to pipeline plugins.
        private delegate void PipelinePluginEventHandler(object sender, PipelinePluginEventArgs e);

        /// Delegate that will be used to handle events when pipeline plugins are added.
        private delegate void PipelinePluginAddEventHandler(object sender, PipelinePluginAddEventArgs e);

        /// Delegate that will be used to handle events when two pipeline plugins switch places.
        private delegate void PipelinePluginSwitchEventHandler(object sender, PipelinePluginSwitchEventArgs e);

        /// Delegate that will be used to handle events when pipeline plugins are imported.
        private delegate void PipelinePluginsImportEventHandler(object sender, PipelinePluginsImportedEventArgs e);

        /// Event triggered when a pipeline plugin is added.
        private event PipelinePluginAddEventHandler PipelinePluginAdded;

        /// Event triggered when a pipeline plugin changes.
        private event PipelinePluginEventHandler PipelinePluginChanged;

        /// Event triggered when two pipeline plugins switch places.
        private event PipelinePluginSwitchEventHandler PipelinePluginMoved;

        /// Event triggered when a pipeline plugin is deleted.
        private event PipelinePluginEventHandler PipelinePluginDeleted;

        /// Event triggered when pipeline plugins have been imported.
        private event PipelinePluginsImportEventHandler PipelinePluginsImported;

        /// Object used to access user settings.
        private UserSettings settings;

        /// List of immutable plugins that cannot move or be deleted.
        private List<Plugin> immutablePlugins;

        //************************************************************************
        /// <summary>
        /// Constructor
        /// </summary>
        //************************************************************************
        public MainForm()
        {
            InitializeComponent();

            // Disable the grid view's auto-creation of columns since we've
            // entered them by hand.
            PluginsDataGrid.AutoGenerateColumns = false;

            // The checkbox columns are bound to properties that return strings.
            // We'll set the true and false values accordingly.
            for (int i = 1; i <= 2; ++i) {
                DataGridViewCheckBoxColumn column = (DataGridViewCheckBoxColumn) PluginsDataGrid.Columns[i];
                column.TrueValue = "1";
                column.FalseValue = String.Empty;
            }

            // Our DataGridView's checkboxes don't show up properly in larger font sizes
            // unless the AutoSizeRowsMode is set to AllCells (or AllCellsExceptHeaders).
            // However I find that it looks a bit too large to set it to that height...
            // What we'll do is try to detect non-standard DPI and if found, we'll set
            // the property (it'll produce tall rows but at least the checkboxes will show up.)
            using (Graphics gridViewGraphics = PluginsDataGrid.CreateGraphics()) {
                if (((int) gridViewGraphics.DpiX) != STANDARD_DPI || ((int) gridViewGraphics.DpiY) != STANDARD_DPI) {
                    PluginsDataGrid.AutoSizeRowsMode = DataGridViewAutoSizeRowsMode.AllCellsExceptHeaders;
                }
            }

            // Init about "box" controls.
            ProductAndVersionLbl.Text = String.Format(ProductAndVersionLbl.Text,
                Assembly.GetExecutingAssembly().GetName().Version.ToString());
            while (ProductAndVersionLbl.Text.EndsWith(".0")) {
                ProductAndVersionLbl.Text = ProductAndVersionLbl.Text.Substring(0, ProductAndVersionLbl.Text.Length - 2);
            }
            DonationLinkLbl.Links[0].LinkData = DONATIONS_PAGE_URI;

            // Hook to events triggered when the pipeline plugin changes.
            PipelinePluginAdded += MainForm_PipelinePluginAdded;
            PipelinePluginChanged += MainForm_PipelinePluginChanged;
            PipelinePluginMoved += MainForm_PipelinePluginMoved;
            PipelinePluginDeleted += MainForm_PipelinePluginDeleted;
            PipelinePluginsImported += MainForm_PipelinePluginsImported;
        }

        //************************************************************************
        /// <summary>
        /// Loads the options and populates the form.
        /// </summary>
        //************************************************************************
        private void LoadUserSettings()
        {
            Debug.Assert(settings != null);

            // Get list of immutable plugins: default & COM.
            immutablePlugins = DefaultPlugins.GetDefaultPlugins();
            List<Plugin> comPlugins = COMPlugins.GetCOMPlugins();
            if (comPlugins.Count != 0) {
                immutablePlugins.Add(new SeparatorPlugin());
                immutablePlugins.AddRange(comPlugins);
            }

            // Get list of pipeline plugin infos.
            List<PipelinePluginInfo> pipelinePluginInfos = settings.PipelinePlugins;

            // Build list of plugins for lists that need them all.
            List<Plugin> plugins = new List<Plugin>(immutablePlugins);
            if (pipelinePluginInfos.Count != 0) {
                plugins.Add(new SeparatorPlugin());
                plugins.AddRange(PipelinePluginInfo.ToPlugins(pipelinePluginInfos));
            }

            // Get lists of plugin states from settings.
            List<Guid> pluginsNotInSubmenu = settings.PluginsNotInSubmenu;
            List<Guid> pluginsInMainMenu = settings.PluginsInMainMenu;
            if (pluginsInMainMenu == null) {
                // If no main menu plugins are specified, use the default.
                pluginsInMainMenu = new List<Guid>();
                pluginsInMainMenu.Add(new Guid(Resources.LONG_PATH_PLUGIN_ID));
            }

            // Prepare plugin display info for the data grid.
            BindingList<PluginDisplayInfo> pluginDisplayInfos = new BindingList<PluginDisplayInfo>();
            foreach (Plugin plugin in plugins) {
                PluginDisplayInfo displayInfo = new PluginDisplayInfo(plugin);
                displayInfo.ShowInMainMenu = pluginsInMainMenu.IndexOf(plugin.ID) != -1;
                displayInfo.ShowInSubmenu = pluginsNotInSubmenu.IndexOf(plugin.ID) == -1 && !plugin.IsSeparator;
                pluginDisplayInfos.Add(displayInfo);
            }

            // Set data grid's data source so that it updates its content.
            PluginsDataGrid.DataSource = pluginDisplayInfos;

            // Disable all rows that correspond to separators and remove their checkboxes.
            foreach (object rawRow in PluginsDataGrid.Rows) {
                DataGridViewRow row = (DataGridViewRow) rawRow;
                if (((PluginDisplayInfo) row.DataBoundItem).Plugin.IsSeparator) {
                    for (int i = 1; i <= 2; ++i) {
                        row.Cells[i] = new DataGridViewTextBoxCell();
                    }
                    row.ReadOnly = true;
                }
            }

            // Add plugin infos to the pipeline plugins list.
            PipelinePluginsLst.BeginUpdate();
            try {
                PipelinePluginsLst.Items.AddRange(pipelinePluginInfos.ToArray());
            } finally {
                PipelinePluginsLst.EndUpdate();
            }

            // Set options checkboxes.
            AddQuotesChk.Checked = settings.AddQuotes;
            HiddenSharesChk.Checked = settings.UseHiddenShares;
            AlwaysShowSubmenuChk.Checked = settings.AlwaysShowSubmenu;
            UseIconForSubmenuChk.Checked = settings.UseIconForSubmenu;
            EnableSoftwareUpdateChk.Checked = !settings.DisableSoftwareUpdate;

            // All those changes probably enabled the "Apply" button so disable it again here.
            ApplyBtn.Enabled = false;

            Debug.Assert(immutablePlugins != null);
            Debug.Assert(pipelinePluginInfos != null);
        }

        //************************************************************************
        /// <summary>
        /// Saves the current options in the user settings.
        /// </summary>
        //************************************************************************
        private void SaveUserSettings()
        {
            Debug.Assert(settings != null);

            // Scan plugin states and build lists to save in config.
            List<Guid> pluginsNotInSubmenu = new List<Guid>();
            List<Guid> pluginsInMainMenu = new List<Guid>();
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            foreach (var stateInfo in pluginDisplayInfos) {
                if (!stateInfo.Plugin.IsSeparator) {
                    if (!stateInfo.ShowInSubmenu) {
                        pluginsNotInSubmenu.Add(stateInfo.Plugin.ID);
                    }
                    if (stateInfo.ShowInMainMenu) {
                        pluginsInMainMenu.Add(stateInfo.Plugin.ID);
                    }
                }
            }

            // If the only plugin left to show in the main menu is our default one,
            // clear the value instead.
            if (pluginsInMainMenu.Count == 1 && pluginsInMainMenu[0] == new Guid(Resources.LONG_PATH_PLUGIN_ID)) {
                pluginsInMainMenu = null;
            }

            // Save plugin state lists in config.
            settings.PluginsNotInSubmenu = pluginsNotInSubmenu;
            settings.PluginsInMainMenu = pluginsInMainMenu;

            // Save the pipeline plugins.
            List<PipelinePluginInfo> pipelinePluginInfos = new List<PipelinePluginInfo>();
            foreach (object info in PipelinePluginsLst.Items) {
                pipelinePluginInfos.Add((PipelinePluginInfo) info);
            }
            settings.PipelinePlugins = pipelinePluginInfos;

            // Save advanced options.
            if (AddQuotesChk.Checked != settings.AddQuotes) {
                settings.AddQuotes = AddQuotesChk.Checked;
            }
            if (HiddenSharesChk.Checked != settings.UseHiddenShares) {
                settings.UseHiddenShares = HiddenSharesChk.Checked;
            }
            if (AlwaysShowSubmenuChk.Checked != settings.AlwaysShowSubmenu) {
                settings.AlwaysShowSubmenu = AlwaysShowSubmenuChk.Checked;
            }
            if (UseIconForSubmenuChk.Checked != settings.UseIconForSubmenu) {
                settings.UseIconForSubmenu = UseIconForSubmenuChk.Checked;
            }
            if (EnableSoftwareUpdateChk.Checked != (!settings.DisableSoftwareUpdate)) {
                settings.DisableSoftwareUpdate = !EnableSoftwareUpdateChk.Checked;
            }

            // Now that everything is saved, disable the "Apply button".
            ApplyBtn.Enabled = false;
        }

        //************************************************************************
        /// <summary>
        /// Adds a separator at the end of the list of plugins to prepare to add
        /// new pipeline plugins after it.
        /// </summary>
        //************************************************************************
        private void AddPipelinePluginSeparator()
        {
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            pluginDisplayInfos.Add(new PluginDisplayInfo(new SeparatorPlugin()));
            // Also fix the row to avoid showing checkboxes.
            DataGridViewRow row = PluginsDataGrid.Rows[pluginDisplayInfos.Count - 1];
            for (int i = 1; i <= 2; ++i) {
                row.Cells[i] = new DataGridViewTextBoxCell();
            }
        }

        //************************************************************************
        /// <summary>
        /// Removes the separator that is at the end of the list of plugins since
        /// there are no more pipeline plugins after it.
        /// </summary>
        //************************************************************************
        private void RemovePipelinePluginSeparator()
        {
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            Debug.Assert(pluginDisplayInfos[pluginDisplayInfos.Count - 1].Plugin.IsSeparator);
            pluginDisplayInfos.RemoveAt(pluginDisplayInfos.Count - 1);
        }

        //************************************************************************
        /// <summary>
        /// Event raised when a new pipeline plugin is added. We need to update
        /// the other listboxes of the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_PipelinePluginAdded(object sender, PipelinePluginAddEventArgs e)
        {
            // If this is the first pipeline plugin, add a separator.
            if (PipelinePluginsLst.Items.Count == e.Count) {
                AddPipelinePluginSeparator();
            }

            // Create display infos using default settings and add them.
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            for (int i = e.PluginIndex; i < (e.PluginIndex + e.Count); ++i) {
                Plugin newPlugin = ((PipelinePluginInfo) PipelinePluginsLst.Items[i]).ToPlugin();
                PluginDisplayInfo newDisplayInfo = new PluginDisplayInfo(newPlugin);
                newDisplayInfo.ShowInMainMenu = false;
                newDisplayInfo.ShowInSubmenu = true;
                pluginDisplayInfos.Add(newDisplayInfo);
            }
        }

        //************************************************************************
        /// <summary>
        /// Event raised when a pipeline plugin changes. We need to update
        /// the other listboxes of the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_PipelinePluginChanged(object sender, PipelinePluginEventArgs e)
        {
            // Get display info for the plugin and update it.
            ((IList<PluginDisplayInfo>) PluginsDataGrid.DataSource)[immutablePlugins.Count + 1 + e.PluginIndex].Plugin =
                ((PipelinePluginInfo) PipelinePluginsLst.Items[e.PluginIndex]).ToPlugin();
        }

        //************************************************************************
        /// <summary>
        /// Event raised when a pipeline plugin switches place with another.
        /// We need to update the other listboxes of the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_PipelinePluginMoved(object sender, PipelinePluginSwitchEventArgs e)
        {
            // Get display infos for the plugins to move.
            int lowIndex = immutablePlugins.Count + 1 + Math.Min(e.PluginIndex, e.SecondPluginIndex);
            int highIndex = immutablePlugins.Count + 1 + Math.Max(e.PluginIndex, e.SecondPluginIndex);
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            PluginDisplayInfo lowDisplayInfo = pluginDisplayInfos[lowIndex];
            PluginDisplayInfo highDisplayInfo = pluginDisplayInfos[highIndex];

            // Note if one of the moved plugin was currently selected.
            // If so we'll have to move the selection afterwards.
            int selectedIndex = -1;
            if (PluginsDataGrid.SelectedRows.Count > 0) {
                if (PluginsDataGrid.SelectedRows[0].Index == lowIndex) {
                    selectedIndex = lowIndex;
                } else if (PluginsDataGrid.SelectedRows[0].Index == highIndex) {
                    selectedIndex = highIndex;
                }
            }

            // Switch the plugin display infos.
            pluginDisplayInfos.RemoveAt(highIndex);
            pluginDisplayInfos.RemoveAt(lowIndex);
            pluginDisplayInfos.Insert(lowIndex, highDisplayInfo);
            pluginDisplayInfos.Insert(highIndex, lowDisplayInfo);

            // Change selection if we have to.
            if (selectedIndex == lowIndex) {
                PluginsDataGrid.Rows[highIndex].Selected = true;
            } else if (selectedIndex == highIndex) {
                PluginsDataGrid.Rows[lowIndex].Selected = true;
            }
        }

        //************************************************************************
        /// <summary>
        /// Event raised when a pipeline plugin is removed. We need to update
        /// the other listboxes of the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_PipelinePluginDeleted(object sender, PipelinePluginEventArgs e)
        {
            // Find display info for the plugin and simply remove it.
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            pluginDisplayInfos.RemoveAt(immutablePlugins.Count + 1 + e.PluginIndex);

            // If this leaves us without pipeline plugins, remove the separator.
            // (Note that this is called before actually removing the plugin from
            // the list of pipeline plugins!)
            if (PipelinePluginsLst.Items.Count == 1) {
                RemovePipelinePluginSeparator();
            }
        }

        //************************************************************************
        /// <summary>
        /// Event raised when some pipeline plugins are imported by the user.
        /// We need to add the new plugins to the other listboxes.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event argument.</param>
        //************************************************************************
        private void MainForm_PipelinePluginsImported(object sender, PipelinePluginsImportedEventArgs e)
        {
            // If these are the first pipeline plugins, add a separator.
            if (PipelinePluginsLst.Items.Count == e.NewCount) {
                AddPipelinePluginSeparator();
            }

            // Scan pipeline plugins that were already in the list, but update
            // their Plugin reference in case it got overwritten.
            IList<PluginDisplayInfo> pluginDisplayInfos = (IList<PluginDisplayInfo>) PluginsDataGrid.DataSource;
            for (int i = 0; i < (PipelinePluginsLst.Items.Count - e.NewCount); ++i) {
                PipelinePluginInfo pluginInfo = (PipelinePluginInfo) PipelinePluginsLst.Items[i];
                pluginDisplayInfos[immutablePlugins.Count + 1 + i].Plugin = pluginInfo.ToPlugin();
            }

            // Now add the new plugins.
            for (int i = 0; i < e.NewCount; ++i) {
                PipelinePluginInfo pluginInfo =
                    (PipelinePluginInfo) PipelinePluginsLst.Items[PipelinePluginsLst.Items.Count - e.NewCount + i];
                PluginDisplayInfo newDisplayInfo = new PluginDisplayInfo(pluginInfo.ToPlugin());
                newDisplayInfo.ShowInMainMenu = false;
                newDisplayInfo.ShowInSubmenu = true;
                pluginDisplayInfos.Add(newDisplayInfo);
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the form is first loaded. We initialize controls here.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_Load(object sender, EventArgs e)
        {
            // Create user settings object and load settings in the form.
            settings = new UserSettings();
            LoadUserSettings();
        }

        //************************************************************************
        /// <summary>
        /// Called when the form finally closes. We need to do cleanup work here.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            // Dispose of user settings object.
            Debug.Assert(settings != null);
            settings.Dispose();
        }

        //************************************************************************
        /// <summary>
        /// Called when the user clicks the "Apply" button. We'll need to save
        /// the user settings.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void ApplyBtn_Click(object sender, EventArgs e)
        {
            // Save user settings but do not close the form.
            SaveUserSettings();
        }

        //************************************************************************
        /// <summary>
        /// Called when the user clicks the "Cancel" button. We close the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void CancelBtn_Click(object sender, EventArgs e)
        {
            // Simply close the form.
            Close();
        }

        //************************************************************************
        /// <summary>
        /// Called when the user clicks the "OK" button. We save the user settings
        /// then close the form.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void OKBtn_Click(object sender, EventArgs e)
        {
            // Save user settings then close the form.
            SaveUserSettings();
            Close();
        }

        //************************************************************************
        /// <summary>
        /// Called when a cell in the plugins data grid changes its "dirty" state
        /// (e.g., someone edited the value). Since we have checkboxes, we want to
        /// commit any changes to the data source immediately when their state changes.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PluginsDataGrid_CurrentCellDirtyStateChanged(object sender, EventArgs e)
        {
            if (PluginsDataGrid.IsCurrentCellDirty && PluginsDataGrid.CurrentCell is DataGridViewCheckBoxCell) {
                PluginsDataGrid.CommitEdit(DataGridViewDataErrorContexts.Commit);
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the value of a cell in the plugins data grid changes. We
        /// use this opportunity to enable our "Apply" button.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PluginsDataGrid_CellValueChanged(object sender, DataGridViewCellEventArgs e)
        {
            // Separators have been disabled when the list was first loaded,
            // so we can be sure that any changes are legitimately on plugins.
            ApplyBtn.Enabled = true;
        }

        //************************************************************************
        /// <summary>
        /// Called when the selection in the pipeline plugins list changes. We use
        /// this opportunity to update the enabled state of the various related buttons.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PipelinePluginsLst_SelectedIndexChanged(object sender, EventArgs e)
        {
            MovePipelinePluginUpBtn.Enabled = PipelinePluginsLst.SelectedIndices.Count == 1 &&
                PipelinePluginsLst.SelectedIndex > 0;
            MovePipelinePluginDownBtn.Enabled = PipelinePluginsLst.SelectedIndices.Count == 1 &&
                PipelinePluginsLst.SelectedIndex >= 0 &&
                PipelinePluginsLst.SelectedIndex < (PipelinePluginsLst.Items.Count - 1);
            EditPipelinePluginBtn.Enabled = PipelinePluginsLst.SelectedIndices.Count == 1 &&
                !((PipelinePluginInfo) PipelinePluginsLst.SelectedItem).Global;
            RemovePipelinePluginBtn.Enabled = PipelinePluginsLst.SelectedIndices.Count == 1 &&
                !((PipelinePluginInfo) PipelinePluginsLst.SelectedItem).Global;
            ExportPipelinePluginsBtn.Enabled = PipelinePluginsLst.SelectedIndices.Count > 0;
        }

        //************************************************************************
        /// <summary>
        /// Called when the user double-clicks on a plugin in the plugins list.
        /// We will perform the same task as the "Edit" button.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PipelinePluginsLst_DoubleClick(object sender, EventArgs e)
        {
            // Simulate press of "Edit" button if it's enabled.
            if (EditPipelinePluginBtn.Enabled) {
                EditPipelinePluginBtn_Click(sender, e);
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the "Move Up" button is pressed in the pipeline plugins page.
        /// We need to modify the ordering and move a plugin up.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MovePipelinePluginUpBtn_Click(object sender, EventArgs e)
        {
            // Remove the currently-selected item, re-insert it one index
            // before that and re-select it.
            int previousIndex = PipelinePluginsLst.SelectedIndex;
            Debug.Assert(previousIndex > 0);
            PipelinePluginInfo info = (PipelinePluginInfo) PipelinePluginsLst.Items[previousIndex];
            PipelinePluginsLst.BeginUpdate();
            try {
                PipelinePluginsLst.Items.RemoveAt(previousIndex);
                PipelinePluginsLst.Items.Insert(previousIndex - 1, info);
                PipelinePluginsLst.SelectedIndex = previousIndex - 1;
            } finally {
                PipelinePluginsLst.EndUpdate();
            }

            // Warn listeners of the change.
            PipelinePluginMoved(sender, new PipelinePluginSwitchEventArgs(previousIndex - 1, previousIndex));

            // Changing the ordering actually modifies the settings.
            ApplyBtn.Enabled = true;
        }
        //************************************************************************
        /// <summary>
        /// Called when the "Move Down" button is pressed in the pipeline plugins page.
        /// We need to modify the ordering and move a plugin down.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void MovePipelinePluginDownBtn_Click(object sender, EventArgs e)
        {
            // Remove the currently-selected item, re-insert it one index
            // after that and re-select it.
            int previousIndex = PipelinePluginsLst.SelectedIndex;
            Debug.Assert(previousIndex >= 0);
            Debug.Assert(previousIndex < (PipelinePluginsLst.Items.Count - 1));
            PipelinePluginInfo info = (PipelinePluginInfo) PipelinePluginsLst.Items[previousIndex];
            PipelinePluginsLst.BeginUpdate();
            try {
                PipelinePluginsLst.Items.RemoveAt(previousIndex);
                PipelinePluginsLst.Items.Insert(previousIndex + 1, info);
                PipelinePluginsLst.SelectedIndex = previousIndex + 1;
            } finally {
                PipelinePluginsLst.EndUpdate();
            }

            // Warn listeners of the change.
            PipelinePluginMoved(sender, new PipelinePluginSwitchEventArgs(previousIndex, previousIndex + 1));

            // Changing the ordering actually modifies the settings.
            ApplyBtn.Enabled = true;
        }

        //************************************************************************
        /// <summary>
        /// Called when the "New" button of the pipeline plugins page is pressed.
        /// We need to show a dialog to allow the user to create a new plugin.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void NewPipelinePluginBtn_Click(object sender, EventArgs e)
        {
            // Ask user to create a new plugin.
            using (PipelinePluginForm editForm = new PipelinePluginForm()) {
                PipelinePluginInfo pluginInfo = editForm.EditPlugin(this, null);
                if (pluginInfo != null) {
                    // User pressed OK. Add the new plugin to the list and
                    // deselect any selection we had.
                    PipelinePluginsLst.BeginUpdate();
                    try {
                        PipelinePluginsLst.Items.Add(pluginInfo);
                        PipelinePluginsLst.ClearSelected();
                    } finally {
                        PipelinePluginsLst.EndUpdate();
                    }

                    // Warn listeners of the change.
                    PipelinePluginAdded(sender, new PipelinePluginAddEventArgs(PipelinePluginsLst.Items.Count - 1, 1));

                    // All this will enable the "Apply" button.
                    ApplyBtn.Enabled = true;
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the "Edit" button of the pipeline plugins page is pressed.
        /// We need to show a dialog to allow the user to edit an existing plugin.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void EditPipelinePluginBtn_Click(object sender, EventArgs e)
        {
            // Get reference to the selected pipeline plugin.
            int pluginInfoIndex = PipelinePluginsLst.SelectedIndex;
            Debug.Assert(pluginInfoIndex != -1);
            PipelinePluginInfo pluginInfo = (PipelinePluginInfo) PipelinePluginsLst.Items[pluginInfoIndex];

            // Ask user to edit the plugin info.
            using (PipelinePluginForm editForm = new PipelinePluginForm()) {
                PipelinePluginInfo newPluginInfo = editForm.EditPlugin(this, pluginInfo);
                if (newPluginInfo != null) {
                    // Replace the existing plugin object in the listbox.
                    Debug.Assert(newPluginInfo.Id == pluginInfo.Id);
                    PipelinePluginsLst.Items[pluginInfoIndex] = newPluginInfo;

                    // Warn listeners of the change.
                    PipelinePluginChanged(sender, new PipelinePluginEventArgs(pluginInfoIndex));

                    // All this will enable the "Apply" button.
                    ApplyBtn.Enabled = true;
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the "Remove" button of the pipeline plugins page is pressed.
        /// We need to remove the selected pipeline plugins from settings.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void RemovePipelinePluginBtn_Click(object sender, EventArgs e)
        {
            // First confirm since this is an irreversible action. We have a string
            // in resources for this message and it includes a format placeholder
            // for the plugin description.
            int infoIndex = PipelinePluginsLst.SelectedIndex;
            PipelinePluginInfo info = (PipelinePluginInfo) PipelinePluginsLst.Items[infoIndex];
            DialogResult res = MessageBox.Show(this, String.Format(Resources.REMOVE_PIPELINE_PLUGIN_MESSAGE, info.Description),
                Resources.REMOVE_PIPELINE_PLUGIN_TITLE, MessageBoxButtons.OKCancel, MessageBoxIcon.Warning);
            if (res == DialogResult.OK) {
                // Warn listeners of the change. NOTE: this needs to be done BEFORE
                // actually removing the plugin this time since that's the only way
                // a listener would know what plugin to look for.
                PipelinePluginDeleted(sender, new PipelinePluginEventArgs(infoIndex));

                // Remove currently-selected plugin, update controls and enable the Apply button.
                PipelinePluginsLst.BeginUpdate();
                try {
                    PipelinePluginsLst.Items.RemoveAt(infoIndex);
                    PipelinePluginsLst.ClearSelected();
                } finally {
                    PipelinePluginsLst.EndUpdate();
                }

                ApplyBtn.Enabled = true;
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the user presses the "Export" button to export selected
        /// pipeline plugins to a file.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void ExportPipelinePluginsBtn_Click(object sender, EventArgs e)
        {
            // Ask user where to save the data.
            if (ExportPipelinePluginsSaveDlg.ShowDialog(this) == DialogResult.OK) {
                // Create a pipeline plugins collection and add all selected plugins to it.
                PipelinePluginCollection collection = new PipelinePluginCollection();
                foreach (var pluginInfo in PipelinePluginsLst.SelectedItems) {
                    collection.Plugins.Add((PipelinePluginInfo) pluginInfo);
                }

                // Open file where to save data and serialize collection in XML.
                using (FileStream fstream = new FileStream(ExportPipelinePluginsSaveDlg.FileName, FileMode.Create)) {
                    collection.ToXML(fstream);
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the user presses the "Import" button to import pipeline
        /// plugins exported by another user (presumably).
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void ImportPipelinePluginsBtn_Click(object sender, EventArgs e)
        {
            // Ask user to select a file.
            if (ImportPipelinePluginsOpenDlg.ShowDialog(this) == DialogResult.OK) {
                // Open the file for reading and deserialize a collection of pipeline plugins.
                PipelinePluginCollection collection = null;
                using (FileStream fstream = new FileStream(ImportPipelinePluginsOpenDlg.FileName, FileMode.Open)) {
                    try {
                        collection = PipelinePluginCollection.FromXML(fstream);
                    } catch (InvalidOperationException) {
                        Debug.Assert(collection == null);
                    }
                }

                // Make sure it worked.
                if (collection != null) {
                    // Go through list of plugins and locate any plugin overwrites.
                    ImportedPipelinePluginOverwrites pluginOverwrites = new ImportedPipelinePluginOverwrites();
                    for (int i = 0; i < PipelinePluginsLst.Items.Count; ++i) {
                        PipelinePluginInfo existingInfo = (PipelinePluginInfo) PipelinePluginsLst.Items[i];
                        PipelinePluginInfo newInfo = collection.Plugins.Find(pl => pl.Equals(existingInfo));
                        if (newInfo != null) {
                            pluginOverwrites.OverwriteInfos[newInfo] = new ImportedPipelinePluginOverwrites.OverwriteInfo(existingInfo, i);
                        }
                    }

                    // Popup a form asking the user which plugins to import.
                    bool shouldContinue;
                    using (ImportPipelinePluginsForm importForm = new ImportPipelinePluginsForm()) {
                        shouldContinue = importForm.SelectPlugins(this, ref collection, ref pluginOverwrites);
                    }
                    if (shouldContinue) {
                        // Check if any plugin to import will overwrite existing ones.
                        // If so, ask user to confirm overwrite.
                        if (pluginOverwrites.OverwriteInfos.Count != 0) {
                            shouldContinue = MessageBox.Show(this, Resources.MainForm_Confirm_ImportPipelinePluginOverwrite,
                                Resources.MainForm_Confirm_ImportMsgTitle, MessageBoxButtons.OKCancel,
                                MessageBoxIcon.Warning) == DialogResult.OK;
                        }
                        if (shouldContinue) {
                            // Reset all Global flags so that imported plugins are user-specific.
                            collection.Plugins.ForEach(pl => pl.Global = false);

                            // Scan plugins to import. If the plugin already exists, overwrite the existing
                            // one, otherwise add it at the end.
                            PipelinePluginsLst.BeginUpdate();
                            try {
                                foreach (var newPluginInfo in collection.Plugins) {
                                    ImportedPipelinePluginOverwrites.OverwriteInfo overwriteInfo;
                                    if (pluginOverwrites.OverwriteInfos.TryGetValue(newPluginInfo, out overwriteInfo)) {
                                        PipelinePluginsLst.Items[overwriteInfo.OldIndex] = newPluginInfo;
                                    } else {
                                        PipelinePluginsLst.Items.Add(newPluginInfo);
                                    }
                                }
                                PipelinePluginsLst.ClearSelected();
                            } finally {
                                PipelinePluginsLst.EndUpdate();
                            }

                            // Warn listeners of the import.
                            PipelinePluginsImported(sender, new PipelinePluginsImportedEventArgs(
                                collection.Plugins.Count, pluginOverwrites.OverwriteInfos.Count));

                            // All this will enable the "Apply" button.
                            ApplyBtn.Enabled = true;
                        }
                    }
                } else {
                    // Error deserializing the XML document for some reason.
                    MessageBox.Show(this, Resources.MainForm_Error_ImportPipelinePluginBadXMLFormat,
                        Resources.MainForm_Confirm_ImportMsgTitle, MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the user checks or unchecks a checkbox in the Advanced tab.
        /// We'll enable the "Apply" button to allow the user to apply those changes.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void Chk_CheckedChanged(object sender, EventArgs e)
        {
            ApplyBtn.Enabled = true;
        }

        //************************************************************************
        /// <summary>
        /// Bean class used to display plugins in a <c>DataGridView</c>.
        /// </summary>
        //************************************************************************
        private class PluginDisplayInfo
        {
            /// <summary>
            /// <see cref="Plugin"/> object we're displaying.
            /// </summary>
            public Plugin Plugin
            {
                get;
                set;
            }

            /// <summary>
            /// Name of the plugin to show in the grid view, AKA plugin description.
            /// </summary>
            public string Name
            {
                get {
                    return Plugin.Description;
                }
            }

            /// <summary>
            /// Whether to show the plugin in the main contextual menu.
            /// </summary>
            public bool ShowInMainMenu
            {
                get;
                set;
            }

            /// <summary>
            /// Whether to show the plugin in the PCC submenu.
            /// </summary>
            public bool ShowInSubmenu
            {
                get;
                set;
            }

            /// <summary>
            /// Equivalent of <see cref="ShowInMainMenu"/> that
            /// uses an empty string for <c>false</c> and <c>1</c>
            /// for <c>true</c>. Used to be able to avoid displaying
            /// checkboxes for separators in the data grid.
            /// </summary>
            public string ShowInMainMenuStr
            {
                get {
                    return this.ShowInMainMenu ? "1" : String.Empty;
                }
                set {
                    this.ShowInMainMenu = value.Length != 0;
                }
            }

            /// <summary>
            /// Equivalent of <see cref="ShowInSubmenu"/> that
            /// uses an empty string for <c>false</c> and <c>1</c>
            /// for <c>true</c>. Used to be able to avoid displaying
            /// checkboxes for separators in the data grid.
            /// </summary>
            public string ShowInSubmenuStr
            {
                get {
                    return this.ShowInSubmenu ? "1" : String.Empty;
                }
                set {
                    this.ShowInSubmenu = value.Length != 0;
                }
            }

            //************************************************************************
            /// <summary>
            /// Constructor.
            /// </summary>
            /// <param name="plugin"><see cref="Plugin"/> object to wrap.</param>
            //************************************************************************
            public PluginDisplayInfo(Plugin plugin)
            {
                this.Plugin = plugin;
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the user clicks on the link to visit the PCC website.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void CodePlexSiteLinkLbl_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            // Link is the control's text.
            Process.Start(CodePlexSiteLinkLbl.Text);
        }

        //************************************************************************
        /// <summary>
        /// Called when the user clicks on the link to see the license file.
        /// We need to locate the license file on disk and open it.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void LicenseTxtLinkLbl_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            // In a normal installation, the license file will be located right next to the executable.
            string licensePath = Path.Combine(Path.GetDirectoryName(Path.GetFullPath(
                Assembly.GetExecutingAssembly().Location)), LICENSE_TXT_FILE);
            FileInfo licenseFileInfo = new FileInfo(licensePath);
            if (!licenseFileInfo.Exists) {
                // In a dev environment, license file is three directories up.
                FileInfo[] fileInfos = null;
                DirectoryInfo dirInfo = licenseFileInfo.Directory;
                for (int i = 0; dirInfo != null && i < 3; ++i) {
                    dirInfo = dirInfo.Parent;
                }
                if (dirInfo != null) {
                    fileInfos = dirInfo.GetFiles(LICENSE_TXT_FILE);
                }
                if (fileInfos != null && fileInfos.Length != 0) {
                    Debug.Assert(fileInfos.Length == 1);
                    licenseFileInfo = fileInfos[0];
                } else {
                    licenseFileInfo = null;
                }
            }
            if (licenseFileInfo != null) {
                // Found license file, launch it in default editor.
                Process.Start(licenseFileInfo.FullName);
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when a link has been clicked related to the donations page.
        /// The URL to follow is in the link data.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void DonationLinkLbl_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            Process.Start(e.Link.LinkData as string);
        }
    }
}
