﻿// PipelinePluginForm.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Windows.Forms;
using PathCopyCopy.Settings.Properties;
using PCC;

namespace PathCopyCopy.Settings
{
    //************************************************************************
    /// <summary>
    /// Form used to create or edit a pipeline plugin. To use, create the form
    /// and call the <see cref="EditPlugin"/> method.
    /// </summary>
    //************************************************************************
    public partial class PipelinePluginForm : Form
    {
        /// Plugin info for the plugin we're editing.
        private PipelinePluginInfo pluginInfo;

        //************************************************************************
        /// <summary>
        /// Constructor.
        /// </summary>
        //************************************************************************
        public PipelinePluginForm()
        {
            InitializeComponent();
        }

        //************************************************************************
        /// <summary>
        /// Main method to use the form. Will show the form as a modal dialog,
        /// allowing the user to edit the given pipeline plugin. When the user
        /// is done, the method will return the new <see cref="PipelinePluginInfo"/>
        /// if the user accepted the changes.
        /// </summary>
        /// <param name="owner">Owner of this dialog. Can be <c>null</c>.</param>
        /// <param name="oldInfo">Info about a pipeline plugin. If set, we'll
        /// populate the form with the plugin's values to allow the user to
        /// edit the plugin.</param>
        /// <returns>Info about the new plugin that user edited. Will be
        /// <c>null</c> if user cancels editing.</returns>
        //************************************************************************
        public PipelinePluginInfo EditPlugin(IWin32Window owner, PipelinePluginInfo oldInfo)
        {
            // Save old info so that the OnLoad event handler can use it.
            pluginInfo = oldInfo;

            // Show the form and check result.
            DialogResult dialogRes = ShowDialog(owner);

            // If user saved, return the new info.
            Debug.Assert(dialogRes != DialogResult.OK || pluginInfo != null);
            return dialogRes == DialogResult.OK ? pluginInfo : null;
        }

        //************************************************************************
        /// <summary>
        /// Called when the form is loaded. We use this opportunity to populate
        /// the controls with info about a plugin we're editing, if we have one.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PipelinePluginForm_Load(object sender, EventArgs e)
        {
            // First load list of plugins to display in the listbox for the base
            // plugin. We only load default and COM plugins for this since we
            // don't want a pipeline plugin to be based off another (for now at least).
            List<Plugin> plugins = DefaultPlugins.GetDefaultPlugins();
            List<Plugin> comPlugins = COMPlugins.GetCOMPlugins();
            if (comPlugins.Count != 0) {
                plugins.Add(new SeparatorPlugin());
                plugins.AddRange(comPlugins);
            }

            // Add all plugins to the list box.
            BasePluginLst.Items.AddRange(plugins.ToArray());

            // Check if we have a plugin info to load.
            Guid? basePluginId = null;
            if (pluginInfo != null) {
                // Decode the plugin's pipeline elements.
                Pipeline pipeline = PipelineDecoder.DecodePipeline(pluginInfo.EncodedElements);

                // Populate our controls.
                NameTxt.Text = pluginInfo.Description;
                PipelineElement element = pipeline.Elements.Find(el => el is ApplyPluginPipelineElement);
                if (element != null) {
                    basePluginId = ((ApplyPluginPipelineElement) element).PluginID;
                }
                QuotesChk.Checked = pipeline.Elements.Find(el => el is QuotesPipelineElement) != null;
                if (pipeline.Elements.Find(el => el is BackToForwardSlashesPipelineElement) != null) {
                    BackToForwardSlashesRadio.Checked = true;
                } else if (pipeline.Elements.Find(el => el is ForwardToBackslashesPipelineElement) != null) {
                    ForwardToBackslashesRadio.Checked = true;
                } else {
                    Debug.Assert(NoSlashesChangeRadio.Checked);
                }
                element = pipeline.Elements.Find(el => el is RegexPipelineElement);
                if (element != null) {
                    UseRegexChk.Checked = true;
                    IgnoreCaseChk.Enabled = true;
                    TestRegexBtn.Enabled = true;
                    RegexPipelineElement regexElement = (RegexPipelineElement) element;
                    FindTxt.Text = regexElement.Regex;
                    ReplaceTxt.Text = regexElement.Format;
                    IgnoreCaseChk.Checked = regexElement.IgnoreCase;
                } else {
                    Debug.Assert(!IgnoreCaseChk.Enabled);
                    Debug.Assert(!TestRegexBtn.Enabled);
                    element = pipeline.Elements.Find(el => el is FindReplacePipelineElement);
                    if (element != null) {
                        FindTxt.Text = ((FindReplacePipelineElement) element).OldValue;
                        ReplaceTxt.Text = ((FindReplacePipelineElement) element).NewValue;
                    }
                }
            }

            // Validate base plugin ID if found.
            if (basePluginId != null) {
                if (plugins.Find(pl => pl.ID == basePluginId) == null) {
                    // Invalid, clear it so that we'll fall back
                    // on the default below.
                    basePluginId = null;
                }
            }

            // If we didn't find a base plugin ID yet, use the "Long Path" plugin ID.
            if (basePluginId == null) {
                basePluginId = new Guid(Resources.LONG_PATH_PLUGIN_ID);
            }

            // Scan list of plugins to select the base plugin.
            for (int i = 0; i < BasePluginLst.Items.Count; ++i) {
                if (!(BasePluginLst.Items[i] is SeparatorPlugin) &&
                    ((Plugin) BasePluginLst.Items[i]).ID == basePluginId) {

                    // This is the default plugin, select it and break.
                    BasePluginLst.SelectedIndex = i;
                    break;
                }
            }
            Debug.Assert(BasePluginLst.SelectedIndex != -1);
        }

        //************************************************************************
        /// <summary>
        /// Called when the form is about to close. If user pressed OK, we save
        /// the content of the plugin here.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void PipelinePluginForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            // If user chose to press OK, save plugin info.
            if (this.DialogResult == DialogResult.OK) {
                // Create a pipeline based on form controls.
                Pipeline pipeline = new Pipeline();
                if (BasePluginLst.SelectedIndex != -1) {
                    pipeline.Elements.Add(new ApplyPluginPipelineElement(
                        ((Plugin) BasePluginLst.SelectedItem).ID));
                }
                if (FindTxt.Text.Length > 0) {
                    if (UseRegexChk.Checked) {
                        pipeline.Elements.Add(new RegexPipelineElement(FindTxt.Text, ReplaceTxt.Text, IgnoreCaseChk.Checked));
                    } else {
                        pipeline.Elements.Add(new FindReplacePipelineElement(FindTxt.Text, ReplaceTxt.Text));
                    }
                }
                if (BackToForwardSlashesRadio.Checked) {
                    pipeline.Elements.Add(new BackToForwardSlashesPipelineElement());
                } else if (ForwardToBackslashesRadio.Checked) {
                    pipeline.Elements.Add(new ForwardToBackslashesPipelineElement());
                }
                if (QuotesChk.Checked) {
                    pipeline.Elements.Add(new QuotesPipelineElement());
                }

                // Create plugin info if we don't already have one.
                if (pluginInfo == null) {
                    pluginInfo = new PipelinePluginInfo();
                    pluginInfo.Id = Guid.NewGuid();
                }

                // Save info in plugin info wrapper.
                pluginInfo.Description = NameTxt.Text;
                pluginInfo.EncodedElements = pipeline.Encode();
                Debug.Assert(!pluginInfo.Global);
            }
        }

        //************************************************************************
        /// <summary>
        /// Called when the user checks or unchecks the "Use regular expressions"
        /// checkbox. We need to enable or disable the "ignore case" checkbox
        /// when this occurs, as this flag is only used with regular expressions.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void UseRegexChk_CheckedChanged(object sender, EventArgs e)
        {
            IgnoreCaseChk.Enabled = UseRegexChk.Checked;
            TestRegexBtn.Enabled = UseRegexChk.Checked;
        }

        //************************************************************************
        /// <summary>
        /// Called when the user presses the button to test a regular expression.
        /// We will show a dialog allowing the user to test the currently-provided
        /// expression and optionally change it.
        /// </summary>
        /// <param name="sender">Event sender.</param>
        /// <param name="e">Event arguments.</param>
        //************************************************************************
        private void TestRegexBtn_Click(object sender, EventArgs e)
        {
            // Grab content of controls so that the tester form
            // can optionally modify them.
            string regex = FindTxt.Text;
            string format = ReplaceTxt.Text;
            bool ignoreCase = IgnoreCaseChk.Checked;

            // Show form and ask user to test. See what happens.
            using (RegexTesterForm testerForm = new RegexTesterForm()) {
                DialogResult res = testerForm.TestRegex(ref regex, ref format, ref ignoreCase, this);
                if (res == DialogResult.OK) {
                    // User pressed OK which means we have updated values.
                    FindTxt.Text = regex;
                    ReplaceTxt.Text = format;
                    IgnoreCaseChk.Checked = ignoreCase;
                }
            }
        }
    }
}
