﻿// PipelinePlugins.cs
// (c) 2011-2012, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using System.Xml;

namespace PCC
{
    //************************************************************************
    /// <summary>
    /// Wrapper for info about a pipeline plugin found in the registry.
    /// </summary>
    //************************************************************************
    public sealed class PipelinePluginInfo : IComparable<PipelinePluginInfo>, IComparable,
        IEquatable<PipelinePluginInfo>
    {
        /// <summary>
        /// ID of this pipeline plugin.
        /// </summary>
        [XmlIgnore]
        public Guid Id
        {
            get;
            set;
        }

        /// <summary>
        /// ID of this pipeline plugin, as a string. Used for XML serialization.
        /// </summary>
        /// <seealso cref="Id"/>
        [XmlElement("Id")]
        public string IdAsString
        {
            get {
                return Id.ToString("B");
            }
            set {
                Id = new Guid(value);
            }
        }

        /// <summary>
        /// Description associated with the pipeline plugin.
        /// </summary>
        [XmlElement]
        public string Description
        {
            get;
            set;
        }

        /// <summary>
        /// String containing encoded pipeline elements.
        /// </summary>
        [XmlElement("Pipeline")]
        public string EncodedElements
        {
            get;
            set;
        }

        /// <summary>
        /// Whether this pipeline plugin is stored in the global key.
        /// Global plugins cannot be deleted or modified.
        /// </summary>
        [XmlElement]
        public bool Global
        {
            get;
            set;
        }

        //************************************************************************
        /// <summary>
        /// Constructor with no arguments.
        /// </summary>
        //************************************************************************
        public PipelinePluginInfo()
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with information.
        /// </summary>
        /// <param name="id">ID of the pipeline plugin.</param>
        /// <param name="description">Pipeline plugin description.</param>
        /// <param name="encodedElements">String containing encoded pipeline
        /// elements.</param>
        /// <param name="isGlobal">Whether this plugin is in the global key.</param>
        //************************************************************************
        public PipelinePluginInfo(Guid id, string description, string encodedElements, bool isGlobal)
        {
            Id = id;
            Description = description;
            EncodedElements = encodedElements;
            Global = isGlobal;
        }

        //************************************************************************
        /// <summary>
        /// Returns a textual representation of the plugin info (e.g. its description).
        /// </summary>
        /// <returns>Plugin info description.</returns>
        //************************************************************************
        public override string ToString()
        {
            return Description;
        }

        //************************************************************************
        /// <summary>
        /// Converts this pipeline plugin info into a plugin wrapper.
        /// </summary>
        /// <returns>Plugin wrapper.</returns>
        //************************************************************************
        public Plugin ToPlugin()
        {
            return new Plugin(Id, Description);
        }

        //************************************************************************
        /// <summary>
        /// Utility method to convert a list of pipeline plugin infos into plugins.
        /// </summary>
        /// <param name="pluginInfos">List of pipeline plugin infos.</param>
        /// <returns>List of plugins.</returns>
        //************************************************************************
        public static List<Plugin> ToPlugins(List<PipelinePluginInfo> pluginInfos)
        {
            Debug.Assert(pluginInfos != null);
            return pluginInfos.ConvertAll(info => info.ToPlugin());
        }

        #region IComparable<PipelinePluginInfo> Members

        //************************************************************************
        /// <summary>
        /// Compares this pipeline plugin info with another. They are compared
        /// through their <see cref="PipelinePluginInfo.Id"/>.
        /// </summary>
        /// <param name="other">Other pipeline plug to compare ourselves to.</param>
        /// <returns>Relative position of <c>this</c> compared to
        /// <paramref name="other"/>.</returns>
        //************************************************************************
        public int CompareTo(PipelinePluginInfo other)
        {
            return other != null ? Id.CompareTo(other.Id) : 1;
        }

        #endregion

        #region IComparable Members

        //************************************************************************
        /// <summary>
        /// Compares this pipeline plugin info with another. They are compared
        /// through their <see cref="PipelinePluginInfo.Id"/>. This is the generic
        /// version for the <see cref="T:IComparable"/> interface.
        /// </summary>
        /// <param name="obj">Other pipeline plug to compare ourselves to.</param>
        /// <returns>Relative position of <c>this</c> compared to
        /// <paramref name="obj"/>.</returns>
        //************************************************************************
        int IComparable.CompareTo(object obj)
        {
            return (obj != null && obj is PipelinePluginInfo) ? this.CompareTo((PipelinePluginInfo) obj) : 1;
        }

        #endregion

        #region IEquatable<PipelinePluginInfo> Members

        //************************************************************************
        /// <summary>
        /// Compares two pipeline plugin infos for equality. They are compared
        /// through their <see cref="PipelinePluginInfo.Id"/>.
        /// </summary>
        /// <param name="other">Pipeline plugin info to compare ourselves with.</param>
        /// <returns>Whether <c>this</c> is equal to <paramref name="other"/>.</returns>
        //************************************************************************
        public bool Equals(PipelinePluginInfo other)
        {
            return other != null ? Id.Equals(other.Id) : false;
        }

        #endregion
    }

    //************************************************************************
    /// <summary>
    /// Collection of <see cref="T:PipelinePluginInfo"/> objects. Can be used
    /// to save and load pipeline plugins to and from XML.
    /// </summary>
    //************************************************************************
    [XmlRoot(Namespace = PipelinePluginCollection.PIPELINE_PLUGINS_XML_NAMESPACE)]
    public class PipelinePluginCollection
    {
        /// List storing the pipeline plugins in the collection.
        private List<PipelinePluginInfo> infos = new List<PipelinePluginInfo>();

        /// <summary>
        /// XML namespace used to serialize pipeline plugins.
        /// </summary>
        public const string PIPELINE_PLUGINS_XML_NAMESPACE = "http://pathcopycopy.codeplex.com/xsd/PipelinePlugins/V1";

        /// <summary>
        /// Pipeline plugins in the collection.
        /// </summary>
        [XmlArray]
        public List<PipelinePluginInfo> Plugins
        {
            get {
                return infos;
            }
            set {
                infos = value;
            }
        }

        //************************************************************************
        /// <summary>
        /// Serializes this pipeline plugins collection to XML format that will
        /// later be accepted by the <see cref="FromXML"/> method.
        /// </summary>
        /// <param name="stream"><see cref="T:Stream"/> where to store the XML data.</param>
        //************************************************************************
        public void ToXML(Stream stream)
        {
            Debug.Assert(stream != null);

            XmlSerializer serializer = new XmlSerializer(typeof(PipelinePluginCollection), PIPELINE_PLUGINS_XML_NAMESPACE);
            using (XmlWriter writer = XmlWriter.Create(stream)) {
                serializer.Serialize(writer, this);
            }
        }

        //************************************************************************
        /// <summary>
        /// Deserializes an <see cref="T:PipelinePluginCollection"/> object from
        /// XML data.
        /// </summary>
        /// <param name="stream"><see cref="T:Stream"/> containing the XML.</param>
        /// <returns>New pipeline plugins collection from XML.</returns>
        //************************************************************************
        public static PipelinePluginCollection FromXML(Stream stream)
        {
            Debug.Assert(stream != null);

            XmlSerializer serializer = new XmlSerializer(typeof(PipelinePluginCollection), PIPELINE_PLUGINS_XML_NAMESPACE);
            using (XmlReader reader = XmlReader.Create(stream)) {
                return (PipelinePluginCollection) serializer.Deserialize(reader);
            }
        }
    }

    //************************************************************************
    /// <summary>
    /// Base class for pipeline elements to be chained in a pipeline plugin.
    /// Since we only care about configuration in the settings app, it merely
    /// contains methods to encode element data to a string.
    /// </summary>
    //************************************************************************
    abstract public class PipelineElement
    {
        /// <summary>
        /// Every pipeline element type must be represented by a specific code.
        /// The pipeline uses this code to encode its elements.
        /// </summary>
        abstract public char Code
        {
            get;
        }

        //************************************************************************
        /// <summary>
        /// Method that must be overridden in subclasses to encode the data of
        /// the element in a string. This will be saved in the registry in the
        /// plugin's encoded elements.
        /// </summary>
        /// <param name="p_Param">Parameter summary</param>
        /// <returns>Return value summary</returns>
        //************************************************************************
        abstract public string Encode();

        //************************************************************************
        /// <summary>
        /// Encodes the specified int value in a format suitable to be included in
        /// encoded element data.
        /// </summary>
        /// <param name="toEncode">Int value to encode.</param>
        /// <returns>Encoded string.</returns>
        //************************************************************************
        public static string EncodeInt(int toEncode)
        {
            // Encode as four characters.
            Debug.Assert(toEncode < 10000);
            return toEncode.ToString("D4");
        }

        //************************************************************************
        /// <summary>
        /// Encodes the specified string in a format suitable to be included in
        /// encoded element data.
        /// </summary>
        /// <param name="toEncode">String to encode.</param>
        /// <returns>Encoded string.</returns>
        //************************************************************************
        public static string EncodeString(string toEncode)
        {
            // First write the string length, then the string content.
            return EncodeInt(toEncode.Length) + toEncode;
        }

        //************************************************************************
        /// <summary>
        /// Encodes the specified boolean in a format suitable to be included in
        /// encoded element data.
        /// </summary>
        /// <param name="toEncode">Bool to encode.</param>
        /// <returns>Encoded string.</returns>
        //************************************************************************
        public static string EncodeBool(bool toEncode)
        {
            // Encode as 0 or 1.
            return toEncode ? "1" : "0";
        }
    }

    //************************************************************************
    /// <summary>
    /// Class representing the pipeline used by a pipeline plugin. Contains an
    /// ordered list of pipeline elements.
    /// </summary>
    /// <remarks>
    /// Please refer to the C++ code in "PluginPipelineElements.cpp" and
    /// "PluginPipelineDecoder" to see the opposite side of the encoding/decoding
    /// work.</remarks>
    //************************************************************************
    public sealed class Pipeline
    {
        /// List containing all pipeline elements.
        private List<PipelineElement> pipelineElements;

        /// <summary>
        /// List containing all the pipeline elements.
        /// </summary>
        public List<PipelineElement> Elements
        {
            get {
                return pipelineElements;
            }
        }

        //************************************************************************
        /// <summary>
        /// Constructor.
        /// </summary>
        //************************************************************************
        public Pipeline()
        {
            pipelineElements = new List<PipelineElement>();
        }

        //************************************************************************
        /// <summary>
        /// Encodes the pipeline's elements in a string to be saved in the plugin data.
        /// </summary>
        /// <returns>Encoded pipeline elements.</returns>
        //************************************************************************
        public string Encode()
        {
            // First write the number of elements in the pipeline as a two-char number.
            Debug.Assert(Elements.Count < 100);
            StringBuilder encodedElements = new StringBuilder();
            encodedElements.Append(Elements.Count.ToString("D2"));

            // Encode each pipeline element by first appending its code then
            // asking the element itself to encode its data.
            foreach (PipelineElement element in Elements) {
                encodedElements.Append(element.Code);
                encodedElements.Append(element.Encode());
            }

            return encodedElements.ToString();
        }
    }

    //************************************************************************
    /// <summary>
    /// Exception class used to represent an invalid pipeline encoded string.
    /// </summary>
    //************************************************************************
    public class InvalidPipelineException : Exception
    {
        /// <summary>
        /// Encoded elements in the invalid pipeline.
        /// </summary>
        public string EncodedElements
        {
            get;
            protected set;
        }

        //************************************************************************
        /// <summary>
        /// Constructor.
        /// </summary>
        //************************************************************************
        public InvalidPipelineException()
            : base("Invalid pipeline.")
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with encoded elements.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements.</param>
        //************************************************************************
        public InvalidPipelineException(string encodedElements)
            : base(String.Format("Invalid pipeline: {0}", encodedElements))
        {
            EncodedElements = encodedElements;
        }

        //************************************************************************
        /// <summary>
        /// Constructor with inner exception.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements.</param>
        /// <param name="innerException">Inner exception object.</param>
        //************************************************************************
        public InvalidPipelineException(string encodedElements, Exception innerException)
            : base(String.Format("Invalid pipeline: {0}", encodedElements), innerException)
        {
            EncodedElements = encodedElements;
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that instructs Path Copy Copy to add quotes around
    /// the returned path.
    /// </summary>
    //************************************************************************
    public class QuotesPipelineElement : PipelineElement
    {
        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '"';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // No other data to encode.
            return string.Empty;
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that replaces all backslashes found in the path with
    /// forward slashes.
    /// </summary>
    //************************************************************************
    public class BackToForwardSlashesPipelineElement : PipelineElement
    {
        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '\\';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // No other data to encode.
            return string.Empty;
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that replaces all forward slashes found in the path
    /// with backslashes.
    /// </summary>
    //************************************************************************
    public class ForwardToBackslashesPipelineElement : PipelineElement
    {
        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '/';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // No other data to encode.
            return string.Empty;
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that performs a find & replace operation in the path.
    /// </summary>
    //************************************************************************
    public class FindReplacePipelineElement : PipelineElement
    {
        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '?';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        /// <summary>
        /// Value to look for in the path.
        /// </summary>
        /// <seealso cref="NewValue"/>
        public string OldValue
        {
            get;
            set;
        }

        /// <summary>
        /// Replacement value.
        /// </summary>
        /// <seealso cref="OldValue"/>
        public string NewValue
        {
            get;
            set;
        }

        //************************************************************************
        /// <summary>
        /// Default constructor.
        /// </summary>
        //************************************************************************
        public FindReplacePipelineElement()
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with old and new values.
        /// </summary>
        /// <param name="oldValue">Value to look for.</param>
        /// <param name="newValue">Replacement value.</param>
        //************************************************************************
        public FindReplacePipelineElement(string oldValue, string newValue)
        {
            OldValue = oldValue;
            NewValue = newValue;
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // Encode old and new value one after the other.
            return EncodeString(OldValue) + EncodeString(NewValue);
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that performs find/replace operations using regex.
    /// </summary>
    //************************************************************************
    public class RegexPipelineElement : PipelineElement
    {
        /// <summary>
        /// Version number used to identify encoded data.
        /// </summary>
        public const int INITIAL_VERSION = 1;

        /// <summary>
        /// Maximum data version understood by this code.
        /// </summary>
        public const int MAX_VERSION = INITIAL_VERSION;

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '^';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        /// <summary>
        /// Regular expression to use to find matches.
        /// </summary>
        /// <remarks>
        /// The shell extension doesn't use the .NET regex syntax since it
        /// uses the std::regex class to perform matches. It uses the
        /// ECMAScript regex syntax.
        /// </remarks>
        public string Regex
        {
            get;
            set;
        }

        /// <summary>
        /// Format of replacement string.
        /// </summary>
        public string Format
        {
            get;
            set;
        }

        /// <summary>
        /// Whether to ignore case when looking for matches.
        /// </summary>
        public bool IgnoreCase
        {
            get;
            set;
        }

        //************************************************************************
        /// <summary>
        /// Default constructor.
        /// </summary>
        //************************************************************************
        public RegexPipelineElement()
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with arguments.
        /// </summary>
        /// <param name="regex">Regular expression to use.</param>
        /// <param name="format">Format of replacement string.</param>
        /// <param name="ignoreCase">Whether to ignore case when looking for
        /// matches.</param>
        //************************************************************************
        public RegexPipelineElement(string regex, string format, bool ignoreCase)
        {
            Regex = regex;
            Format = format;
            IgnoreCase = ignoreCase;
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // First write currently-used version number.
            StringBuilder encoder = new StringBuilder();
            encoder.Append(EncodeInt(INITIAL_VERSION));

            // Now encode regex, format string and ignore case flag.
            encoder.Append(EncodeString(Regex));
            encoder.Append(EncodeString(Format));
            encoder.Append(EncodeBool(IgnoreCase));

            return encoder.ToString();
        }
    }

    //************************************************************************
    /// <summary>
    /// Pipeline element that applies the effect of another plugin on the path.
    /// </summary>
    //************************************************************************
    public class ApplyPluginPipelineElement : PipelineElement
    {
        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public const char CODE = '{';

        /// <summary>
        /// Code representing this pipeline element type.
        /// </summary>
        public override char Code
        {
            get {
                return CODE;
            }
        }

        /// <summary>
        /// ID of plugin to apply to the path.
        /// </summary>
        public Guid PluginID
        {
            get;
            set;
        }

        //************************************************************************
        /// <summary>
        /// Default constructor.
        /// </summary>
        //************************************************************************
        public ApplyPluginPipelineElement()
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with plugin ID.
        /// </summary>
        /// <param name="pluginId">ID of plugin to apply to the path.</param>
        //************************************************************************
        public ApplyPluginPipelineElement(Guid pluginId)
        {
            PluginID = pluginId;
        }

        //************************************************************************
        /// <summary>
        /// Encodes this pipeline element in a string.
        /// </summary>
        /// <returns>Encoded element data.</returns>
        //************************************************************************
        public override string Encode()
        {
            // We encode our guid without length info, since the C++
            // side will know how many characters to read.
            return PluginID.ToString("B");
        }
    }

    //************************************************************************
    /// <summary>
    /// Static class that can decode a pipeline of multiple elements from an
    /// encoded string. This is the C# equivalent of the C++'s PipelineDecoder.
    /// </summary>
    //************************************************************************
    public static class PipelineDecoder
    {
        //************************************************************************
        /// <summary>
        /// Decodes the content of a pipeline using encoded element data.
        /// </summary>
        /// <param name="encodedElements">Encoded elements data.</param>
        /// <returns>New <see cref="Pipeline"/> object.</returns>
        //************************************************************************
        public static Pipeline DecodePipeline(string encodedElements)
        {
            // Initialize new pipeline object.
            Pipeline pipeline = new Pipeline();

            try {
                // The first two characters of the string should be the number of elements.
                if (encodedElements.Length < 2) {
                    throw new InvalidPipelineException();
                }
                int numElements = int.Parse(encodedElements.Substring(0, 2));

                // Loop to read each element.
                int curChar = 2;
                for (int i = 0; i < numElements; ++i) {
                    pipeline.Elements.Add(DecodeElement(encodedElements, ref curChar));
                }
                if (curChar < encodedElements.Length) {
                    throw new InvalidPipelineException();
                }
            } catch (Exception e) {
                // Re-throw an exception with encoded elements info.
                throw new InvalidPipelineException(encodedElements, e);
            }

            return pipeline;
        }

        //************************************************************************
        /// <summary>
        /// Decodes a pipeline element at a specific point in an encoded string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the element data is located.
        /// Upon return, this value will point just after the element data.</param>
        /// <returns>New <see cref="PipelineElement"/> instance.</returns>
        //************************************************************************
        private static PipelineElement DecodeElement(string encodedElements, ref int curChar)
        {
            PipelineElement element = null;

            // First read the element code.
            if (curChar == encodedElements.Length) {
                throw new InvalidPipelineException();
            }
            char elementCode = encodedElements[curChar++];
            switch (elementCode) {
                case QuotesPipelineElement.CODE: {
                    element = new QuotesPipelineElement();
                    break;
                }
                case BackToForwardSlashesPipelineElement.CODE: {
                    element = new BackToForwardSlashesPipelineElement();
                    break;
                }
                case ForwardToBackslashesPipelineElement.CODE: {
                    element = new ForwardToBackslashesPipelineElement();
                    break;
                }
                case FindReplacePipelineElement.CODE: {
                    element = DecodeFindReplaceElement(encodedElements, ref curChar);
                    break;
                }
                case RegexPipelineElement.CODE: {
                    element = DecodeRegexPipelineElement(encodedElements, ref curChar);
                    break;
                }
                case ApplyPluginPipelineElement.CODE: {
                    element = DecodeApplyPluginElement(encodedElements, ref curChar);
                    break;
                }
                default:
                    // Invalid pipeline. PCC downgrade, maybe?
                    throw new InvalidPipelineException();
            }

            Debug.Assert(element != null);
            return element;
        }

        //************************************************************************
        /// <summary>
        /// Decodes a <see cref="FindReplacePipelineElement"/> from an encoded
        /// element string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the element data is to be found
        /// in the string (not counting the element code). Upon return, this will
        /// point just after the element data.</param>
        //************************************************************************
        private static FindReplacePipelineElement DecodeFindReplaceElement(string encodedElements,
            ref int curChar)
        {
            // The element data contains the old and new value.
            string oldValue = DecodeString(encodedElements, ref curChar);
            string newValue = DecodeString(encodedElements, ref curChar);
            return new FindReplacePipelineElement(oldValue, newValue);
        }

        //************************************************************************
        /// <summary>
        /// Decodes a <see cref="RegexPipelineElement"/> from an encoded element
        /// string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the element data is to be found
        /// in the string (not counting the element code). Upon return, this will
        /// point just after the element data.</param>
        //************************************************************************
        private static RegexPipelineElement DecodeRegexPipelineElement(string encodedElements,
            ref int curChar)
        {
            // First read version number and validate.
            int version = DecodeInt(encodedElements, ref curChar);
            if (version > RegexPipelineElement.MAX_VERSION) {
                throw new InvalidPipelineException();
            }

            // Read initial version data: regex, format and ignore case flag.
            string regex = DecodeString(encodedElements, ref curChar);
            string format = DecodeString(encodedElements, ref curChar);
            bool ignoreCase = DecodeBool(encodedElements, ref curChar);

            // Create and return element object.
            return new RegexPipelineElement(regex, format, ignoreCase);
        }

        //************************************************************************
        /// <summary>
        /// Decodes an <see cref="ApplyPluginPipelineElement"/> from an encoded
        /// element string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the element data is to be found
        /// in the string (not counting the element code). Upon return, this will
        /// point just after the element data.</param>
        //************************************************************************
        private static ApplyPluginPipelineElement DecodeApplyPluginElement(string encodedElements,
            ref int curChar)
        {
            // The element data is the ID of the plugin to apply. There's no length saved
            // for this since it's always the same length. Format is as follows:
            // {xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}
            int guidLength = Guid.Empty.ToString("B").Length;
            if (curChar > (encodedElements.Length - guidLength)) {
                throw new InvalidPipelineException();
            }
            Guid pluginId = new Guid(encodedElements.Substring(curChar, guidLength));
            curChar += guidLength;
            return new ApplyPluginPipelineElement(pluginId);
        }

        //************************************************************************
        /// <summary>
        /// Decodes an integer value from an encoded element string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the integer data is to be found
        /// in <paramref name="encodedElements"/>. Upon return, this will point
        /// just after the integer data.</param>
        /// <returns>Decoded integer value.</returns>
        //************************************************************************
        private static int DecodeInt(string encodedElements, ref int curChar)
        {
            // Encoded as four characters.
            if (curChar > (encodedElements.Length - 4)) {
                throw new InvalidPipelineException();
            }
            int intValue = int.Parse(encodedElements.Substring(curChar, 4));
            curChar += 4;
            return intValue;
        }

        //************************************************************************
        /// <summary>
        /// Decodes a string from an encoded element string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the string data is to be found
        /// in <paramref name="encodedElements"/>. Upon return, this will point
        /// just after the string data.</param>
        /// <returns>Decoded string.</returns>
        //************************************************************************
        private static string DecodeString(string encodedElements, ref int curChar)
        {
            // Read string length.
            int stringLength = DecodeInt(encodedElements, ref curChar);

            // Return string content.
            if (curChar > (encodedElements.Length - stringLength)) {
                throw new InvalidPipelineException();
            }
            int prevChar = curChar;
            curChar += stringLength;
            return encodedElements.Substring(prevChar, stringLength);
        }

        //************************************************************************
        /// <summary>
        /// Decodes a boolean from an encoded element string.
        /// </summary>
        /// <param name="encodedElements">String of encoded elements data.</param>
        /// <param name="curChar">Position where the boolean data is to be found
        /// in <paramref name="encodedElements"/>. Upon return, this will point
        /// just after the boolean data.</param>
        /// <returns>Decoded boolean value.</returns>
        //************************************************************************
        private static bool DecodeBool(string encodedElements, ref int curChar)
        {
            // Encoded as 0 or 1.
            if (curChar == encodedElements.Length) {
                throw new InvalidPipelineException();
            }
            if (encodedElements[curChar] != '0' && encodedElements[curChar] != '1') {
                throw new InvalidPipelineException();
            }
            return encodedElements[curChar++] == '1';
        }
    }

    //************************************************************************
    /// <summary>
    /// Container of information used when importing pipeline plugins. It will
    /// store information about plugins overwriting existing ones.
    /// </summary>
    //************************************************************************
    public sealed class ImportedPipelinePluginOverwrites
    {
        /// Dictionary storing overwrite information.
        private Dictionary<PipelinePluginInfo, OverwriteInfo> overwriteInfos = new Dictionary<PipelinePluginInfo, OverwriteInfo>();

        /// <summary>
        /// Dictionary storing information about overwritten plugins. The key
        /// is the new pipeline plugin infos and the value is the container
        /// containing information about the old plugin being overwritten.
        /// </summary>
        /// <seealso cref="T:ImportedPipelinePluginOverwrites.OverwriteInfo"/>
        public Dictionary<PipelinePluginInfo, OverwriteInfo> OverwriteInfos
        {
            get {
                return overwriteInfos;
            }
        }

        //************************************************************************
        /// <summary>
        /// Information about an overwritten pipeline plugin.
        /// </summary>
        //************************************************************************
        public sealed class OverwriteInfo
        {
            /// <summary>
            /// Reference to the old pipeline plugin info that will be overwritten.
            /// </summary>
            public PipelinePluginInfo OldInfo
            {
                get;
                private set;
            }

            /// <summary>
            /// Index of the pipeline plugin that will be overwritten in the list
            /// of existing pipeline plugins.
            /// </summary>
            public int OldIndex
            {
                get;
                private set;
            }

            //************************************************************************
            /// <summary>
            /// Constructor.
            /// </summary>
            /// <param name="oldInfo">Reference to old pipeline plugin info.</param>
            /// <param name="oldIndex">Index of old pipeline plugin in the list of
            /// existing pipeline plugins.</param>
            //************************************************************************
            public OverwriteInfo(PipelinePluginInfo oldInfo, int oldIndex)
            {
                Debug.Assert(oldInfo != null);

                this.OldInfo = oldInfo;
                this.OldIndex = oldIndex;
            }
        }
    }
}
