﻿// Plugin.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

namespace PCC
{
    //************************************************************************
    /// <summary>
    /// Bean class representing a Path Copy Copy plugin. Contains the plugin
    /// description as well as its unique ID.
    /// </summary>
    //************************************************************************
    public class Plugin
    {
        /// <summary>
        /// Unique plugin ID: a GUID.
        /// </summary>
        public Guid ID
        {
            get;
            protected set;
        }

        /// <summary>
        /// Plugin description, as appears in the PCC contextual menu.
        /// </summary>
        public string Description
        {
            get;
            protected set;
        }

        /// <summary>
        /// Whether plugin is a separator in the menu.
        /// </summary>
        /// <remarks>
        /// This should never be overridden by subclasses
        /// except <seealso cref="SeparatorPlugin"/>.
        /// </remarks>
        public virtual bool IsSeparator
        {
            get {
                return false;
            }
        }

        //************************************************************************
        /// <summary>
        /// Creates a new Plugin bean.
        /// </summary>
        /// <param name="id">Plugin unique ID.</param>
        /// <param name="description">Plugin description.</param>
        //************************************************************************
        public Plugin(Guid id, string description)
        {
            ID = id;
            Description = description;
        }

        //************************************************************************
        /// <summary>
        /// Creates a new Plugin bean from a string ID.
        /// </summary>
        /// <param name="id">Plugin unique ID's string representation.</param>
        /// <param name="description">Plugin description.</param>
        //************************************************************************
        public Plugin(string id, string description)
        {
            ID = new Guid(id);
            Description = description;
        }

        //************************************************************************
        /// <summary>
        /// Returns a string representation of the plugin (its description).
        /// </summary>
        //************************************************************************
        public override string ToString()
        {
            return Description;
        }
    }

    //************************************************************************
    /// <summary>
    /// Special implementation of the <see cref="Plugin"/> class that represents
    /// a separator in the contextual menu.
    /// </summary>
    //************************************************************************
    public sealed class SeparatorPlugin : Plugin
    {
        /// Unique ID of the separator plugin.
        private static readonly Guid PLUGIN_SEPARATOR_ID = new Guid("{AFA4D1E1-BA73-4330-A6AB-E574FF39ECC3}");

        /// Description used for the separator plugin.
        private const string PLUGIN_SEPARATOR_DESCRIPTION = "-----------------------------------------";

        /// <summary>
        /// Whether plugin is a separator in the menu. In our case, yes.
        /// </summary>
        public override bool IsSeparator
        {
            get {
                return true;
            }
        }

        //************************************************************************
        /// <summary>
        /// Creates a separator plugin bean.
        /// </summary>
        //************************************************************************
        public SeparatorPlugin()
            : base(PLUGIN_SEPARATOR_ID, PLUGIN_SEPARATOR_DESCRIPTION)
        {
        }
    }
}
