﻿// Program.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Windows.Forms;
using PathCopyCopy.Settings;
using PathCopyCopy.Settings.Properties;
using PCC;
using System.Threading;

namespace PathCopyCopySettings
{
    //************************************************************************
    /// <summary>
    /// Main program class.
    /// </summary>
    //************************************************************************
    static class Program
    {
        /// Name of global mutex used for blocking multiple instances.
        const string MUTEX_NAME = "PathCopyCopySettings.Mutex";

        /// Name of global mutex used for blocking multiple update checks.
        const string UPDATE_CHECK_MUTEX_NAME = "PathCopyCopySettings.UpdateCheck.Mutex";

        /// <summary>
        /// Possible commands passed to the executable.
        /// </summary>
        enum Command
        {
            Settings,
            CheckForUpdates
        }

        //************************************************************************
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        //************************************************************************
        [STAThread]
        static void Main(string[] args)
        {
            // Assume default command.
            Command command = Command.Settings;

            // Check arguments. Two numbers should follow: one represents 32/64 bits
            // and the other whether this is wow64.
            if (args.Length > 0) {
                int is64;
                if (int.TryParse(args[0], out is64) && is64 != 0) {
                    PCCEnvironment.Is64Bit = true;
                }
                if (args.Length > 1) {
                    int isWow64;
                    if (int.TryParse(args[1], out isWow64) && isWow64 != 0) {
                        PCCEnvironment.IsWow64 = true;
                    }

                    // Check for command.
                    if (args.Length > 2) {
                        if (args[2] == "--updatecheck") {
                            command = Command.CheckForUpdates;
                        }
                    }
                }
            }

            // Setup visual styles.
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            // Check if we need to simply check for updates.
            bool created;
            if (command == Command.CheckForUpdates) {
                // Make sure we only do this once.
                
                using (Mutex mutex = new Mutex(true, UPDATE_CHECK_MUTEX_NAME, out created)) {
                    if (created) {
                        using (UserSettings settings = new UserSettings()) {
                            try {
                                SoftwareUpdater updater = new SoftwareUpdater(settings);
                                updater.CheckForUpdate(null);
                            } catch (SoftwareUpdateException) {
                                // Silence this. Can't check for updates, too bad.
                            }
                            return;
                        }
                    } else {
                        return;
                    }
                }
            }

            // Block multiple instances.
            using (Mutex mutex = new Mutex(true, MUTEX_NAME, out created)) {
                if (!created) {
                    return;
                }

                // Make sure editing is allowed.
                using (UserSettings settings = new UserSettings()) {
                    if (settings.EditingDisabled) {
                        MessageBox.Show(Resources.EDITING_DISABLED_MESSAGE, Resources.EDITING_DISABLED_TITLE,
                            MessageBoxButtons.OK, MessageBoxIcon.Stop);
                        return;
                    }
                }

                // If we reached this point, run the application.
                Application.Run(new MainForm());
            }
        }
    }
}
