﻿// SoftwareUpdates.cs
// (c) 2012, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using System.Diagnostics;
using System.Xml;
using System.Windows.Forms;
using System.Net;
using System.Reflection;
using PathCopyCopy.Settings;

namespace PCC
{
    //************************************************************************
    /// <summary>
    /// Bean containing info about software updates. Can be read from XML.
    /// </summary>
    //************************************************************************
    [XmlRoot(Namespace = SoftwareUpdateInfo.SOFTWARE_UPDATE_XML_NAMESPACE)]
    public class SoftwareUpdateInfo
    {
        /// <summary>
        /// XML namespace used by the software update info schema.
        /// </summary>
        public const string SOFTWARE_UPDATE_XML_NAMESPACE = "http://pathcopycopy.codeplex.com/xsd/SoftwareUpdate/V1";

        /// <summary>
        /// Version number of the latest build.
        /// </summary>
        [XmlIgnore]
        public Version Version
        {
            get;
            set;
        }

        /// <summary>
        /// Version number's string representation.
        /// </summary>
        [XmlElement("Version")]
        public string VersionAsString
        {
            get {
                return this.Version.ToString();
            }
            set {
                this.Version = new Version(value);
            }
        }

        /// <summary>
        /// Name of the latest build.
        /// </summary>
        [XmlElement]
        public string Name
        {
            get;
            set;
        }

        /// <summary>
        /// URL of the build's web page.
        /// </summary>
        [XmlElement]
        public string Url
        {
            get;
            set;
        }

        /// <summary>
        /// Lines containing build's release notes.
        /// </summary>
        [XmlArray]
        public string[] ReleaseNotes
        {
            get;
            set;
        }

        //************************************************************************
        /// <summary>
        /// Constructor.
        /// </summary>
        //************************************************************************
        public SoftwareUpdateInfo()
        {
            this.ReleaseNotes = new string[0];
        }

        //************************************************************************
        /// <summary>
        /// Reads a <see cref="SoftwareUpdateInfo"/> from XML.
        /// </summary>
        /// <param name="xml">XML to decode.</param>
        /// <returns><see cref="SoftwareUpdateInfo"/> instance.</returns>
        //************************************************************************
        public static SoftwareUpdateInfo FromXML(string xml)
        {
            Debug.Assert(xml != null);

            XmlSerializer serializer = new XmlSerializer(typeof(SoftwareUpdateInfo), SOFTWARE_UPDATE_XML_NAMESPACE);
            using (TextReader textReader = new StringReader(xml)) {
                using (XmlReader xmlReader = XmlReader.Create(textReader)) {
                    return (SoftwareUpdateInfo) serializer.Deserialize(xmlReader);
                }
            }
        }
    }

    //************************************************************************
    /// <summary>
    /// Class used to check for software updates give access to build information
    /// when updates are found.
    /// </summary>
    //************************************************************************
    public class SoftwareUpdater
    {
        /// <summary>
        /// Update channel value to set to disable software updates.
        /// </summary>
        public const string NULL_UPDATE_CHANNEL = "none";

        /// UserSettings object used to access settings.
        private UserSettings userSettings;

        //************************************************************************
        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="userSettings"><see cref="UserSettings"/> object. Cannot
        /// be <c>null</c>.</param>
        //************************************************************************
        public SoftwareUpdater(UserSettings userSettings)
        {
            Debug.Assert(userSettings != null);

            this.userSettings = userSettings;
        }

        //************************************************************************
        /// <summary>
        /// Downloads the latest software update info and checks if there is a
        /// software update available. If so, shows a modal form proposing the
        /// update to the user.
        /// </summary>
        /// <param name="owner">Owner window of the modal form. Can be <c>null</c>.
        /// </param>
        /// <returns><c>true</c> if there was a software update shown, <c>false</c>
        /// if the software is up-to-date.</returns>
        //************************************************************************
        public bool CheckForUpdate(IWin32Window owner)
        {
            bool hasUpdate = false;

            if (userSettings.UpdateChannel.ToLower() != NULL_UPDATE_CHANNEL) {
                try {
                    // Download the latest software update info XML.
                    string updateInfoXml;
                    using (WebClient client = new WebClient()) {
                        updateInfoXml = client.DownloadString(userSettings.UpdateChannel);
                    }

                    // Decode the software update info.
                    SoftwareUpdateInfo updateInfo = SoftwareUpdateInfo.FromXML(updateInfoXml);

                    // Compare the latest version of the software with the version of this assembly,
                    // or to the latest ignored version if there is one (whichever is greater).
                    Version latestVersion = updateInfo.Version;
                    Version thisVersion = Assembly.GetExecutingAssembly().GetName().Version;
                    Version ignoredVersion = userSettings.IgnoredUpdate;
                    if (ignoredVersion != null && ignoredVersion.CompareTo(thisVersion) > 0) {
                        thisVersion = ignoredVersion;
                    }
                    hasUpdate = latestVersion != null && latestVersion.CompareTo(thisVersion) > 0;

                    // If there is a software update, show the modal form.
                    bool ignoreUpdate = false;
                    if (hasUpdate) {
                        using (SoftwareUpdateForm updateForm = new SoftwareUpdateForm()) {
                            ignoreUpdate = updateForm.ShowUpdate(updateInfo, owner);
                        }
                    }

                    // If we need to ignore this update, do so now.
                    if (ignoreUpdate) {
                        userSettings.IgnoredUpdate = latestVersion;
                    }
                } catch (Exception e) {
                    throw new SoftwareUpdateException("Error during CheckForUpdate", e);
                }
            }

            return hasUpdate;
        }
    }

    //************************************************************************
    /// <summary>
    /// Exception used by the <see cref="SoftwareUpdater"/> when an error occurs
    /// when checking for software updates.
    /// </summary>
    //************************************************************************
    public class SoftwareUpdateException : Exception
    {
        //************************************************************************
        /// <summary>
        /// Default constructor.
        /// </summary>
        //************************************************************************
        public SoftwareUpdateException()
            : base()
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with message.
        /// </summary>
        /// <param name="message">Exception message.</param>
        //************************************************************************
        public SoftwareUpdateException(string message)
            : base(message)
        {
        }

        //************************************************************************
        /// <summary>
        /// Constructor with inner exception.
        /// </summary>
        /// <param name="message">Exception message.</param>
        /// <param name="innerException">Inner exception.</param>
        //************************************************************************
        public SoftwareUpdateException(string message, Exception innerException)
            : base(message, innerException)
        {
        }
    }
}
