﻿// UserSettings.cs
// (c) 2011, Charles Lechasseur
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using Microsoft.Win32;
using PathCopyCopy.Settings;
using PathCopyCopy.Settings.Properties;

namespace PCC
{
    //************************************************************************
    /// <summary>
    /// Class used to interact with the user settings of Path Copy Copy.
    /// </summary>
    //************************************************************************
    public class UserSettings : IDisposable
    {
        /// Path of registry key containing global settings.
        private const string PCC_GLOBAL_SETTINGS_KEY = @"Software\clechasseur\PathCopyCopy";

        /// Path of registry key containing global settings when running under WOW64.
        private const string PCC_GLOBAL_SETTINGS_KEY_WOW64 = @"Software\Wow6432Node\clechasseur\PathCopyCopy";

        /// Path of registry key containing user settings.
        private const string PCC_USER_SETTINGS_KEY = @"Software\clechasseur\PathCopyCopy";

        /// Path of registry key containing the pipeline plugins (relative to the main registry key).
        private const string PCC_PIPELINE_PLUGINS_KEY = "PipelinePlugins";

        /// Name of registry value determining whether we use hidden shares in UNC plugins.
        private const string USE_HIDDEN_SHARES_VALUE_NAME = "UseHiddenShares";

        /// Name of registry value determining whether we add quotes to copied paths.
        private const string ADD_QUOTES_VALUE_NAME = "AddQuotes";

        /// Name of registry value determining whether we show an icon next to the default plugin.
        private const string USE_ICON_FOR_DEFAULT_PLUGIN_VALUE_NAME = "UseIconForDefaultPlugin";

        /// Name of registry value determining whether we show an icon next to the submenu.
        private const string USE_ICON_FOR_SUBMENU_VALUE_NAME = "UseIconForSubmenu";

        /// Name of registry value determining whether software updates are disabled.
        private const string DISABLE_SOFTWARE_UPDATE_VALUE_NAME = "DisableSoftwareUpdate";

        /// Name of registry value determining whether the submenu is always shown.
        private const string ALWAYS_SHOW_SUBMENU_VALUE_NAME = "AlwaysShowSubmenu";

        /// Name of registry value containing the IDs of plugins to remove from the submenu.
        private const string PLUGINS_NOT_IN_SUBMENU_VALUE_NAME = "PluginsNotInSubmenu";

        /// Name of registry value containing the IDs of plugins to show in the main contextual menu.
        private const string PLUGINS_IN_MAIN_MENU_VALUE_NAME = "PluginsInMainMenu";

        /// Name of registry value containing a pipeline plugin's description.
        private const string PIPELINE_PLUGIN_DESCRIPTION_VALUE_NAME = "Description";

        /// Name of registry value containing the order in which to display pipeline plugins.
        private const string PIPELINE_PLUGINS_DISPLAY_ORDER_VALUE_NAME = "DisplayOrder";

        /// Name of registry value containing the last ignored software update version.
        private const string IGNORED_UPDATE_VALUE_NAME = "IgnoredUpdate";

        /// Name of registry value that can exist in the global key to indicate
        /// that settings editing is disabled.
        private const string EDITING_DISABLED_VALUE_NAME = "KeyLock";

        /// Name of registry value that can exist in the global key to store
        /// the URL of the software update file to check.
        private const string UPDATE_CHANNEL_VALUE_NAME = "UpdateChannel";

        /// Separator used between plugins in the registry values.
        private const char PLUGINS_VALUE_SEPARATOR = ',';

        /// Default value of the "use hidden shares" setting.
        private const int USE_HIDDEN_SHARES_DEFAULT_VALUE = 0;

        /// Default value of the "add quotes around paths" setting.
        private const int ADD_QUOTES_DEFAULT_VALUE = 0;

        /// Default value of the "use icon for default plugin" setting.
        private const int USE_ICON_FOR_DEFAULT_PLUGIN_DEFAULT_VALUE = 0;

        /// Default value of the "use icon for submenu" setting.
        private const int USE_ICON_FOR_SUBMENU_DEFAULT_VALUE = 1;

        /// Default value of the "always show submenu" setting.
        private const int ALWAYS_SHOW_SUBMENU_DEFAULT_VALUE = 1;

        /// Default value of the "disable software update" setting.
        private const int DISABLE_SOFTWARE_UPDATE_DEFAULT_VALUE = 0;

        /// Separator used between pipeline plugins in the display order string.
        private const char PIPELINE_PLUGINS_DISPLAY_ORDER_SEPARATOR = ',';

        /// Registry key containing global settings for all users.
        private RegistryKey globalKey;

        /// Registry key containing user-specific settings.
        private RegistryKey userKey;

        /// Registry key containing the global pipeline plugins.
        private RegistryKey globalPipelinePluginsKey;

        /// Registry key containing user-specific pipeline plugins.
        private RegistryKey userPipelinePluginsKey;

        /// <summary>
        /// Whether the UNC plugins should use hidden shares or not.
        /// </summary>
        public bool UseHiddenShares
        {
            get {
                return ((int) GetUserOrGlobalValue(USE_HIDDEN_SHARES_VALUE_NAME, USE_HIDDEN_SHARES_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(USE_HIDDEN_SHARES_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// Whether to add quotes around all copied paths.
        /// </summary>
        public bool AddQuotes
        {
            get {
                return ((int) GetUserOrGlobalValue(ADD_QUOTES_VALUE_NAME, ADD_QUOTES_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(ADD_QUOTES_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// Whether to display an icon next to the default plugin in the contextual menu.
        /// </summary>
        public bool UseIconForDefaultPlugin
        {
            get {
                return ((int) GetUserOrGlobalValue(USE_ICON_FOR_DEFAULT_PLUGIN_VALUE_NAME, USE_ICON_FOR_DEFAULT_PLUGIN_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(USE_ICON_FOR_DEFAULT_PLUGIN_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// Whether to display an icon next to the submenu in the contextual menu.
        /// </summary>
        public bool UseIconForSubmenu
        {
            get {
                return ((int) GetUserOrGlobalValue(USE_ICON_FOR_SUBMENU_VALUE_NAME, USE_ICON_FOR_SUBMENU_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(USE_ICON_FOR_SUBMENU_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// Whether to always show the submenu. If not, it will only be shown
        /// when the user holds down the Shift key.
        /// </summary>
        public bool AlwaysShowSubmenu
        {
            get {
                return ((int) GetUserOrGlobalValue(ALWAYS_SHOW_SUBMENU_VALUE_NAME, ALWAYS_SHOW_SUBMENU_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(ALWAYS_SHOW_SUBMENU_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// Whether software updates are disabled.
        /// </summary>
        public bool DisableSoftwareUpdate
        {
            get {
                return ((int) GetUserOrGlobalValue(DISABLE_SOFTWARE_UPDATE_VALUE_NAME, DISABLE_SOFTWARE_UPDATE_DEFAULT_VALUE)) != 0;
            }
            set {
                userKey.SetValue(DISABLE_SOFTWARE_UPDATE_VALUE_NAME, value ? 1 : 0);
            }
        }

        /// <summary>
        /// List of plugins to remove from the PCC submenu.
        /// </summary>
        public List<Guid> PluginsNotInSubmenu
        {
            get {
                return LoadPluginsFromValue(PLUGINS_NOT_IN_SUBMENU_VALUE_NAME, false);
            }
            set {
                SavePluginsInValue(PLUGINS_NOT_IN_SUBMENU_VALUE_NAME, value);
            }
        }

        /// <summary>
        /// List of plugins to show in the main contextual menu.
        /// Will return <c>null</c> if not set.
        /// </summary>
        public List<Guid> PluginsInMainMenu
        {
            get {
                return LoadPluginsFromValue(PLUGINS_IN_MAIN_MENU_VALUE_NAME, true);
            }
            set {
                if (value != null) {
                    SavePluginsInValue(PLUGINS_IN_MAIN_MENU_VALUE_NAME, value);
                } else {
                    // Delete the value in the registry instead.
                    userKey.DeleteValue(PLUGINS_IN_MAIN_MENU_VALUE_NAME, false);
                }
            }
        }

        /// <summary>
        /// List of pipeline plugins. Contains both global and user-specific
        /// ones. Stored in the order they are to be displayed.
        /// </summary>
        public List<PipelinePluginInfo> PipelinePlugins
        {
            get {
                return LoadPipelinePlugins();
            }
            set {
                SavePipelinePlugins(value);
            }
        }

        /// <summary>
        /// Last ignored software update version. Returns <c>null</c> if no update has been ignored.
        /// </summary>
        public Version IgnoredUpdate
        {
            get {
                string ignoredValue = (string) GetUserOrGlobalValue(IGNORED_UPDATE_VALUE_NAME, String.Empty);
                return !String.IsNullOrEmpty(ignoredValue) ? new Version(ignoredValue) : null;
            }
            set {
                if (value != null) {
                    userKey.SetValue(IGNORED_UPDATE_VALUE_NAME, value.ToString(4));
                } else {
                    userKey.DeleteValue(IGNORED_UPDATE_VALUE_NAME);
                }
            }
        }

        /// <summary>
        /// Whether editing settings has been disabled by the administrator.
        /// This is a read-only setting.
        /// </summary>
        public bool EditingDisabled
        {
            get {
                object keyLockValue = globalKey != null ? globalKey.GetValue(EDITING_DISABLED_VALUE_NAME) : null;
                return keyLockValue != null ? ((int) keyLockValue) != 0 : false;
            }
        }

        /// <summary>
        /// URL of the software update channel to use.
        /// </summary>
        public string UpdateChannel
        {
            get {
                object channelValue = globalKey != null ? globalKey.GetValue(UPDATE_CHANNEL_VALUE_NAME) : null;
                return channelValue != null ? (string) channelValue : Resources.UserSettings_DefaultUpdateChannel;
            }
        }

        //************************************************************************
        /// <summary>
        /// Constructor. Creates the registry key right away to read the settings,
        /// both global and user-specific.
        /// </summary>
        //************************************************************************
        public UserSettings()
        {
            // Get path of global key. If we're being called by a 32-bit extension
            // running under WOW64 we'll have to adjust.
            string globalKeyPath;
            if (!PCCEnvironment.Is64Bit && PCCEnvironment.IsWow64) {
                globalKeyPath = PCC_GLOBAL_SETTINGS_KEY_WOW64;
            } else {
                globalKeyPath = PCC_GLOBAL_SETTINGS_KEY;
            }

            // Open the global key for reading only.
            globalKey = Registry.LocalMachine.OpenSubKey(globalKeyPath, false);

            // Open the user key for reading and writing.
            userKey = Registry.CurrentUser.CreateSubKey(PCC_USER_SETTINGS_KEY);

            // Open the global pipeline plugins key for reading only.
            globalPipelinePluginsKey = globalKey.OpenSubKey(PCC_PIPELINE_PLUGINS_KEY, false);

            // Open the user pipeline plugins key for reading and writing.
            userPipelinePluginsKey = userKey.CreateSubKey(PCC_PIPELINE_PLUGINS_KEY);
        }

        //************************************************************************
        /// <summary>
        /// Called when the object is disposed. We close the registry keys.
        /// </summary>
        //************************************************************************
        public void Dispose()
        {
            userPipelinePluginsKey.Close();
            if (globalPipelinePluginsKey != null) {
                globalPipelinePluginsKey.Close();
            }
            userKey.Close();
            if (globalKey != null) {
                globalKey.Close();
            }
        }

        //************************************************************************
        /// <summary>
        /// Function called to load a list of plugin IDs from a registry value.
        /// </summary>
        /// <param name="valueName">Name of registry value containing the data.</param>
        /// <param name="wantsNull">Whether to return <c>null</c> (<c>true</c>)
        /// or an empty list (<c>false</c>) when the registry value is not found.</param>
        /// <returns>List of disabled plugins.</returns>
        //************************************************************************
        private List<Guid> LoadPluginsFromValue(string valueName, bool wantsNull)
        {
            List<Guid> plugins = null;

            // Load registry value containing the plugins.
            object pluginsValue = GetUserOrGlobalValue(valueName);
            if (pluginsValue != null && pluginsValue is string) {
                // Split the values so we get individual IDs.
                string[] pluginIds = ((string) pluginsValue).Split(PLUGINS_VALUE_SEPARATOR);

                // Convert each ID and add it to the list of plugins.
                plugins = new List<Guid>();
                foreach (string id in pluginIds) {
                    try {
                        if (id.Length != 0) {
                            plugins.Add(new Guid(id));
                        }
                    } catch (FormatException) {
                    } catch (OverflowException) {
                    }
                }
            } else if (!wantsNull) {
                // Create an empty list instead.
                plugins = new List<Guid>();
            }

            return plugins;
        }

        //************************************************************************
        /// <summary>
        /// Saves a list of plugin IDs in a user settings' registry value.
        /// </summary>
        /// <param name="valueName">Name of registry value where to save plugins.</param>
        /// <param name="plugins">List of plugins to save.</param>
        //************************************************************************
        private void SavePluginsInValue(string valueName, List<Guid> plugins)
        {
            // Scan guids and build a registry value.
            string regValue = String.Empty;
            foreach (Guid id in plugins) {
                string idAsString = id.ToString("B");
                if (regValue.Length != 0) {
                    regValue += PLUGINS_VALUE_SEPARATOR;
                }
                regValue += idAsString;
            }

            // Save this value content in the registry.
            userKey.SetValue(valueName, regValue);
        }

        //************************************************************************
        /// <summary>
        /// Loads the list of pipeline plugins, both global and user-specific.
        /// </summary>
        /// <returns>List of pipeline plugins, in the order they should be
        /// displayed.</returns>
        //************************************************************************
        private List<PipelinePluginInfo> LoadPipelinePlugins()
        {
            // First load the plugins found in the global and user-specific keys.
            List<PipelinePluginInfo> pluginInfos = new List<PipelinePluginInfo>();
            if (globalPipelinePluginsKey != null) {
                LoadPipelinePluginsFromKey(globalPipelinePluginsKey, ref pluginInfos, true);
            }
            LoadPipelinePluginsFromKey(userPipelinePluginsKey, ref pluginInfos, false);

            // Check if we have a value to display the pipeline plugins in a specific
            // order. There could be one in either the user or global key.
            object displayOrderValue = userPipelinePluginsKey.GetValue(PIPELINE_PLUGINS_DISPLAY_ORDER_VALUE_NAME);
            if (displayOrderValue == null && globalPipelinePluginsKey != null) {
                displayOrderValue = globalPipelinePluginsKey.GetValue(PIPELINE_PLUGINS_DISPLAY_ORDER_VALUE_NAME);
            }
            if (displayOrderValue != null && displayOrderValue is string) {
                // The display order is represented as a comma-separated list of
                // pipeline plugin IDs.
                string[] idsAsString = ((string) displayOrderValue).Split(PIPELINE_PLUGINS_DISPLAY_ORDER_SEPARATOR);
                List<Guid> ids = new List<Guid>();
                foreach (string idAsString in idsAsString) {
                    try {
                        ids.Add(new Guid(idAsString));
                    } catch (FormatException) {
                    } catch (OverflowException) {
                    }
                }

                // Sort the pipeline plugin infos using the display order IDs.
                // Note: to match the C++ implementation, we position unknown
                // pipeline plugins at the end of the display order.
                pluginInfos.Sort((leftInfo, rightInfo) => {
                    int leftIndex = ids.FindIndex(id => id.Equals(leftInfo.Id));
                    int rightIndex = ids.FindIndex(id => id.Equals(rightInfo.Id));
                    if (leftIndex < 0) {
                        leftIndex = int.MaxValue;
                    }
                    if (rightIndex < 0) {
                        rightIndex = int.MaxValue;
                    }
                    return leftIndex - rightIndex;
                });
            }

            return pluginInfos;
        }

        //************************************************************************
        /// <summary>
        /// Saves the list of pipeline plugins. User-specific plugins will be saved
        /// to the registry; global plugins are not modified, but the display order
        /// will be honored.
        /// </summary>
        /// <param name="pipelinePlugins">List of pipeline plugins in the order
        /// they should be displayed.</param>
        //************************************************************************
        private void SavePipelinePlugins(List<PipelinePluginInfo> pipelinePlugins)
        {
            Debug.Assert(pipelinePlugins != null);

            // Build display order string and save it.
            List<string> idsAsString = pipelinePlugins.ConvertAll(plugin => plugin.Id.ToString("B"));
            string displayOrder = string.Join(PIPELINE_PLUGINS_DISPLAY_ORDER_SEPARATOR.ToString(), idsAsString.ToArray());
            userPipelinePluginsKey.SetValue(PIPELINE_PLUGINS_DISPLAY_ORDER_VALUE_NAME, displayOrder);

            // Here's the algo to save the plugins:
            // 1. Save existing list of pipeline plugins in user key.
            // 2. Scan provided list and for each user-specific plugin...
            //    2a. Create registry key for plugin and save info.
            // 3. For each plugin that no longer exists...
            //    3a. Delete its registry key.

            // 1.
            string[] existingPlugins = userPipelinePluginsKey.GetSubKeyNames();

            // 2.
            foreach (var pluginInfo in pipelinePlugins) {
                if (!pluginInfo.Global) {
                    // 2a.
                    using (RegistryKey pluginKey = userPipelinePluginsKey.CreateSubKey(pluginInfo.Id.ToString("B"))) {
                        pluginKey.SetValue(PIPELINE_PLUGIN_DESCRIPTION_VALUE_NAME, pluginInfo.Description);
                        pluginKey.SetValue(null, pluginInfo.EncodedElements);
                    }
                }
            }

            // 3.
            foreach (string existingPlugin in existingPlugins) {
                if (!pipelinePlugins.Exists(plugin => plugin.Id.ToString("B").Equals(existingPlugin))) {
                    // 3a.
                    userPipelinePluginsKey.DeleteSubKeyTree(existingPlugin);
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Loads the list of pipeline plugins stored in the given registry key.
        /// Does not check the display order.
        /// </summary>
        /// <param name="regKey">Registry key from which to load plugins.</param>
        /// <param name="pipelinePlugins">Where to store the pipeline plugins
        /// loaded.</param>
        /// <param name="isGlobal">Whether this is the global key we're parsing.</param>
        //************************************************************************
        private void LoadPipelinePluginsFromKey(RegistryKey regKey,
            ref List<PipelinePluginInfo> pipelinePlugins, bool isGlobal)
        {
            Debug.Assert(regKey != null);
            Debug.Assert(pipelinePlugins != null);

            // Plugins are stored as subkeys of the registry key.
            string[] pluginKeyNames = regKey.GetSubKeyNames();
            foreach (string idAsString in pluginKeyNames) {
                // Convert this ID to a guid and make sure it works.
                try {
                    Guid id = new Guid(idAsString);

                    // Open the subkey and read values for the description
                    // and encoded elements.
                    string description, encodedElements;
                    using (RegistryKey subKey = regKey.OpenSubKey(idAsString, false)) {
                        description = (string) subKey.GetValue(PIPELINE_PLUGIN_DESCRIPTION_VALUE_NAME);
                        encodedElements = (string) subKey.GetValue(null);
                    }

                    // If we made it here we have the plugin info, add it to the list.
                    pipelinePlugins.Add(new PipelinePluginInfo(id, description, encodedElements, isGlobal));
                } catch (FormatException) {
                } catch (OverflowException) {
                }
            }
        }

        //************************************************************************
        /// <summary>
        /// Reads a registry value from the user key first, and if not found,
        /// from the global key.
        /// </summary>
        /// <param name="name">Name of the registry value to load.</param>
        /// <returns>Value, or <c>null</c> if the value is not found.</returns>
        //************************************************************************
        private object GetUserOrGlobalValue(string name)
        {
            object value = userKey.GetValue(name);
            if (value == null && globalKey != null) {
                value = globalKey.GetValue(name);
            }
            return value;
        }

        //************************************************************************
        /// <summary>
        /// Reads a registry value from the user key first, and if not found,
        /// from the global key. If not found at all, a default value is returned.
        /// </summary>
        /// <param name="name">Name of the registry value to load.</param>
        /// <param name="defaultValue">Default value to return if not found.</param>
        /// <returns>Value, or <paramref name="defaultValue"/> if not found.</returns>
        //************************************************************************
        private object GetUserOrGlobalValue(string name, object defaultValue)
        {
            object value = GetUserOrGlobalValue(name);
            return value != null ? value : defaultValue;
        }
    }
}
